#include "callcheck.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <rbx/lua/customfn/customfn.h>
#include "../minhook/MinHook.h"

#include <unordered_map>
#include <iostream>

namespace temple::checks::callcheck
{
	using createevent_fn = HANDLE(__stdcall*)(LPSECURITY_ATTRIBUTES lpEventAttributes, BOOL bManualReset, BOOL bInitialState, LPCSTR lpName);

	int cfunction_handler(std::uintptr_t thread);
	createevent_fn createevent_cb = nullptr;
	std::unordered_map<std::uintptr_t, std::string> pooprecord;
	std::unordered_map<std::uintptr_t, std::uintptr_t> cclosure_map{};

	std::uintptr_t return_addy;
	bool cchooked = false;

	__declspec(naked)
	void createevent_detour()
	{
		__asm
		{
			mov eax, dword ptr[esp]
			cmp eax, return_addy
			jne nrml
			add esp, 4
			push dword ptr[esp + 4]
			call cfunction_handler
			add esp, 4
			retn

			nrml :
			jmp[createevent_cb]
		}
	}

	int cfunction_handler(std::uintptr_t thread)
	{
		const auto& iterator = cclosure_map.find(functions::lua::state::get_closure(thread));

		if (iterator != cclosure_map.end())
		{
			std::uintptr_t custom_fn = iterator->second;

			/*const auto& it = pooprecord.find(custom_fn);

			if (it != pooprecord.end())
			{
				//std::cout << "AH: " << it->second << "\n";
			}*/

			return reinterpret_cast<int(__cdecl*)(std::uintptr_t)>(custom_fn)(thread);
		}
		return 0;
	}

	bool hook_createevent()
	{
		if (cchooked == false)
		{
			return_addy = memdata::memloc::createevent_trigger + 6;

			if (MH_CreateHook(&CreateEventA, &createevent_detour,
				reinterpret_cast<LPVOID*>(&createevent_cb)) != MH_OK)
				return false;

			if (MH_EnableHook(&CreateEventA) != MH_OK)
				return false;
		}
		cchooked = true;

		return true;
	}

	std::optional<std::uintptr_t> get_associated_function(std::uintptr_t closure)
	{
		if (functions::lua::closure::get_source(closure) == memdata::memloc::createevent_trigger)
		{
			const auto& it = cclosure_map.find(closure);

			if (it != cclosure_map.end())
			{
				return it->second;
			}
		}

		return std::nullopt;
	}

	void push_temple_function(std::uintptr_t thread, std::uintptr_t func, int nups, int idx)
	{
		functions::lua::pushcclosure(thread, memdata::memloc::createevent_trigger, nups, 0, 0);

		cclosure_map[functions::lua::index2adr(thread, idx == 0 ? -1 : idx - 1)->v.gcobj] = func;

	}

	void copy_temple_function(std::uintptr_t new_closure, std::uintptr_t orig_closure)
	{
		functions::lua::closure::set_source(new_closure, functions::lua::closure::get_source(orig_closure));

		const auto& it = cclosure_map.find(orig_closure);

		if (it != cclosure_map.end())
		{
			cclosure_map[new_closure] = it->second;
		}
	}

	bool is_temple_cclosure(std::uintptr_t closure)
	{
		if (functions::lua::closure::get_source(closure) == memdata::memloc::createevent_trigger)
			return true;
		else
			return false;
	}
}