#include "bytecode.h"

#include <rbx/memory/functions/functions.h>
#include "zstd/common/xxhash.h"
#include "zstd/zstd.h"
#include <xorstr.h>

#include <vector>

namespace temple::lua::bytecode
{
	std::optional<std::string> decompress_bytecode(const std::string& code)
	{
		if (code.length() < 16)
			return std::nullopt;

		std::string buffer = code;

		const char* bytecode_magic = xorstr("RSB1");
		std::uint32_t bytecode_hashmul = 41;
		std::uint32_t bytecode_hashseed = 42;

		char hashbytes[4];
		std::memcpy(hashbytes, &buffer[0], 4);

		for (std::size_t i = 0; i < 4; ++i)
		{
			hashbytes[i] ^= bytecode_magic[i];
			hashbytes[i] -= i * bytecode_hashmul;
		}

		for (std::size_t i = 0; i < buffer.size(); ++i)
			buffer[i] ^= hashbytes[i % 4] + i * bytecode_hashmul;

		unsigned int hash;
		std::memcpy(&hash, hashbytes, 4);

		auto rehash = XXH32(&buffer[0], buffer.size(), bytecode_hashseed);

		if (rehash != hash)
			return std::nullopt;

		int data_size;
		std::memcpy(&data_size, &buffer[4], 4);

		std::vector<char> data(data_size);

		ZSTD_decompress(&data[0], data_size, &buffer[0] + 8, buffer.size() - 8);

		return std::string (data.begin(), data.end());
	}
}