#include "customfn.h"

#include <xorstr.h>

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <util/bitwise.h>

namespace temple::customfn
{
	int bdiv(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) /
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int badd(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) +
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bsub(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) -
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bmul(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) *
			functions::lua::touinteger32(thread, 2);

		//std::cout << "num: " << (static_cast<std::uint64_t>(functions::lua::tointeger32(thread, 1))) << "\n";

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int band(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) &
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bor(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) |
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bxor(std::uintptr_t thread)
	{
		//std::cout << "xor\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) ^
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bnot(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);

		auto result = functions::lua::touinteger32(thread, 1);
	
		result = ~result;

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int bswap(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);

		std::uint32_t result = util::bitwise::bswap(functions::lua::touinteger32(thread, 1));

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int rol(std::uintptr_t thread)
	{
		//std::cout << "rol\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		auto result = util::bitwise::rotl32(functions::lua::touinteger32(thread, 1),
			functions::lua::touinteger32(thread, 2));

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int ror(std::uintptr_t thread)
	{
		//std::cout << "ror\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		auto result = util::bitwise::rotr32(functions::lua::touinteger32(thread, 1),
			functions::lua::touinteger32(thread, 2));

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int tohex(std::uintptr_t thread)
	{
		//std::cout << "tohex\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);

		auto result = util::bitwise::to_hex(functions::lua::touinteger32(thread, 1));

		functions::lua::pushstring(thread, result.c_str());

		return 1;
	}

	int tobit(std::uintptr_t thread)
	{
		//std::cout << "tobit\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);

		auto n = functions::lua::touinteger32(thread, 1);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(n));

		return 1;
	}

	int lshift(std::uintptr_t thread)
	{
		//std::cout << "lshift\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) <<
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int rshift(std::uintptr_t thread)
	{
		//std::cout << "rshift\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		std::uint32_t result = functions::lua::touinteger32(thread, 1) >>
			functions::lua::touinteger32(thread, 2);

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	int arshift(std::uintptr_t thread)
	{
		//std::cout << "arshift\n";
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tnumber);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tnumber);

		auto n = functions::lua::tointeger32(thread, 1);
		auto disp = static_cast<std::int32_t>(functions::lua::tonumber(thread, 2));
		auto result = n;

		if (disp > 0)
		{
			result >>= disp;
		}
		else if(disp < 0)
		{
			result <<= abs(disp);
		}

		if (disp > 0)
		{
			if (n & 0x80000000)
			{
				result |= 0x80000000;
			}
		}

		//std::cout << std::hex << result << "\n";
		//std::cout << std::dec << static_cast<std::int32_t>(result) << "\n";

		functions::lua::pushinteger(thread, static_cast<std::int32_t>(result));

		return 1;
	}

	namespace initialiser
	{
		void load_bit_functions(temple::lua::state thread)
		{
			auto bit_str = xorstr("bit");

			functions::lua::createtable(thread, 0, 0);
			functions::lua::setfield(thread, -10002, bit_str);

			set_table_function(thread, bdiv, xorstr("bdiv"), bit_str);
			set_table_function(thread, badd, xorstr("badd"), bit_str);
			set_table_function(thread, bsub, xorstr("bsub"), bit_str);
			set_table_function(thread, bmul, xorstr("bmul"), bit_str);
			set_table_function(thread, band, xorstr("band"), bit_str);
			set_table_function(thread, bor, xorstr("bor"),  bit_str);
			set_table_function(thread, bxor, xorstr("bxor"), bit_str);
			set_table_function(thread, bnot, xorstr("bnot"), bit_str);
			set_table_function(thread, bswap, xorstr("bswap"), bit_str);
			set_table_function(thread, rol, xorstr("rol"), bit_str);
			set_table_function(thread, ror, xorstr("ror"), bit_str);
			set_table_function(thread, tohex, xorstr("tohex"), bit_str);
			set_table_function(thread, tobit, xorstr("tobit"), bit_str);
			set_table_function(thread, lshift, xorstr("lshift"), bit_str);
			set_table_function(thread, rshift, xorstr("rshift"), bit_str);
			set_table_function(thread, arshift, xorstr("arshift"), bit_str);
		}
	}
}