#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <util/console.h>
#include <xorstr.h>

namespace temple::customfn
{
	using namespace memdata::lua::types;
	int consolecreate(std::uintptr_t thread)
	{
		util::console::create();

		return 0;
	}

	int consoledestroy(std::uintptr_t thread)
	{
		util::console::destroy();

		return 0;
	}

	std::string format_datatype(std::uintptr_t thread, int idx)
	{
		auto type = functions::lua::type(thread, idx);

		if (type == lua_tnil)
		{
			return "nil";
		}
		else if (type == lua_tboolean)
		{
			return "boolean: " + functions::lua::toboolean(thread, idx) ? "true" : "false";
		}
		else if (type == lua_tnumber)
		{
			return "number: " + std::to_string(functions::lua::tonumber(thread, idx));
		}
		else if (type == lua_tlightuserdata)
		{
			return "lightuserdata";
		}
		else if (type == lua_tvector)
		{
			auto obj = functions::lua::index2adr(thread, idx);

			return "vector: " + std::to_string(obj->v.v[0]) + ", " + std::to_string(obj->v.v[1]) + ", " + std::to_string(*reinterpret_cast<float*>(&obj->extra));
		}
		else if (type == lua_tstring)
		{
			return functions::lua::tostring(thread, idx);
		}
		else if (type == lua_tfunction)
		{
			auto name = functions::lua::closure::get_name(functions::lua::index2adr(thread, idx)->v.gcobj);
			if (!name)
				name = "unnamed";

			return "function: " + std::string(name);
		}
		else if (type == lua_tuserdata)
		{
			std::string content;

			if (*reinterpret_cast<std::uint8_t*>(functions::lua::index2adr(thread, 1)->v.gcobj + memdata::offsets::lua::userdata::tag) == 11)
			{
				auto instance = *reinterpret_cast<rbx::instance**>(functions::lua::touserdata(thread, idx));
				content = "name: " + functions::instance::get_name(instance) +
					" classname: " + functions::instance::get_classname(instance);
			}
			else
			{
				content = "classname: " + functions::lua::getobjecttype(thread, 1);
			}
			return "userdata: " + content;
		}
		else if (type == lua_ttable)
		{
			return "table: size: " + std::to_string(functions::lua::table::get_sizearray(functions::lua::index2adr(thread, 1)->v.gcobj));
		}
		else if (type == lua_tthread)
		{
			return "thread";
		}
		return "unknown";
	}

	int rconsoleprint(std::uintptr_t thread)
	{
		get_argument(thread, 1, 0, true, false);

		util::console::print(format_datatype(thread, 1));

		return 0;
	}

	int rconsolename(std::uintptr_t thread)
	{
		get_argument(thread, 1, memdata::lua::types::lua_tstring);

		util::console::set_title(functions::lua::tostring(thread, 1));

		return 0;
	}

	int rconsoleinfo(std::uintptr_t thread)
	{
		get_argument(thread, 1, 0, true, false);

		util::console::print("@@LIGHT_CYAN@@");
		util::console::print("[INFO]: ", false);
		util::console::print("@@WHITE@@");
		util::console::print(format_datatype(thread, 1));

		return 0;
	}

	int rconsolewarn(std::uintptr_t thread)
	{
		get_argument(thread, 1, 0, true, false);

		util::console::print("@@YELLOW@@");
		util::console::print("[WARNING]: ", false);
		util::console::print("@@WHITE@@");
		util::console::print(format_datatype(thread, 1));

		return 0;
	}

	int rconsoleerr(std::uintptr_t thread)
	{
		get_argument(thread, 1, 0, true, false);

		util::console::print("@@LIGHT_RED@@");
		util::console::print("[ERROR]: ", false);
		util::console::print("@@WHITE@@");
		util::console::print(format_datatype(thread, 1));

		return 0;
	}

	int rconsoleinput(std::uintptr_t thread)
	{
		std::string out;
		std::getline(std::cin, out);

		functions::lua::pushstring(thread, out.c_str());

		return 1;
	}

	int rconsoleclear(std::uintptr_t thread)
	{
		system("cls");

		return 0;
	}

	namespace initialiser
	{
		void load_console_functions(temple::lua::state thread)
		{
			set_global_function(thread, consoledestroy, xorstr("consoledestroy"));
			set_global_function(thread, rconsoleprint, xorstr("rconsoleprint"));
			set_global_function(thread, rconsoleclear, xorstr("rconsoleclear"));
			set_global_function(thread, rconsoleinput, xorstr("rconsoleinput"));
			set_global_function(thread, consolecreate, xorstr("consolecreate"));
			set_global_function(thread, rconsoleinfo, xorstr("rconsoleinfo"));
			set_global_function(thread, rconsolewarn, xorstr("rconsolewarn"));
			set_global_function(thread, rconsolename, xorstr("rconsolename"));
			set_global_function(thread, rconsoleerr, xorstr("rconsoleerr"));
		}
	}
}