#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <themida/ThemidaSDK.h>
#include <rbx/memory/memdata/memdata.h>
#include "../execution.h"
#include <xorstr.h>

#include <string>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	std::string get_typename(std::int32_t type)
	{

		if (type == lua_tnil)
		{
			return xorstr("nil");
		}
		else if (type == lua_tboolean)
		{
			return xorstr("boolean");
		}
		else if (type == lua_tlightuserdata)
		{
			return xorstr("lightuserdata");
		}
		else if (type == lua_tvector)
		{
			return xorstr("vector");
		}
		else if (type == lua_tnumber)
		{
			return xorstr("number");
		}
		else if (type == lua_tstring)
		{
			return xorstr("string");
		}
		else if (type == lua_tfunction)
		{
			return xorstr("function");
		}
		else if (type == lua_tthread)
		{
			return xorstr("thread");
		}
		else if (type == lua_ttable)
		{
			return xorstr("table");
		}
		else if (type == lua_tuserdata)
		{
			return xorstr("userdata");
		}

		return xorstr("unknown");
	}

	lua::tvalue* get_argument(temple::lua::state thread, int idx, int type, bool extraarg, bool typematters, bool one_type)
	{
		if (functions::lua::gettop(*thread) >= idx)
		{
			auto tv = functions::lua::index2adr(*thread, idx);

			if (tv->tt == type || !typematters)
			{
				return tv;
			}
			else if(one_type)
			{
				std::string exception = xorstr("invalid argument #");
				exception += std::to_string(idx) + " (" + get_typename(type) + xorstr(" expected, got ") + get_typename(tv->tt) + ")";

				throw std::exception(exception.c_str());
			}
		}
		else
		{
			if (extraarg)
			{
				std::string exception = xorstr("missing argument #");
				exception += std::to_string(idx) + " (" + (typematters ? get_typename(type) : xorstr("any type")) + xorstr(" expected)");

				throw std::exception(exception.c_str());
			}
		}

		return 0; //wont reach here anyways :lmao3d:
	}

	void set_global_function(temple::lua::state thread, int(__cdecl *fn)(std::uintptr_t), const char* name)
	{
		checks::callcheck::push_temple_function(*thread, reinterpret_cast<std::uintptr_t>(fn));
		functions::lua::setfield(*thread, -10002, name);

		//checks::callcheck::pooprecord[std::uintptr_t(fn)] = std::string{ name };
	}

	void set_table_function(temple::lua::state thread, int(__cdecl* fn)(std::uintptr_t), const char* name, const char* table)
	{
		functions::lua::getfield(thread, -10002, table);
		checks::callcheck::push_temple_function(*thread, reinterpret_cast<std::uintptr_t>(fn));
		functions::lua::setfield(thread, -2, name);

		functions::lua::pop(thread, 1);
	}

	void set_subtable_function(temple::lua::state thread, int(__cdecl* fn)(std::uintptr_t), const char* name, const char* table, const char* subtable)
	{
		functions::lua::getfield(thread, -10002, table);
		functions::lua::getfield(thread, -1, subtable);
		checks::callcheck::push_temple_function(*thread, reinterpret_cast<std::uintptr_t>(fn));
		functions::lua::setfield(thread, -2, name);

		functions::lua::pop(thread, 2);
	}

	void load_custom_functions(temple::lua::state thread)
	{
		functions::lua::createtable(thread, 0, 0);
		functions::lua::setfield(thread, -10000, xorstr("tmp_fc"));
		VM_START;
		initialiser::load_crypto_functions(thread);
		initialiser::load_reflection_functions(thread);
		initialiser::load_drawing_functions(thread);
		initialiser::load_debug_functions(thread);
		initialiser::load_rlua_functions(thread);
		initialiser::load_win_functions(thread);
		initialiser::load_bit_functions(thread);
		initialiser::load_temple_functions(thread);
		initialiser::load_rbx_functions(thread);
		initialiser::load_console_functions(thread);
		VM_END;
	}
}
