#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <xorstr.h>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	int getsizeconstants(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
			functions::lua::pushnumber(thread, functions::lua::proto::get_sizek(proto));
		else
			functions::lua::pushnil(thread);

		return 1;
	}

	int getconstant(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);

		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
		{
			temple::lua::tvalue obj;
			
			if (!functions::lua::proto::get_k(proto, static_cast<int>(functions::lua::tonumber(thread, 2)), obj))
				throw std::exception(xorstr("Constant index is out of bounds"));

			auto& top = functions::lua::state::get_top(thread);
			top->v = obj.v;
			top->tt = obj.tt;

			top++;
		}
		else
			functions::lua::pushnil(thread);

		return 1;
	}

	int setconstant(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);
		auto arg3 = get_argument(thread, 3, 0, 1, 0);

		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
			if (!functions::lua::proto::set_k(proto, static_cast<int>(functions::lua::tonumber(thread, 2)), *arg3))
				throw std::exception(xorstr("Constant index is out of bounds"));

		return 0;
	}

	int getsizeupvalues(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		functions::lua::pushnumber(thread, functions::lua::closure::get_nups(arg1->v.gcobj));

		return 1;
	}

	int getupvalue(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);

		temple::lua::tvalue obj;
		if(!functions::lua::closure::get_ups(arg1->v.gcobj, static_cast<int>(functions::lua::tonumber(thread, 2)), obj))
			throw std::exception(xorstr("Upvalue index is out of bounds"));

		auto& top = functions::lua::state::get_top(thread);
		top->v = obj.v;
		top->tt = obj.tt;

		top++;

		return 1;
	}

	int setupvalue(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);
		auto arg3 = get_argument(thread, 3, 0, 1, 0);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
		{
			if (!functions::lua::closure::set_ups(arg1->v.gcobj, static_cast<int>(functions::lua::tonumber(thread, 2)), *arg3))
				throw std::exception(xorstr("Constant index is out of bounds"));
		}
		else
		{
			throw std::exception(xorstr("can not set upvalues of a c function"));
		}
		
		return 0;
	}

	int getsizeprotos(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
			functions::lua::pushnumber(thread, functions::lua::proto::get_sizep(proto));
		else
			functions::lua::pushnil(thread);

		return 1;
	}

	int getproto(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);
		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
		{
			temple::lua::tvalue obj;
			if(!functions::lua::proto::get_p(thread, proto, static_cast<int>(functions::lua::tonumber(thread, 2)), obj))
				throw std::exception(xorstr("Constant index is out of bounds"));

			auto& top = functions::lua::state::get_top(thread);
			top->v = obj.v;
			top->tt = obj.tt;

			top++;
		}
		else
		{
			throw std::exception(xorstr("can not get proto of a c function"));
		}

		return 1;
	}

	int setproto(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);
		auto arg2 = get_argument(thread, 2, lua_tnumber);
		auto arg3 = get_argument(thread, 3, lua_tfunction, 1, 0);
		auto proto = functions::lua::closure::get_proto(arg1->v.gcobj);

		if (!functions::lua::closure::iscclosure(arg1->v.gcobj))
			if (!functions::lua::proto::set_p(proto, static_cast<int>(functions::lua::tonumber(thread, 2)), functions::lua::closure::get_proto(arg3->v.gcobj)))
				throw std::exception(xorstr("Constant index is out of bounds"));

		return 0;
	}

	int getclosurename(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		functions::lua::pushstring(thread, functions::lua::closure::get_name(arg1->v.gcobj));

		return 1;
	}

	int getclosuresource(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		functions::lua::pushstring(thread, functions::lua::proto::get_source(arg1->v.gcobj));

		return 1;
	}

	int getinfo(std::uintptr_t thread)
	{
		int idx = 0;

		if (functions::lua::gettop(thread) >= 1)
		{
			auto type = functions::lua::type(thread, 1);

			if (type == memdata::lua::types::lua_tfunction)
			{
				idx = -functions::lua::gettop(thread);
			}
			else if (type == memdata::lua::types::lua_tnumber)
			{
				idx = functions::lua::tonumber(thread, 1);
			}
			else
			{
				get_argument(thread, 1, lua_tfunction);
			}
		}
		else
		{
			get_argument(thread, 1, lua_tfunction);
		}

		auto arg2 = get_argument(thread, 2, lua_tstring, false);

		const char* desc = xorstr("slaun");

		if (arg2)
		{
			desc = functions::lua::tostring(thread, 2);
		}

		temple::lua::lua_Debug ar1;

		functions::lua::getinfo(thread, desc, &ar1, idx);

		functions::lua::createtable(thread, 0, 0);

		if (std::strchr(desc, 'u'))
		{
			functions::lua::pushnumber(thread, ar1.nupvals);
			functions::lua::setfield(thread, -2, xorstr("nups"));
		}

		if (std::strchr(desc, 'n'))
		{
			functions::lua::pushstring(thread, ar1.name);
			functions::lua::setfield(thread, -2, xorstr("name"));
		}

		if (std::strchr(desc, 'l'))
		{
			functions::lua::pushnumber(thread, ar1.currentline);
			functions::lua::setfield(thread, -2, xorstr("currentline"));
		}

		if (std::strchr(desc, 's'))
		{
			functions::lua::pushstring(thread, ar1.source);
			functions::lua::setfield(thread, -2, xorstr("source"));

			functions::lua::pushstring(thread, ar1.short_src);
			functions::lua::setfield(thread, -2, xorstr("short_src"));

			functions::lua::pushstring(thread, ar1.what);
			functions::lua::setfield(thread, -2, xorstr("what"));

			functions::lua::pushnumber(thread, ar1.linedefined);
			functions::lua::setfield(thread, -2, xorstr("linedefined"));
		}


		if (std::strchr(desc, 'a'))
		{
			functions::lua::pushnumber(thread, static_cast<std::int32_t>(ar1.isvararg));
			functions::lua::setfield(thread, -2, xorstr("isvararg"));

			functions::lua::pushnumber(thread, ar1.nparams);
			functions::lua::setfield(thread, -2, xorstr("numparams"));
		}

		if (functions::lua::type(thread, 1) == memdata::lua::types::lua_tfunction)
		{
			functions::lua::pushvalue(thread, 1);
			functions::lua::setfield(thread, -2, xorstr("func"));
		}

		return 1;
	}

	int getstack(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);
		auto arg2 = get_argument(thread, 2, lua_tnumber, false);

		auto stack_idx = static_cast<int>(functions::lua::tonumber(thread, 1));
		auto obj_idx = static_cast<int>(functions::lua::tonumber(thread, 2));

		auto ci_base = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);
		//std::cout << (arg1) << "\n";

		if (stack_idx <= 0)
			throw std::exception(xorstr("out of range"));

		if (obj_idx <= 0)
			throw std::exception(xorstr("out of range"));

		ci_base += 24 * stack_idx;

		if (ci_base > *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::end_ci))
			throw std::exception(xorstr("stack out of range"));

		auto cifn = (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::func));

		if (functions::lua::closure::iscclosure(cifn->v.gcobj))
			throw std::exception(xorstr("can not use getstack on C function"));
	
		auto cibase = (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::base));
		auto citop = (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::top));

		size_t idx = 0;

		if (!arg2)
		{
			functions::lua::createtable(thread, 0, 0);

			while (cibase < citop);
			{
				auto& top = functions::lua::state::get_top(thread);;
				top->v = cibase->v;
				top->tt = cibase->tt;

				top++;

				functions::lua::rawseti(thread, -2, ++idx);

				cibase++;

			}
		}
		else
		{
			if(cibase + obj_idx < citop)
			{ 
				cibase += obj_idx;

				auto& top = functions::lua::state::get_top(thread);;
				top->v = cibase->v;
				top->tt = cibase->tt;

				top++;
			}

		}

		return 1;
	}

	int setstack(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);
		auto arg2 = get_argument(thread, 2, lua_tnumber);
		auto arg3 = get_argument(thread, 3, 0, 1, 0);

		auto ci_base = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);
		//std::cout << (arg1) << "\n";

		auto stack_idx = static_cast<int>(functions::lua::tonumber(thread, 1));
		auto obj_idx = static_cast<int>(functions::lua::tonumber(thread, 2));

		if (stack_idx <= 0 || obj_idx <= 0)
			throw std::exception(xorstr("out of range"));

		ci_base += 24 * stack_idx;

		if (ci_base >= *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::end_ci))
			throw std::exception(xorstr("stack out of range"));

		auto base = (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::base));

		base +=  obj_idx;

		if (base >= (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::top)))
			throw std::exception(xorstr("index out of range"));

		auto cifn = (*reinterpret_cast<temple::lua::tvalue**>(ci_base + memdata::offsets::lua::state::callinfo::func));

		if (functions::lua::closure::iscclosure(cifn->v.gcobj))
			throw std::exception(xorstr("can not use getstack on C function"));
		
		if (base->tt == arg3->tt)
		{
			base->v = arg3->v;
			base->tt = arg3->tt;
		}
		else
		{
			throw std::exception(xorstr("source type does not match value type"));
		}

		return 0;
	}


	namespace initialiser
	{
		void load_debug_functions(temple::lua::state thread)
		{
			auto debug_str = xorstr("temple_debug");

			functions::lua::createtable(thread, 0, 0);
			functions::lua::setfield(thread, -10002, debug_str);

			set_table_function(thread, getsizeconstants, xorstr("getsizeconstants"), debug_str);
			set_table_function(thread, getsizeupvalues, xorstr("getsizeupvalues"), debug_str);
			set_table_function(thread, getsizeprotos, xorstr("getsizeprotos"), debug_str);

			set_table_function(thread, setconstant, xorstr("setconstant"), debug_str);
			set_table_function(thread, setupvalue, xorstr("setupvalue"), debug_str);
			set_table_function(thread, setstack, xorstr("setstack"), debug_str);

			set_table_function(thread, getconstant, xorstr("getconstant"), debug_str);
			set_table_function(thread, getupvalue, xorstr("getupvalue"), debug_str);
			set_table_function(thread, getproto, xorstr("getproto"), debug_str);
			set_table_function(thread, getstack, xorstr("getstack"), debug_str);

			set_table_function(thread, getclosuresource, xorstr("getclosuresource"), debug_str);
			set_table_function(thread, getclosurename, xorstr("getclosurename"), debug_str);
			set_table_function(thread, getinfo, xorstr("getinfo"), debug_str);
		}
	}
}