#include "customfn.h"

#include <themida/ThemidaSDK.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <overlay/overlay.h>
#include <xorstr.h>

#include <elements/elements.h>

using namespace temple::overlay;

namespace temple::customfn
{
	void create_color3(temple::lua::state thread, std::uint8_t r, std::uint8_t g, std::uint8_t b)
	{
		functions::lua::getfield(thread, -10002, xorstr("Color3"));
		functions::lua::getfield(thread, -1, xorstr("fromRGB"));
		functions::lua::pushnumber(thread, r);
		functions::lua::pushnumber(thread, g);
		functions::lua::pushnumber(thread, b);
		functions::lua::pcall(thread, 3, 1, 0);
	}

	void create_vector2(temple::lua::state thread, float x, float y)
	{
		functions::lua::getfield(thread, -10002, xorstr("Vector2"));
		functions::lua::getfield(thread, -1, xorstr("new"));
		functions::lua::pushnumber(thread, x);
		functions::lua::pushnumber(thread, y);
		functions::lua::pcall(thread, 2, 1, 0);
	}

	elements::color3 get_color3(temple::lua::state thread, int idx)
	{
		elements::color3 color{};

		functions::lua::getfield(thread, idx, xorstr("R"));
		color.r = functions::lua::tonumber(thread, -1) * 255;
		functions::lua::pop(thread, 1);

		functions::lua::getfield(thread, idx, xorstr("G"));
		color.g = functions::lua::tonumber(thread, -1) * 255;
		functions::lua::pop(thread, 1);

		functions::lua::getfield(thread, idx, xorstr("B"));
		color.b = functions::lua::tonumber(thread, -1) * 255;
		functions::lua::pop(thread, 1);

		return color;
	}

	elements::vector2 get_vector2(temple::lua::state thread, int idx)
	{
		elements::vector2 vector{};

		functions::lua::getfield(thread, idx, xorstr("X"));
		vector.x = functions::lua::tonumber(thread, -1);
		functions::lua::pop(thread, 1);

		functions::lua::getfield(thread, idx, xorstr("Y"));
		vector.y = functions::lua::tonumber(thread, -1);
		functions::lua::pop(thread, 1);

		return vector;
	}

	int drawing_remove(std::uintptr_t thread)
	{
		if (functions::lua::gettop(thread) >= 1 && functions::lua::type(thread, 1) == memdata::lua::types::lua_tuserdata)
		{
			auto element = *reinterpret_cast<elements::base**>(functions::lua::touserdata(thread, 1));

			element->remove();
		}

		return 0;
	}

	int drawing_index(std::uintptr_t thread)
	{
		auto element = *reinterpret_cast<elements::base**>(functions::lua::touserdata(thread, 1));
		auto idx = std::string(functions::lua::tostring(thread, 2));

		if (idx  == xorstr("Visible"))
		{
			functions::lua::pushboolean(thread, element->visible);

			return 1;
		}
		else if (idx == xorstr("Transparency"))
		{
			functions::lua::pushnumber(thread, element->color.a / 255);

			return 1;
		}
		else if (idx == xorstr("Color"))
		{
			create_color3(thread, element->color.r, element->color.g, element->color.b);

			return 1;
		}
		else if (idx == xorstr("Remove"))
		{
			checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(drawing_remove));

			return 1;
		}

		switch (element->get_type())
		{
		case elements::element_type::line:
		{
			auto line = static_cast<elements::line*>(element);

			if (idx == xorstr("Thickness"))
			{
				functions::lua::pushnumber(thread, line->thickness);
			}
			else if (idx == xorstr("From"))
			{
				create_vector2(thread, line->p1.x, line->p1.y);
			}
			else if (idx == xorstr("To"))
			{
				create_vector2(thread, line->p2.x, line->p2.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			break;

		}
		case elements::element_type::circle:
		{
			auto circle = static_cast<elements::circle*>(element);

			if (idx == xorstr("Thickness"))
			{
				functions::lua::pushnumber(thread, circle->thickness);
			}
			else if (idx == xorstr("Radius"))
			{
				functions::lua::pushnumber(thread, circle->radius);
			}
			else if (idx == xorstr("NumSides"))
			{
				functions::lua::pushnumber(thread, circle->edges);
			}
			else if (idx == xorstr("Filled"))
			{
				functions::lua::pushboolean(thread, circle->filled);
			}
			else if (idx == xorstr("Position"))
			{
				create_vector2(thread, circle->position.x, circle->position.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			break;
		}
		case elements::element_type::square:
		{
			auto square = static_cast<elements::square*>(element);
			if (idx == xorstr("Thickness"))
			{
				functions::lua::pushnumber(thread, square->thickness);
			}
			else if (idx == xorstr("Filled"))
			{
				functions::lua::pushboolean(thread, square->filled);
			}
			else if (idx == xorstr("Position"))
			{
				create_vector2(thread, square->position.x, square->position.y);
			}
			else if (idx == xorstr("Size"))
			{
				create_vector2(thread, square->size.x, square->size.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::triangle:
		{
			auto triangle = static_cast<elements::triangle*>(element);
			if (idx == xorstr("Thickness"))
			{
				functions::lua::pushnumber(thread, triangle->thickness);
			}
			else if (idx == xorstr("Filled"))
			{
				functions::lua::pushboolean(thread, triangle->filled);
			}
			else if (idx == xorstr("PointA"))
			{
				create_vector2(thread, triangle->p1.x, triangle->p1.y);
			}
			else if (idx == xorstr("PointB"))
			{
				create_vector2(thread, triangle->p2.x, triangle->p2.y);
			}
			else if (idx == xorstr("PointC"))
			{
				create_vector2(thread, triangle->p3.x, triangle->p3.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::quad:
		{
			auto quad = static_cast<elements::quad*>(element);
			if (idx == xorstr("Thickness"))
			{
				functions::lua::pushnumber(thread, quad->thickness);
			}
			else if (idx == xorstr("Filled"))
			{
				functions::lua::pushboolean(thread, quad->filled);
			}
			else if (idx == xorstr("PointA"))
			{
				create_vector2(thread, quad->p1.x, quad->p1.y);
			}
			else if (idx == xorstr("PointB"))
			{
				create_vector2(thread, quad->p2.x, quad->p2.y);
			}
			else if (idx == xorstr("PointC"))
			{
				create_vector2(thread, quad->p3.x, quad->p3.y);
			}
			else if (idx == xorstr("PointD"))
			{
				create_vector2(thread, quad->p4.x, quad->p4.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::text:
		{
			auto text = static_cast<elements::text*>(element);
			if (idx == xorstr("Text"))
			{
				functions::lua::pushstring(thread, text->content.c_str());
			}
			else if (idx == xorstr("Font"))
			{
				functions::lua::pushstring(thread, text->font.c_str());
			}
			else if (idx == xorstr("Size"))
			{
				functions::lua::pushnumber(thread, text->size);
			}
			else if (idx == xorstr("Outline"))
			{
				functions::lua::pushboolean(thread, text->outline);
			}
			else if (idx == xorstr("OutlineColor"))
			{
				create_color3(thread, text->outline_color.r, text->outline_color.g, text->outline_color.b);
			}
			else if (idx == xorstr("Center"))
			{
				functions::lua::pushboolean(thread, text->center);
			}
			else if (idx == xorstr("Position"))
			{
				create_vector2(thread, text->position.x, text->position.y);
			}
			else if (idx == xorstr("TextBounds"))
			{
				create_vector2(thread, text->text_bounds.x, text->text_bounds.y);
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			break;
		}
		default:
		{
			throw std::exception(xorstr("Why the fuck would you call this index with that"));
			break;
		}
		}

		return 1;
	}

	void verify_type(temple::lua::state thread, int idx, std::string expected_type)
	{
		auto type = functions::lua::getobjecttype(thread, idx);

		if(type != expected_type)
			throw std::exception((xorstr("invalid argument #3 (") + expected_type + xorstr(" expected, got ") + type + ")").c_str());
	}

	int drawing_newindex(std::uintptr_t thread)
	{
		auto element = *reinterpret_cast<elements::base**>(functions::lua::touserdata(thread, 1));
		auto idx = std::string(functions::lua::tostring(thread, 2));

		if (idx == xorstr("Transparency"))
		{
			verify_type(thread, -1, xorstr("number"));

			auto val = functions::lua::tonumber(thread, -1) * 255;
			element->color.a = val > 255 ? 255 : val;

			return 0;
		}
		else if (idx == xorstr("ZIndex"))
		{
			verify_type(thread, -1, xorstr("number"));

			auto val = functions::lua::tonumber(thread, -1);
			element->move(val);

			return 0;
		}
		else if (idx == xorstr("Visible"))
		{
			verify_type(thread, -1, xorstr("boolean"));

			auto val = functions::lua::toboolean(thread, -1);
			element->visible = val;

			return 0;
		}
		else if (idx == xorstr("Color"))
		{
			verify_type(thread, -1, xorstr("Color3"));

			auto color = get_color3(thread, -1);

			element->color = color;
			return 0;
		}

		switch (element->get_type())
		{
		case elements::element_type::line:
		{
			auto line = static_cast<elements::line*>(element);

			if (idx == xorstr("Thickness"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				line->thickness = val;
			}
			else if (idx == xorstr("From"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				line->p1 = vector;
			}
			else if (idx == xorstr("To"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				line->p2 = vector;
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			break;

		}
		case elements::element_type::circle:
		{
			auto circle = static_cast<elements::circle*>(element);

			if (idx == xorstr("Thickness"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				circle->thickness = val;
			}
			else if (idx == xorstr("Radius"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				circle->radius = val;
			}
			else if (idx == xorstr("NumSides"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				circle->edges = val;
			}
			else if (idx == xorstr("Filled"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				circle->filled = val;
			}
			else if (idx == xorstr("Position"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				circle->position = vector;
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			break;
		}
		case elements::element_type::square:
		{
			auto square = static_cast<elements::square*>(element);
			if (idx == xorstr("Thickness"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				square->thickness = val;
			}
			else if (idx == xorstr("Filled"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				square->filled = val;
			}
			else if (idx == xorstr("Position"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				square->position = vector;
			}
			else if (idx == xorstr("Size"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				square->size = vector;
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::triangle:
		{
			auto triangle = static_cast<elements::triangle*>(element);
			if (idx == xorstr("Thickness"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				triangle->thickness = val;
			}
			else if (idx == xorstr("Filled"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				triangle->filled = val;
			}
			else if (idx == xorstr("PointA"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				triangle->p1 = vector;
			}
			else if (idx == xorstr("PointB"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				triangle->p2 = vector;
			}
			else if (idx == xorstr("PointC"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				triangle->p3 = vector;
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::quad:
		{
			auto quad = static_cast<elements::quad*>(element);
			if (idx == xorstr("Thickness"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				quad->thickness = val;
			}
			else if (idx == xorstr("Filled"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				quad->filled = val;
			}
			else if (idx == xorstr("PointA"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				quad->p1 = vector;
			}
			else if (idx == xorstr("PointB"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				quad->p2 = vector;
			}
			else if (idx == xorstr("PointC"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				quad->p3 = vector;
			}
			else if (idx == xorstr("PointD"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				quad->p4 = vector;
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}
			break;
		}
		case elements::element_type::text:
		{
			auto text = static_cast<elements::text*>(element);

			if (idx == xorstr("Text"))
			{
				verify_type(thread, -1, xorstr("string"));

				auto val = functions::lua::tostring(thread, -1);
				text->content = val;
			}
			else if (idx == xorstr("Font"))
			{
				verify_type(thread, -1, xorstr("string"));

				auto val = functions::lua::tostring(thread, -1);
				text->font = val;
			}
			else if (idx == xorstr("Size"))
			{
				verify_type(thread, -1, xorstr("number"));

				auto val = functions::lua::tonumber(thread, -1);
				text->size = val;
			}
			else if (idx == xorstr("Outline"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				text->outline = val;
			}
			else if (idx == xorstr("OutlineColor"))
			{
				verify_type(thread, -1, xorstr("Color3"));

				auto color = get_color3(thread, -1);

				text->outline_color = color;
				return 0;
			}
			else if (idx == xorstr("Center"))
			{
				verify_type(thread, -1, xorstr("boolean"));

				auto val = functions::lua::toboolean(thread, -1);
				text->center = val;
			}
			else if (idx == xorstr("Position"))
			{
				verify_type(thread, -1, xorstr("Vector2"));

				auto vector = get_vector2(thread, -1);

				text->position = vector;
			}
			else if (idx == xorstr("TextBounds"))
			{
				throw std::exception(xorstr("TextBounds is a read-only member of Text"));
			}
			else
			{
				functions::lua::luaG_indexerror(thread, functions::lua::index2adr(thread, 1), functions::lua::index2adr(thread, 2));
			}

			overlay::text_set_bounds(text);

			break;
		}
		default:
		{
			throw std::exception(xorstr("Why the fuck would you call this newindex with that"));
			break;
		}
		}

		return 0;
	}

	int drawing_namecall(std::uintptr_t thread)
	{
		auto method = std::string(functions::lua::getnamecallmethod(thread));

		if (method == xorstr("Remove"))
		{
			if (functions::lua::gettop(thread) == 1)
			{
				auto element = *reinterpret_cast<elements::base**>(functions::lua::touserdata(thread, 1));

				element->remove();
			}
			else
				throw std::exception((std::string(xorstr("Remove does not take ") + std::to_string(functions::lua::gettop(thread) - 1) + xorstr(" arguments."))).c_str());
		}
		else
			throw std::exception((std::string(method) + xorstr(" is not a valid member of Element")).c_str());
	}

	int load_font(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tstring);
		auto arg2 = get_argument(thread, 2, memdata::lua::types::lua_tstring);

		overlay::load_font(functions::lua::tostring(thread, 2), functions::lua::tostring(thread, 1));

		return 0;
	}

	int drawing_new(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, memdata::lua::types::lua_tstring);

		auto type = functions::lua::tostring(thread, 1);
		elements::base* object = nullptr;

		//obj->p1 = Drawing::Vector2(0, 0);
		//obj->p2 = Drawing::Vector2(500, 500);

		if (strcmp(type, xorstr("Line")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::line());
		}
		else if (strcmp(type, xorstr("Circle")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::circle());
		}
		else if (strcmp(type, xorstr("Square")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::square());
		}
		else if (strcmp(type, xorstr("Triangle")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::triangle());
		}
		else if (strcmp(type, xorstr("Quad")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::quad());
		}
		else if (strcmp(type, xorstr("Text")) == 0)
		{
			object = reinterpret_cast<elements::base*>(new elements::text());
		}
		else
		{
			throw std::exception((std::string(type) + xorstr(" is not a valid 2D element")).c_str());
		}


		elements::elements.push_back(object);

		auto res = functions::lua::newuserdata(thread, sizeof(object), 0);
		*reinterpret_cast<std::uintptr_t*>(res) = 
			reinterpret_cast<std::uintptr_t>(object);

		functions::lua::createtable(thread, 0, 0);

		checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(drawing_index), 0);
		functions::lua::setfield(thread, -2, xorstr("__index"));


		checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(drawing_newindex), 0);
		functions::lua::setfield(thread, -2, xorstr("__newindex"));

		//checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(drawing_namecall), 0);
		//functions::lua::setfield(thread, -2, "__namecall");

		functions::lua::setmetatable(thread, -2);

		return 1;
	}

	namespace initialiser
	{
		void load_drawing_functions(temple::lua::state thread)
		{
			auto drawing_str = xorstr("Drawing");

			functions::lua::createtable(thread, 0, 0);
			functions::lua::setfield(thread, -10002, drawing_str);

			set_table_function(thread, drawing_new, xorstr("new"), drawing_str);
			set_table_function(thread, load_font , xorstr("LoadFont"), drawing_str);
		}
	}
}