#include "mthooks.h"

#include "../customfn.h"

#include <rbx/memory/functions/functionhandler/functionhandler.h>
#include <rbx/memory/functions/functions.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <util/util.h>
#include <xorstr.h>
#include <temple/config/config.h>

#include <iostream>
#include <cstdint>

namespace temple::customfn::mthooks
{
	functions::temple_fn<int, std::uintptr_t> mt_index;
	functions::temple_fn<int, std::uintptr_t> mt_namecall;

	std::uintptr_t marketplaceservice = 0;
	std::uintptr_t messagebusservice = 0;
	std::uintptr_t browserservice = 0;
	std::uintptr_t guiservice = 0;

	bool raw_check_caller(std::uintptr_t thread)
	{
		if (functions::lua::state::get_identity(thread) == temple::config::lua::identity)
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	bool check_browserservice_func(const std::string& str)
	{
		auto lowstr = util::string::to_lower(str);

		if (lowstr.find(xorstr("closebrowserwindow")) != std::string::npos || lowstr.find(xorstr("copyauthcookie")) != std::string::npos ||
			lowstr.find(xorstr("emithybrid")) != std::string::npos || lowstr.find(xorstr("javascript")) != std::string::npos ||
			lowstr.find(xorstr("open")) != std::string::npos || lowstr.find(xorstr("sendcommand")) != std::string::npos)
		{
			return true;
		}

		return false;
	}

	std::uintptr_t get_raw_service(std::uintptr_t thread, std::string_view service)
	{
		functions::lua::getfield(thread, -10002, "game");
		functions::lua::getfield(thread, -1, "GetService");
		functions::lua::pushvalue(thread, -2);
		functions::lua::pushstring(thread, service.data());

		functions::lua::pcall(thread, 2, 1, 0);

		auto ud = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, -1));

		functions::lua::pop(thread, 2);

		return ud;
	}

	void index_call(std::uintptr_t thread, int idx, const char* index)
	{
		functions::lua::pushcclosure(thread, mt_index.get_ptr(), 0, 0, 0);

		functions::lua::pushvalue(thread, (idx < 0 && idx > -10000) ? idx - 1 : idx);
		functions::lua::pushstring(thread, index);

		functions::lua::pcall(thread, 2, 1, 0);
	}

	int getobjects(std::uintptr_t thread)
	{
		auto game = get_argument(thread, 1, memdata::lua::types::lua_tuserdata);
		auto fn = get_argument(thread, 2, memdata::lua::types::lua_tstring);

		functions::lua::getfield(thread, -10002, xorstr("game"));
		functions::lua::getfield(thread, -1, xorstr("GetService"));
		
		functions::lua::pushvalue(thread, -2);
		functions::lua::pushstring(thread, xorstr("InsertService"));
		functions::lua::pcall(thread, 2, 1, 0);

		functions::lua::getfield(thread, -1, xorstr("LoadLocalAsset"));

		functions::lua::pushvalue(thread, -2);
		functions::lua::pushvalue(thread, 2);
		functions::lua::pcall(thread, 2, 1, 0);

		if (functions::lua::type(thread, -1) == memdata::lua::types::lua_tstring)
		{
			std::string str{ functions::lua::tostring(thread, -1) };

			throw std::exception(str.c_str());
		}

		functions::lua::createtable(thread, 0, 0);

		functions::lua::pushvalue(thread, -2);

		functions::lua::rawseti(thread, -2, 1);


		return 1;
	}

	int httpget(std::uintptr_t thread)
	{
		auto game = get_argument(thread, 1, memdata::lua::types::lua_tuserdata);
		auto url = get_argument(thread, 2, memdata::lua::types::lua_tstring);

		functions::lua::getfield(thread, -10002, "request");

		functions::lua::createtable(thread, 0, 0);
		
		functions::lua::pushstring(thread, "GET");
		functions::lua::setfield(thread, -2, "Method");

		functions::lua::pushvalue(thread, 2);
		functions::lua::setfield(thread, -2, "Url");

		functions::lua::pcall(thread, 1, 1, 0);
		
		functions::lua::getfield(thread, -1, "Body");

		return 1;
	}

	int httppost(std::uintptr_t thread)
	{
		auto game = get_argument(thread, 1, memdata::lua::types::lua_tuserdata);
		auto url = get_argument(thread, 2, memdata::lua::types::lua_tstring);
		auto data = get_argument(thread, 3, memdata::lua::types::lua_tstring);

		functions::lua::getfield(thread, -10002, "request");

		functions::lua::createtable(thread, 0, 0);

		functions::lua::pushstring(thread, "POST");
		functions::lua::setfield(thread, -2, "Method");

		functions::lua::pushvalue(thread, 2);
		functions::lua::setfield(thread, -2, "Url");

		functions::lua::pushvalue(thread, 3);
		functions::lua::setfield(thread, -2, "Body");

		functions::lua::pcall(thread, 1, 1, 0);

		functions::lua::getfield(thread, -1, "Body");

		return 1;
	}

	int index_hook(std::uintptr_t thread)
	{
		if (raw_check_caller(thread))
		{
			if (functions::lua::gettop(thread) < 2)
			{
				return 0;
			}

			if (functions::lua::type(thread, 1) != memdata::lua::types::lua_tuserdata ||
				functions::lua::type(thread, 2) != memdata::lua::types::lua_tstring)
			{
				return 0;
			}

			auto self = functions::lua::index2adr(thread, 1);
			auto idx = std::string{ functions::lua::tostring(thread, 2) };

			if (util::string::to_lower(idx).find(xorstr("openvideosfolder")) != std::string::npos ||
				util::string::to_lower(idx).find(xorstr("openscreenshotsfolder")) != std::string::npos)
			{
				return 0;
			}

			if (functions::lua::type(thread, 1) == memdata::lua::types::lua_tuserdata)
			{
				auto ud = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, 1));

				if (ud == messagebusservice || ud == browserservice ||
					(ud == guiservice && util::string::to_lower(idx).find(xorstr("openbrowserwindow")) != std::string::npos) ||
					(ud == marketplaceservice && (util::string::to_lower(idx).find(xorstr("prompt")) != std::string::npos 
						|| util::string::to_lower(idx).find(xorstr("purchase")) != std::string::npos)))
				{
					return 0;
				}
			}

			if (idx == xorstr("SimulationRadius"))
			{
				functions::lua::getfield(thread, -10002, xorstr("game"));
				index_call(thread, -1, xorstr("Players"));
				index_call(thread, -1, xorstr("LocalPlayer"));

				auto localplayer = functions::lua::index2adr(thread, -1)->v.gcobj;

				functions::lua::pop(thread, 3);

				if (self->v.gcobj == localplayer)
				{
					functions::lua::getfield(thread, -10002, xorstr("gethiddenproperty"));

					functions::lua::pushvalue(thread, 1);
					functions::lua::pushvalue(thread, 2);

					if (!functions::lua::pcall(thread, 2, 1, 0))
					{
						return 1;
					}

					functions::lua::pop(thread, 1);
				}
			}

			if (idx == xorstr("GetObjects"))
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(getobjects));

				return 1;
			}

			if (idx == xorstr("HttpGet") || idx == xorstr("HttpGetAsync"))
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(httpget));

				return 1;
			}

			if (idx == xorstr("HttpPost") || idx == xorstr("HttpPostAsync"))
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(httppost));

				return 1;
			}

			functions::lua::getfield(thread, -10002, xorstr("game"));

			auto game = functions::lua::index2adr(thread, -1)->v.gcobj;

			functions::lua::pop(thread, 1);

			if (self->v.gcobj == game)
			{
				functions::lua::getfield(thread, -10002, xorstr("game"));
				index_call(thread, -1, xorstr("GetService"));

				functions::lua::pushvalue(thread, 1);
				functions::lua::pushvalue(thread, 2);

				if (!functions::lua::pcall(thread, 2, 1, 0) && functions::lua::type(thread, -1) != memdata::lua::types::lua_tnil)
				{
					return 1;
				}

				functions::lua::pop(thread, 2);
			}
		}

		return mt_index(thread);
	}

	int namecall_hook(std::uintptr_t thread)
	{
		if (raw_check_caller(thread))
		{
			auto method = functions::lua::getnamecallmethod(thread);

			if (method)
			{
				std::string func{ method };

				if (util::string::to_lower(func).find(xorstr("openvideosfolder")) != std::string::npos ||
					util::string::to_lower(func).find(xorstr("openscreenshotsfolder")) != std::string::npos)
				{
					return 0;
				}
				
				if (functions::lua::gettop(thread) >= 2)
				{
					auto idx = functions::lua::tostring(thread, 2);
					if (util::string::to_lower(func).find(xorstr("getservice")) != std::string::npos &&
						(util::string::to_lower(idx).find(xorstr("browserservice")) != std::string::npos ||
							util::string::to_lower(idx).find(xorstr("messagebusservice")) != std::string::npos))
					{
						return 0;
					}
				}
				
				if (functions::lua::gettop(thread) >= 1 && functions::lua::type(thread, 1) == memdata::lua::types::lua_tuserdata)
				{
					auto ud = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, 1));

					if ((ud == browserservice && check_browserservice_func(func)) ||
						(ud == guiservice && util::string::to_lower(func).find(xorstr("openbrowserwindow")) != std::string::npos) ||
						(ud == marketplaceservice && (util::string::to_lower(func).find(xorstr("prompt")) != std::string::npos || util::string::to_lower(func).find(xorstr("purchase")) != std::string::npos)) ||
						(ud == messagebusservice && (util::string::to_lower(func).find(xorstr("getmessageid")) != std::string::npos || util::string::to_lower(func).find(xorstr("publish")) != std::string::npos)))
					{
						return 0;
					}
				}
				
				if (func == xorstr("GetObjects"))
				{
					auto args = functions::lua::gettop(thread);
					checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(getobjects));
					functions::lua::insert(thread, 1);

					functions::lua::pcall(thread, args, 1, 0);

					return 1;
				}
				
				if (func == xorstr("HttpGet") || func == xorstr("HttpGetAsync"))
				{
					auto args = functions::lua::gettop(thread);

					checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(httpget));
					functions::lua::insert(thread, 1);

					functions::lua::pcall(thread, args, -1, 0);

					return functions::lua::gettop(thread);
				}

				if (func == xorstr("HttpPost") || func == xorstr("HttpPostAsync"))
				{
					auto args = functions::lua::gettop(thread);

					checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(httppost));
					functions::lua::insert(thread, 1);

					functions::lua::pcall(thread, args, -1, 0);

					return functions::lua::gettop(thread);
				}
				
			}
		}

		return mt_namecall(thread);
	}

	void prepare_hooks(std::uintptr_t thread)
	{
		mt_index.load_data(temple::memdata::functions::lua::index);
		mt_namecall.load_data(temple::memdata::functions::lua::namecall);

		marketplaceservice = get_raw_service(thread, "MarketplaceService");
		messagebusservice = get_raw_service(thread, "MessageBusService");
		browserservice = get_raw_service(thread, "BrowserService");
		guiservice = get_raw_service(thread, "GuiService");

		functions::lua::getfield(thread, -10002, xorstr("game"));

		functions::lua::getmetatable(thread, -1);

		functions::lua::getfield(thread, -10002, xorstr("hookfunction"));

		functions::lua::getfield(thread, -2, xorstr("__index"));
		checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(index_hook));

		functions::lua::pcall(thread, 2, 0, 0);

		functions::lua::getfield(thread, -10002, xorstr("hookfunction"));

		functions::lua::getfield(thread, -2, xorstr("__namecall"));
		checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(namecall_hook));

		functions::lua::pcall(thread, 2, 0, 0);

		functions::lua::settop(thread, 0);
	}
}