#include "newcclosure.h"

#include <rbx/checks/callcheck/callcheck.h>
#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <util/bitwise.h>
#include <xorstr.h>

#include <unordered_map>
#include <iostream>

namespace temple::customfn::newcclosurehandler
{
	std::unordered_map<std::uintptr_t, std::uintptr_t> closure_map;


	int closure_handler(std::uintptr_t);

	int closure_error_handler(std::uintptr_t thread)
	{
		std::string error = functions::lua::tostring(thread, -1);

		//std::cout << error << "\n";

		if(error == xorstr("attempt to yield across metamethod/C-call boundary") || error.find(xorstr("is not")) != std::string::npos || error.find(":") == std::string::npos)
			return functions::lua::gettop(thread);

		//std::cout << "pero pero" << "\n";

		auto ci = temple::functions::lua::state::get_ci(thread);
		auto base_ci = temple::functions::lua::state::get_base_ci(thread);

		auto current_ci = ci;
		std::uint32_t stack_level = 0;

		do
		{
			auto func = (*reinterpret_cast<temple::lua::tvalue**>(current_ci + 
				memdata::offsets::lua::state::callinfo::func))->v.gcobj;

			auto res = checks::callcheck::get_associated_function(func);

			if (res.has_value() && res == reinterpret_cast<std::uintptr_t>(closure_handler))
				break;

			current_ci -= 24;
		} while (current_ci >= base_ci);

		stack_level = ((ci - current_ci) / 24) + 1;

		temple::lua::lua_Debug ar;

		functions::lua::getinfo(thread, xorstr("sl"), &ar, stack_level);

		auto first_colon = error.find(':');
		auto second_colon = error.find(':', first_colon + 1);

		auto new_error = std::string(ar.short_src) + ":" + std::to_string(ar.currentline) + error.substr(second_colon);

		functions::lua::pop(thread, 1);
		functions::lua::pushstring(thread, new_error.c_str());

		return functions::lua::gettop(thread);
	}

	int closure_handler(std::uintptr_t thread)
	{
		auto nargs = functions::lua::gettop(thread);

		functions::lua::getfield(thread, -10000, xorstr("tmp_fc"));
		functions::lua::getfield(thread, -1, std::to_string(functions::lua::state::get_closure(thread)).c_str());

		functions::lua::insert(thread, 1);
		functions::lua::pop(thread, 1);

		checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(closure_error_handler));
		functions::lua::insert(thread, 1);
		
		auto res = functions::lua::pcall(thread, nargs, -1, 1);
;
		if (res)
		{
			//std::cout << "commerza\n";
			if (functions::lua::type(thread, -1) == memdata::lua::types::lua_tstring)
			{
				//std::cout << "commerze\n";
				const char* msg = functions::lua::tostring(thread, -1);

				if (!strcmp(msg, xorstr("attempt to yield across metamethod/C-call boundary")))
				{
					//std::cout << "caught\n";
					return functions::lua::yield(thread, 0);
				}
				throw std::exception(msg);
			}
			else
			{
				throw std::exception(xorstr("unknown error"));
			}
		}

		return functions::lua::gettop(thread) - 1;
	}
	void push_wrapped_function(temple::lua::state thread, int idx)
	{
		auto wrap_func = functions::lua::index2adr(thread, idx)->v.gcobj;

		checks::callcheck::push_temple_function(*thread, reinterpret_cast<std::uintptr_t>(closure_handler));

		auto closure = functions::lua::index2adr(thread, -1)->v.gcobj;

		functions::lua::getfield(thread, -10000, xorstr("tmp_fc"));
		functions::lua::pushvalue(thread, idx);

		functions::lua::setfield(thread, -2, std::to_string(closure).c_str());

		functions::lua::pop(thread, 1);

	}

	void copy_wrapped_function(temple::lua::state thread, std::uintptr_t new_closure, std::uintptr_t orig_closure)
	{
		checks::callcheck::copy_temple_function(new_closure, orig_closure);

		functions::lua::getfield(thread, -10000, xorstr("tmp_fc"));

		functions::lua::getfield(thread, -1, std::to_string(orig_closure).c_str());
		functions::lua::setfield(thread, -2, std::to_string(new_closure).c_str());
		functions::lua::pop(thread, 1);
	}

	bool is_temple_newcclosure(temple::lua::state thread, std::uintptr_t closure)
	{
		functions::lua::getfield(thread, -10000, xorstr("tmp_fc"));
		functions::lua::getfield(thread, -1, std::to_string(closure).c_str());

		auto res = functions::lua::index2adr(thread, -1)->tt == memdata::lua::types::lua_tfunction;

		functions::lua::pop(thread, 2);

		if (res)
			return true;
		else
			return false;
	}


}