#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <themida/ThemidaSDK.h>
#include <temple/config/config.h>
#include "../execution.h"
#include <rbx/lua/bytecode/bytecode.h>
#include <util/util.h>
#include <xorstr.h>

#include <iostream>
#include <fstream>
namespace temple::customfn
{
	using namespace memdata::lua::types;

	namespace signal
	{
		std::size_t connection_node_offset = 0;
		std::size_t connection_chain_offset = 4;
		std::size_t connection_function_offset = 8;
		std::size_t connection_signal_offset = 12;

		bool node_is_alive(std::uintptr_t connection)
		{
			auto node = *reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset);

			return *reinterpret_cast<std::uint32_t*>(node) == 1;
		}

		bool node_is_connected(std::uintptr_t connection)
		{
			auto node = *reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset);

			auto chain_node = *reinterpret_cast<std::uintptr_t*>(connection + connection_chain_offset);

			if (chain_node)
			{
				chain_node = *reinterpret_cast<std::uintptr_t*>(chain_node);

				while (chain_node)
				{
					if (chain_node == node)
					{
						return true;
					}

					chain_node = *reinterpret_cast<std::uintptr_t*>(chain_node + 16);
				}
			}
			return false;
		}

		int connection_disable(std::uintptr_t thread)
		{
			auto arg1 = get_argument(thread, 1, lua_tuserdata);

			auto ud_type = functions::lua::getobjecttype(thread, 1);

			auto connection = functions::lua::touserdata(thread, 1);
			auto node = *reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset);

			auto chain = *reinterpret_cast<std::uintptr_t*>(connection + connection_chain_offset);
			std::uintptr_t prev_node = 0;

			if (chain && *reinterpret_cast<std::uint32_t*>(node) == 1)
			{
				auto chain_node = *reinterpret_cast<std::uintptr_t*>(chain);

				while (chain_node)
				{
					if (chain_node == node)
					{
						if (prev_node)
						{
							*reinterpret_cast<std::uintptr_t*>(prev_node + 16) = *reinterpret_cast<std::uintptr_t*>(chain_node + 16);
						}
						else
						{
							*reinterpret_cast<std::uintptr_t*>(chain) = *reinterpret_cast<std::uintptr_t*>(chain_node + 16);
						}

						*reinterpret_cast<std::uint32_t*>(chain_node) = 2;
						return 0;
					}

					prev_node = chain_node;
					chain_node = *reinterpret_cast<std::uintptr_t*>(chain_node + 16);
				}
			}

			return 0;
		}

		int connection_enable(std::uintptr_t thread)
		{
			auto arg1 = get_argument(thread, 1, lua_tuserdata);

			auto ud_type = functions::lua::getobjecttype(thread, 1);

			auto connection = functions::lua::touserdata(thread, 1);
			auto node = *reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset);

			auto chain = *reinterpret_cast<std::uintptr_t*>(connection + connection_chain_offset);
			std::uintptr_t prev_node = 0;

			if (chain && *reinterpret_cast<std::uint32_t*>(node) != 1)
			{
				if (*reinterpret_cast<std::uint32_t*>(node) == 2)
				{
					*reinterpret_cast<std::uintptr_t*>(node + 16) = *reinterpret_cast<std::uintptr_t*>(chain);
					*reinterpret_cast<std::uintptr_t*>(chain) = node;

					*reinterpret_cast<std::uint32_t*>(node) = 1;
				}
				else if (*reinterpret_cast<std::uint32_t*>(node) == 0)
				{
					//auto lol = *reinterpret_cast<std::weak_ptr<std::uintptr_t>*>(connection + connection_signal_offset);

					//functions::reflection::instancebridge_push(thread, lol.lock());

					auto wrapper = *reinterpret_cast<std::uintptr_t*>(node + 28);

					if (wrapper)
					{
						functions::lua::rawgeti(thread, -10000, connection);
						functions::lua::rawgeti(thread, -1, 2);

						functions::lua::getfield(thread, -1, "Connect");

						functions::lua::pushvalue(thread, -2);
						functions::lua::rawgeti(thread, -4, 1);

						functions::lua::pcall(thread, 2, 1, 0);

						auto node = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, -1));

						*reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset) = node;
					}
				}
			}

			return 0;
		}

		int connection_fire(std::uintptr_t thread)
		{
			auto arg1 = get_argument(thread, 1, lua_tuserdata);

			auto connection = functions::lua::touserdata(thread, 1);

			auto node = *reinterpret_cast<std::uintptr_t*>(connection + connection_node_offset);

			auto args = functions::lua::gettop(thread) - 1;

			if (node_is_alive(connection))
			{
				auto wrapper = *reinterpret_cast<std::uintptr_t*>(node + 28);
				
				if (wrapper)
				{
					functions::lua::rawgeti(thread, -10000, connection);
					functions::lua::rawgeti(thread, -1, 1);

					functions::lua::insert(thread, 2);
					functions::lua::pop(thread, 1);

					functions::lua::pcall(thread, args, 0, 0);
				}

				return 1;
			}

			return 0;
		}

		int connection_index(std::uintptr_t thread)
		{
			auto connection = functions::lua::touserdata(thread, 1);

			auto node = *reinterpret_cast<std::uintptr_t*>(connection);

			auto method = functions::lua::index2adr(thread, 2)->v.gcobj;

			//std::cout << "calling\n";

			auto index_method = std::string_view{ functions::lua::string::get_buffer(method), 
				functions::lua::string::get_length(method) };

			if (index_method == "State")
			{
				functions::lua::pushboolean(thread, node_is_alive(connection));

				return 1;
			}
			else if (index_method == "Fire")
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(connection_fire));

				return 1;
			}
			else if (index_method == "Disable")
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(connection_disable));

				return 1;
			}
			else if (index_method == "Enable")
			{
				checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(connection_enable));

				return 1;
			}
			else if (index_method == "Function")
			{
				if (node_is_alive(connection))
				{
					auto wrapper = *reinterpret_cast<std::uintptr_t*>(node + 28);

					if (wrapper)
					{
						functions::lua::rawgeti(thread, -10000, connection);
						functions::lua::rawgeti(thread, -1, 1);
					}
					else
					{
						functions::lua::pushnil(thread);
					}

					return 1;
				}

				functions::lua::pushnil(thread);
				return 1;
			}
			else
			{
				throw std::exception(("attempt to index connection with " + std::string{ index_method }).c_str());
			}

			return 0;
		}

		void create_connection_metatable(std::uintptr_t thread)
		{
			functions::lua::createtable(thread, 0, 0);

			checks::callcheck::push_temple_function(thread, reinterpret_cast<std::uintptr_t>(connection_index));

			functions::lua::setfield(thread, -2, xorstr("__index"));
		}

		int getconnections(std::uintptr_t thread)
		{
			auto arg1 = get_argument(thread, 1, lua_tuserdata);

			auto ud_type = functions::lua::getobjecttype(thread, 1);

			if (ud_type != "RBXScriptSignal")
			{
				throw std::exception(("invalid argument #1 (RBXScriptSignal expected, got " + ud_type + ")").c_str());
			}
			/*auto event_instance = functions::lua::touserdata(thread, -1);

			functions::lua::createtable(thread, 0, 0);

			auto raw_ei = *reinterpret_cast<std::uintptr_t*>(event_instance);

			//std::cout << "china\n";

			if (!raw_ei)
				return 1;

			auto source = *reinterpret_cast<std::uintptr_t*>(event_instance + memdata::offsets::instance::event_instance::source);

			//std::cout << "sounds\n";

			if (!source)
				return 1;

			auto first_node_idx = *reinterpret_cast<std::uintptr_t*>(raw_ei + memdata::offsets::instance::event_instance::first_node);

			//std::cout << "like " << std::hex << first_node_idx << " " << source << "\n";

			if (!first_node_idx)
				return 1;

			auto node = *reinterpret_cast<std::uintptr_t*>(source + first_node_idx);

			if (!node || !*reinterpret_cast<std::uint32_t*>(node))
			{
				//std::cout << "vagina\n";
				return 1;
			}*/

			functions::lua::getfield(thread, -1, "Connect");

			functions::lua::pushvalue(thread, -2);
			functions::lua::pushcclosure(thread, 0, 0, 0, 0);

			functions::lua::pcall(thread, 2, 1, 0);

			auto node = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, -1));

			functions::lua::createtable(thread, 0, 0);

			std::size_t i = 1;

			node = *reinterpret_cast<std::uintptr_t*>(node + 16);
			
			while (node)
			{
				auto ud = functions::lua::newuserdata(thread, 8, 0);

				create_connection_metatable(thread);
				functions::lua::setmetatable(thread, -2);

				*reinterpret_cast<std::uintptr_t*>(ud + connection_node_offset) = node;
				*reinterpret_cast<std::uintptr_t*>(ud + connection_chain_offset) = *reinterpret_cast<std::uintptr_t*>(node + 20);

				auto wrapper = *reinterpret_cast<std::uintptr_t*>(node + 28);

				if (wrapper)
				{
					auto weakthreadref = *reinterpret_cast<std::uintptr_t*>(wrapper + 100);

					if (weakthreadref)
					{
						auto weak_thread = *reinterpret_cast<std::uintptr_t*>(weakthreadref + 4);
						auto key = *reinterpret_cast<std::uint32_t*>(weakthreadref + 12);

						//auto fn = reinterpret_cast<void(__fastcall*)(std::uintptr_t, std::uintptr_t)>(0x12D1CA0 + memdata::base);

						//fn(weak_thread, 1);

						//functions::lua::luaC_wakethread(weak_thread);

						functions::lua::rawgeti(weak_thread, -10000, key);

						if (functions::lua::type(weak_thread, -1) == memdata::lua::types::lua_tfunction)
						{

							auto func = functions::lua::index2adr(weak_thread, -1);

							functions::lua::createtable(thread, 0, 0);

							auto state = temple::lua::state(thread);

							auto& top = state.top;

							top->v = func->v;
							top->tt = func->tt;

							top += 1;

							functions::lua::rawseti(thread, -2, 1);
							functions::lua::pushvalue(thread, 1);
							functions::lua::rawseti(thread, -2, 2);
							functions::lua::rawseti(thread, -10000, ud);

							functions::lua::pop(weak_thread, 1);
						}
					}

				}
				functions::lua::rawseti(thread, -2, i++);

				node = *reinterpret_cast<std::uintptr_t*>(node + 16);

			}

			functions::lua::getfield(thread, -2, "Disconnect");
			functions::lua::pushvalue(thread, -3);
			functions::lua::pcall(thread, 1, 0, 0);

			return 1;
		}
	}

	int firesignal(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto ud_type = functions::lua::getobjecttype(thread, 1);

		if (ud_type != "RBXScriptSignal")
		{
			throw std::exception(("invalid argument #1 (RBXScriptSignal expected, got " + ud_type + ")").c_str());
		}

		auto args_num = functions::lua::gettop(thread) - 1;

		functions::lua::getfield(thread, 1, "Connect");

		functions::lua::pushvalue(thread, 1);
		functions::lua::pushcclosure(thread, 0, 0, 0, 0);

		functions::lua::pcall(thread, 2, 1, 0);

		auto node = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, -1));

		std::size_t i = 1;

		node = *reinterpret_cast<std::uintptr_t*>(node + 16);

		while (node)
		{
			auto wrapper = *reinterpret_cast<std::uintptr_t*>(node + 28);

			if (wrapper)
			{
				auto weakthreadref = *reinterpret_cast<std::uintptr_t*>(wrapper + 100);

				auto weak_thread = *reinterpret_cast<std::uintptr_t*>(weakthreadref + 4);
				auto key = *reinterpret_cast<std::uint32_t*>(weakthreadref + 12);

				functions::lua::luaC_wakethread(weak_thread);
				functions::lua::rawgeti(weak_thread, -10000, key);

				if (functions::lua::type(weak_thread, -1) == memdata::lua::types::lua_tfunction)
				{

					auto func = functions::lua::index2adr(weak_thread, -1);

					auto state = temple::lua::state(thread);

					auto& top = state.top;

					top->v = func->v;
					top->tt = func->tt;

					top += 1;

					for (std::size_t i = 2; i <= args_num + 1; i++)
					{
						functions::lua::pushvalue(thread, i);
					}

					functions::lua::pcall(thread, args_num, 0, 0);

					functions::lua::pop(weak_thread, 1);
				}
			}
			node = *reinterpret_cast<std::uintptr_t*>(node + 16);

		}

		functions::lua::getfield(thread, -1, "Disconnect");
		functions::lua::pushvalue(thread, -2);
		functions::lua::pcall(thread, 1, 0, 0);

		return 0;
	}

	int getcallingscript(std::uintptr_t thread)
	{
		//std::cout << "poopfucking\n";
		const auto& script = functions::lua::state::get_extraspace(thread)->script;

		//std::cout << "sexfucking\n";

		if (script)
		{
			//std::cout << "anal punish\n";
			functions::reflection::instancebridge_push(thread, script);
			//std::cout << "ok no punish lol\n";
		}
		else
		{
			functions::lua::pushnil(thread);
		}

		return 1;
	}

	int getsenv(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);
		auto script = *reinterpret_cast<std::uintptr_t*>(arg1->v.gcobj + 16);

		std::uintptr_t state = 0;

		auto node = *reinterpret_cast<std::uintptr_t*>(script + memdata::offsets::weakthreadref::thread_node);
		auto weakthreadref = node ? *reinterpret_cast<std::uintptr_t*>(node + memdata::offsets::weakthreadref::node_first) : 0;

		for (; weakthreadref; weakthreadref = *reinterpret_cast<std::uintptr_t*>(weakthreadref + memdata::offsets::weakthreadref::node_next))
		{
			auto ref_state = *reinterpret_cast<std::uintptr_t*>(weakthreadref + memdata::offsets::weakthreadref::state_off1);

			if (ref_state)
				ref_state = *reinterpret_cast<std::uintptr_t*>(ref_state + memdata::offsets::weakthreadref::state_off2);
			else
				continue;

			if (ref_state)
			{
				const auto& ref_script = functions::lua::state::get_extraspace(ref_state)->script;

				if (ref_script)
				{
					if (reinterpret_cast<std::uintptr_t>(ref_script.get()) == script)
					{
						state = ref_state;
						break;
					}
				}
			}
		}

		if (state)
			functions::lua::pushtable(thread, functions::lua::index2adr(state, -10002)->v.gcobj);
		else
			throw std::exception(xorstr("Script is not loaded into memory!"));

		return 1;
	}

	int fireclickdetector(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);
		auto clickdetector = *reinterpret_cast<std::uintptr_t*>(arg1->v.gcobj + 16);

		functions::lua::getfield(thread, -10002, xorstr("game"));
		functions::lua::getfield(thread, -1, xorstr("Players"));
		functions::lua::getfield(thread, -1, xorstr("LocalPlayer"));

		auto player = *reinterpret_cast<std::uintptr_t*>(functions::lua::index2adr(thread, -1)->v.gcobj + 16);

		functions::lua::pop(thread, 4);

		__asm
		{
			mov eax, 0
			movd xmm1, eax
		}

		reinterpret_cast<void(__thiscall*)(std::uintptr_t, std::uintptr_t)>(memdata::functions::instance::fireclickdetector.ptr)(clickdetector, player);

		return 0;
	}

	int fireproximityprompt (std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);
		auto proximityprompt = *reinterpret_cast<std::uintptr_t*>(arg1->v.gcobj + 16);

		functions::instance::fireproximityprompt(proximityprompt);

		return 0;
	}

	int sethiddenproperty(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);
		auto arg2 = get_argument(thread, 2, lua_tstring);

		auto object = arg2->v.gcobj;

		auto instance = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, 1));

		auto string_hash = *reinterpret_cast<short*>(object + memdata::offsets::lua::string::str_id);

		auto ktablecode = *reinterpret_cast<unsigned int*>(memdata::vars::instance::property_table + string_hash * sizeof(unsigned int));

		if (!ktablecode)
		{
		invalidprop:
			throw std::exception((std::string(functions::lua::tostring(thread, 2)) + std::string(xorstr(" is not a valid member of "))
				+ "\"" + functions::instance::get_name(reinterpret_cast<temple::rbx::instance*>(instance)) + '"').c_str());
		}

		auto instance_property = *reinterpret_cast<std::uintptr_t*>(instance + 12);

		auto property_descriptor = functions::instance::get_property_descriptor(instance_property, ktablecode);

		if (!property_descriptor)
			goto invalidprop;

		property_descriptor = *reinterpret_cast<std::uintptr_t*>(property_descriptor);

		if (!property_descriptor)
			goto invalidprop;

		std::uint32_t data[2];

		data[0] = thread;
		data[1] = 3;

		if (((*reinterpret_cast<std::uint32_t*>(property_descriptor + memdata::offsets::instance::property_descriptor::scriptable)) & 0x20) == 0)
			*reinterpret_cast<std::uint32_t*>(property_descriptor + memdata::offsets::instance::property_descriptor::scriptable) |= 0x20;

		functions::instance::set_instance_property(thread, property_descriptor, property_descriptor, instance, reinterpret_cast<std::uintptr_t>(&temple::config::lua::identity));

		return 0;
	}

	int gethiddenproperty(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);
		auto arg2 = get_argument(thread, 2, lua_tstring);

		auto object = arg2->v.gcobj;

		auto instance = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, 1));

		auto string_hash = *reinterpret_cast<short*>(object + memdata::offsets::lua::string::str_id);

		auto ktablecode = *reinterpret_cast<unsigned int*>(memdata::vars::instance::property_table + string_hash * sizeof(unsigned int));

		if (!ktablecode)
		{
		invalidprop:
			throw std::exception((std::string(functions::lua::tostring(thread, 2)) + std::string(xorstr(" is not a valid member of "))
				+ "\"" + functions::instance::get_name(reinterpret_cast<temple::rbx::instance*>(instance)) + '"').c_str());
		}

		auto instance_property = *reinterpret_cast<std::uintptr_t*>(instance + 12);

		auto property_descriptor = functions::instance::get_property_descriptor(instance_property, ktablecode);

		if (!property_descriptor)
			goto invalidprop;

		property_descriptor = *reinterpret_cast<std::uintptr_t*>(property_descriptor);

		if (!property_descriptor)
			goto invalidprop;

		std::uintptr_t propnobj[2];

		propnobj[0] = property_descriptor;
		propnobj[1] = instance;

		if (((*reinterpret_cast<std::uint32_t*>(property_descriptor + memdata::offsets::instance::property_descriptor::scriptable)) & 0x20) == 0)
			*reinterpret_cast<std::uint32_t*>(property_descriptor + memdata::offsets::instance::property_descriptor::scriptable) |= 0x20;

		functions::instance::get_instance_property(reinterpret_cast<std::uintptr_t>(propnobj), thread, reinterpret_cast<std::uintptr_t>(&temple::config::lua::identity));

		return 1;
	}

	int checkmodulescript(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto userdata = functions::lua::touserdata(thread, -1);

		auto script = *reinterpret_cast<std::uintptr_t*>(userdata);

		//std::cout << static_cast<std::uint32_t>(*reinterpret_cast<std::uint8_t*>(script + 264)) << "\n";

		*reinterpret_cast<std::uint8_t*>(script + 264) = 1;

		return 0;
	}

	int require(std::uintptr_t thread)
	{
		if (functions::lua::gettop(thread) == 1 && functions::lua::type(thread, -1) == memdata::lua::types::lua_tuserdata && functions::lua::state::is_temple_thread(thread))
		{
			auto userdata = functions::lua::touserdata(thread, -1);
			auto script = *reinterpret_cast<std::uintptr_t*>(userdata);

			//std::cout << functions::instance::get_classname(reinterpret_cast<rbx::instance*>(script)) << "\n";

			if (functions::instance::get_classname(reinterpret_cast<rbx::instance*>(script)) == xorstr("ModuleScript"))
			{

				//std::cout << std::hex << memdata::functions::lua::require.ptr - memdata::base + 0x400000 << "\n";

				auto old_flag = *reinterpret_cast<std::uint8_t*>(script + 264) = 1;

				auto identity = functions::lua::state::get_identity(thread);
				if (identity > 5 || identity == 3)
				{
					*reinterpret_cast<std::uint8_t*>(script + 264) = 1;
				}
				else
				{
					*reinterpret_cast<std::uint8_t*>(script + 264) = 0;
				}
				auto result = functions::lua::require(thread);
				*reinterpret_cast<std::uint8_t*>(script + 264) = old_flag;

				return result;
			}
		}

		return functions::lua::require(thread);
	}

	std::string to_hex(std::uint8_t byte)
	{
		std::string hex_table = "0123456789ABCDEF";
		std::string result = "";

		for (int i = 0; i < 2; i++)
		{
			result = hex_table[byte % 16] + result;
			byte /= 16;
		}

		return result;
	}

	int dumpstring(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto userdata = functions::lua::touserdata(thread, -1);
		auto script = *reinterpret_cast<std::uintptr_t*>(userdata);

		auto classname = functions::instance::get_classname(reinterpret_cast<rbx::instance*>(script));

		std::uintptr_t protected_string = 0;

		if (classname == xorstr("LocalScript"))
		{
			protected_string = script + memdata::offsets::instance::localscript::protected_string;//*reinterpret_cast<std::uintptr_t*>(script + memdata::offsets::instance::localscript::protected_string);
		}
		else if (classname == xorstr("ModuleScript"))
		{
			protected_string = script + memdata::offsets::instance::modulescript::protected_string;
		}
		else
		{
			throw std::exception((xorstr("invalid argument #1 (Script expected, got ") + classname + ")").c_str());
		}

		auto str = reinterpret_cast<std::string*>(*reinterpret_cast<std::uintptr_t*>(protected_string + 8) + 16);
		
		auto res = lua::bytecode::decompress_bytecode(*str);

		if (!res.has_value())
		{
			temple::lua::execution::load_script(thread, "");

			return 1;
		}

		auto res_script = *res;

		functions::lua::pushlstring(thread, res_script.c_str(), res_script.length());

		return 1;
	}

	int getscriptclosure(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto userdata = functions::lua::touserdata(thread, -1);
		auto script = *reinterpret_cast<std::uintptr_t*>(userdata);

		auto classname = functions::instance::get_classname(reinterpret_cast<rbx::instance*>(script));

		std::uintptr_t protected_string = 0;

		if (classname == xorstr("LocalScript"))
		{
			protected_string = script + memdata::offsets::instance::localscript::protected_string;//*reinterpret_cast<std::uintptr_t*>(script + memdata::offsets::instance::localscript::protected_string);
		}
		else if (classname == xorstr("ModuleScript"))
		{
			protected_string = script + memdata::offsets::instance::modulescript::protected_string;
		}
		else
		{
			throw std::exception((xorstr("invalid argument #1 (Script expected, got ") + classname + ")").c_str());
		}

		auto str = reinterpret_cast<std::string*>(*reinterpret_cast<std::uintptr_t*>(protected_string + 8) + 16);

		auto res = lua::bytecode::decompress_bytecode(*str);

		if (!res.has_value())
		{
			temple::lua::execution::load_script(thread, "");

			return 1;
		}

		std::string res_script = *res;

		functions::lua::load(thread, "", res_script.c_str(), res_script.length(), 0);

		return 1;
	}

	int isnetworkowner(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto userdata = functions::lua::touserdata(thread, -1);
		auto part = *reinterpret_cast<std::uintptr_t*>(userdata);

		auto memer = *reinterpret_cast<std::uintptr_t*>(part + 52);

		*reinterpret_cast<std::uintptr_t*>(memer + 52) = 0;

		auto classname = functions::instance::get_classname(reinterpret_cast<rbx::instance*>(part));

		if (classname.find(xorstr("Part")) != std::string::npos)
		{
			std::uintptr_t result[2];

			functions::instance::getnetworkowner(part, result);

			if (*result)
			{
				functions::lua::getfield(thread, -10002, xorstr("game"));
				functions::lua::getfield(thread, -1, xorstr("Players"));
				functions::lua::getfield(thread, -1, xorstr("LocalPlayer"));

				auto puserdata = functions::lua::touserdata(thread, -1);
				auto player = *reinterpret_cast<std::uintptr_t*>(puserdata);

				if (player == *result)
				{
					functions::lua::pushboolean(thread, true);

					return 1;
				}
			}
		}
		else
		{
			throw std::exception((xorstr("invalid argument #1 (Part expected, got ") + classname + ")").c_str());
		}

		functions::lua::pushboolean(thread, false);

		return 1;
	}

	int cloneref(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata);

		auto og_ud = arg1->v.gcobj;

		auto ud = functions::lua::newuserdata(thread, 8, *reinterpret_cast<std::uint8_t*>(og_ud + memdata::offsets::lua::userdata::tag));

		functions::lua::getmetatable(thread, -2);
		functions::lua::setmetatable(thread, -2);

		memcpy(reinterpret_cast<void*>(ud), reinterpret_cast<void*>(og_ud + memdata::offsets::lua::userdata::ud), 8);

		//functions::lua::pushstring(thread, std::to_string(ud));

		return 1;
	}

	namespace initialiser
	{
		void load_reflection_functions(temple::lua::state thread)
		{
			set_global_function(thread, require, xorstr("require"));
			//set_global_function(thread, getloadedmodules, "getloadedmodules");
			set_global_function(thread, fireproximityprompt, xorstr("fireproximityprompt"));
			set_global_function(thread, gethiddenproperty, xorstr("gethiddenproperty"));
			set_global_function(thread, sethiddenproperty, xorstr("sethiddenproperty"));
			set_global_function(thread, fireclickdetector, xorstr("fireclickdetector"));
			set_global_function(thread, getcallingscript, xorstr("getcallingscript"));
			set_global_function(thread, signal::getconnections, xorstr("getconnections"));
			set_global_function(thread, getscriptclosure, xorstr("getscriptclosure"));
			//set_global_function(thread, isnetworkowner, "isnetworkowner");
			set_global_function(thread, dumpstring, xorstr("dumpstring"));
			set_global_function(thread, firesignal, xorstr("firesignal"));
			set_global_function(thread, cloneref, xorstr("cloneref"));
			set_global_function(thread, getsenv, xorstr("getsenv"));
		}
	}
}