#include "customfn.h"

#include <temple/taskscheduler/taskscheduler.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <rbx/memory/functions/functions.h>
#include <themida/ThemidaSDK.h>
#include <rbx/memory/memdata/memdata.h>
#include <transpiler/serializer/serializer.h>
#include "newcclosure/newcclosure.h"
#include <temple/config/config.h>
#include "../execution.h"
#include <xorstr.h>

#include <iostream>
#include <string>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	int loadstring(std::uintptr_t thread)
	{
		std::string str = "";
		std::string chunkname = "";
		auto arg1 = get_argument(thread, 1, lua_tstring);

		if (functions::lua::gettop(thread) == 1)
		{
			str = std::string(functions::lua::tostring(thread, 1));

			lua::execution::load_script(thread, str, "");
		}
		else if (functions::lua::gettop(thread) == 2)
		{
			auto arg2 = get_argument(thread, 2, lua_tstring);

			str = std::string(functions::lua::tostring(thread, 1));
			chunkname = std::string(functions::lua::tostring(thread, 2));

			lua::execution::load_script(thread, str, chunkname.c_str());
		}

		if (functions::lua::type(thread, -1) == memdata::lua::types::lua_tstring)
			throw std::exception(functions::lua::tostring(thread, -1));

		return 1;
	}

	int doscript(std::uintptr_t thread)
	{
		return 0;
	}

	int getrawmetatable(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata, false, true, false);

		if (!arg1)
		{
			arg1 = get_argument(thread, 1, lua_ttable, false, true, false);
		}

		if (!arg1)
		{
			arg1 = get_argument(thread, 1, lua_tvector);
		}


		functions::lua::getmetatable(thread, 1);
		functions::lua::setreadonly(thread, -1, false);

		return 1;
	}

	int setrawmetatable(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tuserdata, false, true, false);

		if (!arg1)
		{
			arg1 = get_argument(thread, 1, lua_ttable, false, true, false);
		}

		if (!arg1)
		{
			arg1 = get_argument(thread, 1, lua_tvector);
		}

		auto arg2 = get_argument(thread, 2, lua_ttable);


		functions::lua::setmetatable(thread, -2);

		return 0;
	}

	int setreadonly(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_ttable);
		auto arg2 = get_argument(thread, 2, lua_tboolean);

		functions::lua::setreadonly(thread, 1, arg2->v.b);

		return 0;
	}

	int isreadonly(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_ttable);

		functions::lua::pushboolean(thread, *reinterpret_cast<bool*>(arg1->v.gcobj + memdata::offsets::lua::table::readonly));

		return 1;
	}

	int getthreadcontext(std::uintptr_t thread)
	{
		functions::lua::pushnumber(thread, functions::lua::state::get_identity(thread));

		return 1;
	}

	int setthreadcontext(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);

		functions::lua::state::set_identity(thread, static_cast<std::int32_t>(functions::lua::tonumber(thread, 1)));

		return 0;
	}

	int checkcaller(std::uintptr_t thread)
	{
		if (functions::lua::state::get_identity(thread) == temple::config::lua::identity)
		{
			functions::lua::pushboolean(thread, true);
		}
		else
		{
			functions::lua::pushboolean(thread, false);
		}

		return 1;
	}


	int getnamecallmethod(std::uintptr_t thread)
	{
		auto method = functions::lua::getnamecallmethod(thread);
		if (method)
			functions::lua::pushstring(thread, method);
		else
			functions::lua::pushnil(thread);

		return 1;
	}

	int setnamecallmethod(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		auto str = functions::lua::tostring(thread, 1);

		*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::namecall) = functions::lua::newlstr(thread, str, std::strlen(str));

		return 0;
	}

	int newcclosure(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		newcclosurehandler::push_wrapped_function(thread, 1);

		return 1;
	}

	int iscclosure(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		functions::lua::pushboolean(thread, functions::lua::closure::iscclosure(functions::lua::index2adr(thread, -1)->v.gcobj));

		return 1;
	}

	int islclosure(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tfunction);

		functions::lua::pushboolean(thread, functions::lua::closure::islclosure(arg1->v.gcobj));

		return 1;
	}

	int getcclosure(std::uintptr_t thread)
	{
		auto fn = get_argument(thread, 1, lua_tfunction)->v.gcobj;

		if (functions::lua::closure::iscclosure(fn))
		{
			functions::lua::pushcclosure(thread, functions::lua::closure::get_source(fn), 0, 0, 0);

			return 1;
		}
		return 0;
	}

	int hookfunction(std::uintptr_t thread)
	{
		auto fn1 = get_argument(thread, 1, lua_tfunction)->v.gcobj;
		auto fn2 = get_argument(thread, 2, lua_tfunction)->v.gcobj;

		auto f1_t = functions::lua::closure::iscclosure(fn1);
		auto f2_t = functions::lua::closure::iscclosure(fn2);

		auto f1_s = functions::lua::closure::get_source(fn1);
		auto f2_s = functions::lua::closure::get_source(fn2);

		////std::cout << "el gringos: " << f1_t << " " << f2_t << "\n";

		if ((f1_t && !f2_t) || functions::lua::closure::get_nups(fn2) > 0)
		{
			////std::cout << "wrapped fn2\n";
			newcclosurehandler::push_wrapped_function(thread, 2);
			fn2 = functions::lua::index2adr(thread, -1)->v.gcobj;

			f2_t = functions::lua::closure::iscclosure(fn2);
			f2_s = functions::lua::closure::get_source(fn2);
		}

		////std::cout << "types: " << f1_t << " " << f2_t << "\n";

		if (f1_t)
		{
			if (!functions::lua::closure::is_temple_closure(thread, fn1))
			{
				auto nups = functions::lua::closure::get_nups(fn1);
				auto& top = functions::lua::state::get_top(thread);
				for (std::size_t i = 0; i < nups; i++)
				{
					*top = *reinterpret_cast<temple::lua::tvalue*>(fn1 + memdata::offsets::lua::closure::c_ups + (i * 16));
					top += 1;
				}

				functions::lua::pushcclosure(thread, f1_s, nups, functions::lua::closure::get_name(fn1),
					functions::lua::closure::get_continuation(fn1));
			}
			else
			{
				if (newcclosurehandler::is_temple_newcclosure(thread, fn1))
				{
					////std::cout << "types: temple wrapped closure\n";
					functions::lua::pushcclosure(thread, 0, 0, 0, 0);
					newcclosurehandler::copy_wrapped_function(thread, functions::lua::index2adr(thread, -1)->v.gcobj, fn1);
				}
				else
				{
					////std::cout << "types: temple normal closure\n";
					auto associated = checks::callcheck::get_associated_function(fn1);

					if (associated.has_value() && *associated == reinterpret_cast<std::uintptr_t>(&checkcaller))
					{
						throw std::exception(xorstr("nice try retard"));
					}

					functions::lua::pushcclosure(thread, 0, 0, 0, 0);
					checks::callcheck::copy_temple_function(functions::lua::index2adr(thread, -1)->v.gcobj, fn1);
				}
			}
		}
		else
		{
			functions::lua::pushlclosure(thread, 0, functions::lua::closure::get_env(fn1), f1_s);
		}

		if (f2_t)
		{
			if (!functions::lua::closure::is_temple_closure(thread, fn2))
			{
				////std::cout << "types 2: rbx closure\n";
				functions::lua::closure::setcclosure(fn1, true);
				functions::lua::closure::set_source(fn1, f2_s);
				//functions::lua::closure::set_env(fn1, functions::lua::closure::get_env(fn2));
			}
			else
			{
				if (newcclosurehandler::is_temple_newcclosure(thread, fn2))
				{
					////std::cout << "types 2: temple wrapped closure\n";
					functions::lua::closure::setcclosure(fn1, true);
					//functions::lua::closure::set_nups(fn1, 0);
					//functions::lua::closure::set_maxstacksize(fn1, functions::lua::closure::get_maxstacksize(fn2));
					newcclosurehandler::copy_wrapped_function(thread, fn1, fn2);
					//functions::lua::closure::set_continuations(fn1, 0);
					//functions::lua::closure::set_env(fn1, functions::lua::closure::get_env(fn2));
				}
				else
				{
					////std::cout << "types 2: temple normal closure\n"; 
					functions::lua::closure::setcclosure(fn1, true);
					checks::callcheck::copy_temple_function(fn1, fn2);
					//functions::lua::closure::set_env(fn1, functions::lua::closure::get_env(fn2));
				}
			}
		}
		else
		{
			functions::lua::closure::setcclosure(fn1, false);
			functions::lua::closure::set_source(fn1, f2_s);
			//functions::lua::closure::set_env(fn1, functions::lua::closure::get_env(fn2));

		}


		return 1;
	}

	int getgenv(std::uintptr_t thread)
	{
		functions::lua::pushtable(thread, functions::lua::index2adr(*memdata::classes::lua::state, -10002)->v.gcobj);

		return 1;
	}

	int getrenv(std::uintptr_t thread)
	{
		functions::lua::pushtable(thread, functions::lua::index2adr(*memdata::classes::lua::rstate, -10002)->v.gcobj);

		return 1;
	}

	int getreg(std::uintptr_t thread)
	{
		functions::lua::pushtable(thread, functions::lua::index2adr(thread, -10000)->v.gcobj);

		return 1;
	}

	int getgc(std::uintptr_t thread)
	{
		bool cond = false;
		auto arg1 = get_argument(thread, 1, lua_tboolean, 0);

		if (arg1)
			cond = arg1->v.b;

		functions::lua::createtable(thread, 0, 0);

		std::uintptr_t gcobj = functions::lua::l_g::get_rootgc(functions::lua::state::get_l_g(thread));

		std::size_t gcon = 1;
		
		while (gcobj)
		{
			gcobj += memdata::offsets::lua::gcobj::entry::size;

			auto gcobj_tt = functions::lua::gcobj::get_tt(gcobj);

			if (gcobj_tt == lua_tfunction || (gcobj_tt == lua_ttable && cond))
			{
				auto& top = functions::lua::state::get_top(thread);

				top->v.gcobj = gcobj;
				top->tt = gcobj_tt;
				top++;

				functions::lua::rawseti(thread, -2, gcon++);
			}

			gcobj = functions::lua::gcobj::get_next(gcobj);
		}
		
		return 1;
	}

	int getinstancecache(std::uintptr_t thread)
	{
		functions::lua::createtable(thread, 0, 0);

		functions::lua::pushlightuserdata(thread, reinterpret_cast<void*>(memdata::functions::reflection::instancebridge_push.ptr));
		functions::lua::rawget(thread, -10000);

		if (functions::lua::type(thread, -1) != memdata::lua::types::lua_ttable)
		{
			functions::lua::pop(thread, 1);
			return 1;
		}

		std::size_t i = 1;
		
		functions::lua::pushnil(thread);
		while (functions::lua::next(thread, -2))
		{
			functions::lua::rawseti(thread, -4, i++);

		}

		functions::lua::pop(thread, 1);
		
		return 1;
	}
	
	int istempleclosure(std::uintptr_t thread)
	{
		auto fn = get_argument(thread, 1, lua_tfunction)->v.gcobj;

		functions::lua::pushboolean(thread, functions::lua::closure::is_temple_closure(thread, fn));

		return 1;
	}

	int queue_on_teleport(std::uintptr_t thread)
	{
		auto fn = get_argument(thread, 1, lua_tstring);

		auto result = temple::lua::execution::serialize_script(functions::lua::tostring(thread, 1));

		if (!result[0])
		{
			throw std::exception(&result[1]);
		}

		taskscheduler::add_teleport_script(result);

		return 0;
	}

	namespace initialiser
	{
		void load_rlua_functions(temple::lua::state thread)
		{
			set_global_function(thread, queue_on_teleport, xorstr("queue_on_teleport"));
			set_global_function(thread, getnamecallmethod, xorstr("getnamecallmethod"));
			set_global_function(thread, setnamecallmethod, xorstr("setnamecallmethod"));
			set_global_function(thread, getinstancecache, xorstr("getinstancecache"));
			set_global_function(thread, getthreadcontext, xorstr("getthreadcontext"));
			set_global_function(thread, setthreadcontext, xorstr("setthreadcontext"));
			set_global_function(thread, istempleclosure, xorstr("istempleclosure"));
			set_global_function(thread, getrawmetatable, xorstr("getrawmetatable"));
			set_global_function(thread, setrawmetatable, xorstr("setrawmetatable"));
			set_global_function(thread, hookfunction, xorstr("hookfunction"));
			set_global_function(thread, checkcaller, xorstr("checkcaller"));
			set_global_function(thread, setreadonly, xorstr("setreadonly"));
			set_global_function(thread, newcclosure, xorstr("newcclosure"));
			set_global_function(thread, iscclosure, xorstr("iscclosure"));
			set_global_function(thread, islclosure, xorstr("islclosure"));
			set_global_function(thread, isreadonly, xorstr("isreadonly"));
			set_global_function(thread, loadstring, xorstr("loadstring"));
			set_global_function(thread, getgenv, xorstr("getgenv"));
			set_global_function(thread, getrenv, xorstr("getrenv"));
			set_global_function(thread, getreg, xorstr("getreg"));
			set_global_function(thread, getgc, xorstr("getgc"));
		}
	}
}