#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <rbx/memory/memdata/memdata.h>
#include <auth/hwid_gen/hwid_gen.h>
#include <themida/ThemidaSDK.h>
#include <temple/config/config.h>
#include "../execution.h"
#include <auth/auth.h>
#include <http/http.h>
#include <util/util.h>
#include <xorstr.h>

#include <iostream>
#include <string>
#include <algorithm>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	int request(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_ttable);

		functions::lua::getfield(thread, -1, xorstr("Url"));

		if (functions::lua::type(thread, -1) == lua_tnil)
		{
			throw std::exception(xorstr("Url was not passed to request."));
		}

		std::string uri = functions::lua::tostring(thread, -1);  

        std::size_t start{};
        for (const auto& c : uri)
        {
            if (c != ' ')
                break;

            start++;
        }
        std::string protocol = uri.substr(start);
		protocol = temple::util::string::to_lower(protocol); // block file:/// FILE:/// etc

        if ((protocol.size() >= 5 && protocol.substr(0, 5) == xorstr("file:")) || (protocol.size() >= 7 && protocol.substr(0, 7) == xorstr("telnet:")))
            throw std::exception(xorstr("fail"));

		functions::lua::pop(thread, 1);

		std::string method = xorstr("GET");
		std::string body = "";
		std::vector<std::string> headers;

		auto username = auth::registry_get_username();

		if (!username.has_value())
		{
			throw std::exception(xorstr("Unexpected error when trying to perform request."));
		}

		headers.push_back(xorstr("User-Identifier: ") + *username);
		headers.push_back(xorstr("Fingerprint: ") + auth::hwid_gen::get_tiny_hwid());
		headers.push_back(xorstr("User-Agent: Temple V1"));

		functions::lua::getfield(thread, -1, xorstr("Method"));

		if (functions::lua::type(thread, -1) != lua_tnil)
		{
			method = functions::lua::tostring(thread, -1);

			std::transform(method.begin(), method.end(), method.begin(),
				[](unsigned char c) { return std::toupper(c); });
		}
		functions::lua::pop(thread, 1);

		functions::lua::getfield(thread, -1, xorstr("Headers"));

		if (functions::lua::type(thread, -1) != lua_tnil)
		{
			temple::functions::lua::pushnil(thread);
		
			while (functions::lua::next(thread, -2))
			{
				if (functions::lua::type(thread, -2) == lua_tstring)
				{
					auto header = std::string(functions::lua::tostring(thread, -2));
					std::string value = "";

					auto type_val = functions::lua::type(thread, -1);

					if (type_val == lua_tboolean)
					{
						if (functions::lua::toboolean(thread, -1))
							value = xorstr("true");
						else
							value = xorstr("false");
					}
					else if (type_val == lua_tnumber)
					{
						value = std::to_string(functions::lua::tonumber(thread, -1));
					}
					else if (type_val == lua_tstring)
					{
						value = functions::lua::tostring(thread, -1);
					}
					else if (type_val == lua_tuserdata)
					{
						auto ud = *reinterpret_cast<rbx::instance**>(functions::lua::touserdata(thread, -1));

						value = ud->name;
					}

					headers.push_back(header + ": " + value);

				}

				functions::lua::pop(thread, 1);
			}
		}
		functions::lua::pop(thread, 1);

		functions::lua::getfield(thread, -1, xorstr("Body"));

		if (functions::lua::type(thread, -1) != lua_tnil)
		{
			auto str = functions::lua::index2adr(thread, -1);

			body = std::string(reinterpret_cast<const char*>(str->v.gcobj + memdata::offsets::lua::string::raw),
				functions::lua::string::get_length(str->v.gcobj));
		}
		functions::lua::pop(thread, 1);

		auto result = http::send_request(uri, method, headers, body);

		functions::lua::createtable(thread, 0, 0);

		auto success = false;
		std::string response_body = "";

		if (result.has_value())
		{
			success = true;
			response_body = *result;
		}

		long statuscode = http::http_code;
		std::string successmessage = http::status_message[statuscode];

		functions::lua::pushboolean(thread, success);
		functions::lua::setfield(thread, -2, xorstr("Success"));

		functions::lua::pushstring(thread, successmessage.c_str());
		functions::lua::setfield(thread, -2, xorstr("StatusMessage"));

		functions::lua::pushnumber(thread, statuscode);
		functions::lua::setfield(thread, -2, xorstr("StatusCode"));

		functions::lua::pushstring(thread, response_body.c_str());
		functions::lua::setfield(thread, -2, xorstr("Body"));

		auto response_headers = http::last_headers;

		functions::lua::createtable(thread, 0, 0);

		for (const auto& str : response_headers)
		{
			auto colon = str.find(":");

			if (!colon)
			{
				functions::lua::pushnil(thread);
				functions::lua::setfield(thread, -2, str.c_str());
			}
			else
			{
				auto header = str.substr(0, colon);
				auto value = str.substr(colon + 2);

				functions::lua::pushstring(thread, value.c_str());
				functions::lua::setfield(thread, -2, header.c_str());
			}
		}

		functions::lua::setfield(thread, -2, xorstr("Headers"));

		return 1;
	}

	int konbanwa(std::uintptr_t thread)
	{
		auto player = *reinterpret_cast<std::uintptr_t*>(functions::lua::touserdata(thread, 1));

		//std::cout << "Retrieving ip of: " << functions::instance::get_name(reinterpret_cast<rbx::instance*>(player)) << "\n";

		auto sys_address = player + 740;

		//std::cout << std::hex << "System address: " << sys_address << "\n";

		auto ip = *reinterpret_cast<std::uint32_t*>(sys_address);
		auto port = *reinterpret_cast<std::uint16_t*>(sys_address + 4);

		auto ip_ptr = reinterpret_cast<std::uint8_t*>(&ip);

		//std::cout << std::dec;

		for (std::size_t i = 0; i < 4; i++)
		{
			//std::cout << static_cast<int>(ip_ptr[i]);
			if (i == 3)
			{
				//std::cout << "\n";
			}
			else
			{
				//std::cout << ".";
			}
		}
		//std::cout << std::dec << "port: " << port << "\n";

		return 0;
	}

	namespace initialiser
	{
		void load_temple_functions(temple::lua::state thread)
		{
			set_global_function(thread, request, xorstr("request"));
			//set_global_function(thread, konbanwa, "konbanwa");
		}
	}
}
