#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <temple/config/config.h>
#include <themida/ThemidaSDK.h>
#include "../execution.h"
#include <util/util.h>
#include <xorstr.h>

#include <Windows.h>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	int createdirectory(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		util::file::create_directory(path);

		return 0;
	}

	int isfolder(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		functions::lua::pushboolean(thread, util::file::is_folder(path));

		return 1;
	}

	int isfile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		functions::lua::pushboolean(thread, util::file::is_file(path));

		return 1;
	}

	int appendfile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);
		auto arg2 = get_argument(thread, 2, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		util::file::append_file(path, std::string(functions::lua::tostring(thread, 2), functions::lua::string::get_length(arg2->v.gcobj)));

		return 0;
	}

	int loadfile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		auto ret = util::file::read_file(path);

		if (ret.has_value())
			lua::execution::load_script(thread, ret->c_str());
		else
			throw std::exception(xorstr("File not found!"));

		return 1;
	}

	int dofile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		auto ret = util::file::read_file(path);

		functions::lua::settop(thread, 0);

		if (ret.has_value())
			lua::execution::load_script(thread, ret->c_str());
		else
			throw std::exception(xorstr("File not found!"));

		functions::lua::pcall(thread, 0, -1, 0);

		return 0;
	}

	int readfile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos || path.find(".exe") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		auto ret = util::file::read_file(path);

		if (ret.has_value())
			functions::lua::pushlstring(thread, ret->c_str(), ret->length());
		else
			throw std::exception(xorstr("File not found!"));

		return 1;
	}

	int writefile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);
		auto arg2 = get_argument(thread, 2, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos || temple::util::file::check_illegal_extension(path))
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		util::file::write_file(path, std::string(functions::lua::tostring(thread, 2), functions::lua::string::get_length(arg2->v.gcobj)));

		return 0;
	}

	int listfiles(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		functions::lua::createtable(thread, 0, 0);

		int i = 1;

		for (const auto& path : temple::util::file::get_filenames(path))
		{
			functions::lua::pushstring(thread, path.c_str());
			functions::lua::rawseti(thread, -2, i++);
		}

		return 1;
	}

	int delfile(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		std::string path = temple::config::util::current_path + xorstr("\\workspace\\") + functions::lua::tostring(thread, 1);

		if (path.find("..") != std::string::npos)
			throw std::exception(xorstr("Attempt to be a bloody benchod, stop doing that"));

		util::file::delete_file(path);

		return 0;
	}

	int setclipboard(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tstring);

		util::clipboard::write_text(functions::lua::tostring(thread, 1));

		return 0;
	}

	int messagebox(std::uintptr_t thread)
	{

		auto arg1 = get_argument(thread, 1, lua_tstring);
		auto arg2 = get_argument(thread, 2, lua_tstring);
		auto arg3 = get_argument(thread, 3, lua_tnumber, 0);

		auto Type = (arg3 == 0) ? MB_OK : functions::lua::tonumber(thread, 3);

		auto response = MessageBoxA(NULL, functions::lua::tostring(thread, 2), functions::lua::tostring(thread, 1), static_cast<unsigned int>(Type));

		functions::lua::pushnumber(thread, response);

		return 1;
	}

	int mousemoveabs(std::uintptr_t thread)
	{

		auto arg1 = get_argument(thread, 1, lua_tnumber);
		auto arg2 = get_argument(thread, 2, lua_tnumber);

		util::input::mouse::move(static_cast<int>(functions::lua::tonumber(thread, 1)), static_cast<int>(functions::lua::tonumber(thread, 2)));

		return 0;
	}

	int mousemoverel(std::uintptr_t thread)
	{

		auto arg1 = get_argument(thread, 1, lua_tnumber);
		auto arg2 = get_argument(thread, 2, lua_tnumber);

		util::input::mouse::move(static_cast<int>(functions::lua::tonumber(thread, 1)), static_cast<int>(functions::lua::tonumber(thread, 2)), true);

		return 0;
	}

	int keypress(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);

		util::input::keyboard::input(static_cast<int>(functions::lua::tonumber(thread, 1)), true);

		return 0;
	}
	int keyrelease(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);

		util::input::keyboard::input(static_cast<int>(functions::lua::tonumber(thread, 1)), false);

		return 0;
	}

	int mouse1press(std::uintptr_t thread)
	{
		util::input::mouse::input(true, true);

		return 0;
	}

	int mouse1release(std::uintptr_t thread)
	{
		util::input::mouse::input(true, false);

		return 0;
	}

	int mouse2press(std::uintptr_t thread)
	{
		util::input::mouse::input(false, true);

		return 0;
	}

	int mouse2release(std::uintptr_t thread)
	{
		util::input::mouse::input(false, false);

		return 0;
	}

	int mousescroll(std::uintptr_t thread)
	{
		auto arg1 = get_argument(thread, 1, lua_tnumber);

		util::input::mouse::scroll(functions::lua::tonumber(thread, 1));

		return 0;
	}

	int iswindowactive(std::uintptr_t thread)
	{
		auto active = GetForegroundWindow();
		int title_size = GetWindowTextLengthA(active);
		auto title = new char[title_size + 1];

		GetWindowTextA(active, title, title_size + 1);

		functions::lua::pushboolean(thread, strcmp(title, xorstr("Roblox")) == 0);

		delete[] title;

		return 1;
	}

	namespace initialiser
	{
		void load_win_functions(temple::lua::state thread)
		{
			set_global_function(thread, createdirectory, xorstr("createdirectory"));
			set_global_function(thread, iswindowactive, xorstr("iswindowactive"));
			set_global_function(thread, mouse1release, xorstr("mouse1release"));
			set_global_function(thread, mouse2release, xorstr("mouse2release"));
			set_global_function(thread, mousemoveabs, xorstr("mousemoveabs"));
			set_global_function(thread, mousemoverel, xorstr("mousemoverel"));
			set_global_function(thread, setclipboard, xorstr("setclipboard"));
			set_global_function(thread, mouse1press, xorstr("mouse1press"));
			set_global_function(thread, mouse2press, xorstr("mouse2press"));
			set_global_function(thread, mousescroll, xorstr("mousescroll"));
			set_global_function(thread, messagebox, xorstr("messagebox"));
			set_global_function(thread, keyrelease, xorstr("keyrelease"));
			set_global_function(thread, appendfile, xorstr("appendfile"));
			set_global_function(thread, writefile, xorstr("writefile"));
			set_global_function(thread, listfiles, xorstr("listfiles"));
			set_global_function(thread, keypress, xorstr("keypress"));
			set_global_function(thread, isfolder, xorstr("isfolder"));
			set_global_function(thread, readfile, xorstr("readfile"));
			set_global_function(thread, loadfile, xorstr("loadfile"));
			set_global_function(thread, delfile, xorstr("delfile"));
			set_global_function(thread, dofile, xorstr("dofile"));
			set_global_function(thread, isfile, xorstr("isfile"));
		}
	}
}