#pragma once
#include <rbx/checks/retcheck/retcheck.h>
#include <rbx/memory/memdata/memdata.h>

#include <iostream>

namespace temple::functions
{
	template<class R, class...A>
	class temple_fn
	{
	protected:
		memdata::functions::func_data fd;
		bool retchk = false;
	public:

		std::uintptr_t get_ptr()
		{
			return fd.ptr;
		}

		template<class...Args>
		R operator()(A...a, Args...args)
		{
			if constexpr (std::is_same_v<R, void>)
			{
				if(retchk)
				{
					if (fd.cc == memdata::functions::calling_convention::cc_fastcall ||
						fd.cc == memdata::functions::calling_convention::cc_rbxcall)
						checks::retcheck::ret_fastcall<void, A...>(fd.ptr, a..., args...);
					else
						checks::retcheck::ret_call<void, A...>(fd.ptr, a..., args...);
				}
				else
				{
					switch (fd.cc)
					{
						case memdata::functions::calling_convention::cc_cdecl:
						{
							reinterpret_cast<void(__cdecl*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_stdcall:
						{
							reinterpret_cast<void(__stdcall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_fastcall:
						{
							reinterpret_cast<void(__fastcall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_thiscall:
						{
							reinterpret_cast<void(__thiscall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_rbxcall:
						{
							//reinterpret_cast<void(__fastcall*)(A...)>(&checks::retcheck::arguments_push)(a...);
							//reinterpret_cast<void(__cdecl*)(Args...)>(fd.ptr)(args...);

							constexpr std::size_t arg_size = (0 + ... + sizeof(Args));

							reinterpret_cast<void(__fastcall*)(A..., Args...)>(fd.ptr)(a..., args...);

							__asm add esp, arg_size

							break;
						}
					}
				}
			}
			else
			{
				R result{};

				if (retchk)
				{
					if (fd.cc == memdata::functions::calling_convention::cc_fastcall ||
						fd.cc == memdata::functions::calling_convention::cc_rbxcall)
						result = checks::retcheck::ret_fastcall<R, A...>(fd.ptr, a..., args...);
					else
						result = checks::retcheck::ret_call<R, A...>(fd.ptr, a..., args...);
				}
				else
				{
					switch (fd.cc)
					{
						case memdata::functions::calling_convention::cc_cdecl:
						{
							result = reinterpret_cast<R(__cdecl*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_stdcall:
						{
							result = reinterpret_cast<R(__stdcall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_fastcall:
						{
							result = reinterpret_cast<R(__fastcall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_thiscall:
						{
							result = reinterpret_cast<R(__thiscall*)(A..., Args...)>(fd.ptr)(a..., args...);
							break;
						}
						case memdata::functions::calling_convention::cc_rbxcall:
						{
							//reinterpret_cast<void(__fastcall*)(A...)>(&checks::retcheck::arguments_push)(a...);
							//result = reinterpret_cast<R(__cdecl*)(A..., Args...)>(fd.ptr)(a..., args...);

							constexpr std::size_t arg_size = (0 + ... + sizeof(Args));

							result = reinterpret_cast<R(__fastcall*)(A..., Args...)>(fd.ptr)(a..., args...);

							__asm add esp, arg_size

							break;
						}
					}
				}

				return result;
			}
		}

		void load_data(memdata::functions::func_data fd)
		{
			this->fd = fd;
		}

		temple_fn() {}
		temple_fn(bool retchk) : retchk(retchk) {}
	};
}