#include "functions.h"

#include <rbx/lua/customfn/newcclosure/newcclosure.h>
#include <rbx/checks/callcheck/callcheck.h>
#include <rbx/checks/retcheck/retcheck.h>
#include <temple/config/config.h>
#include "../memdata/memdata.h"
#include "themida/ThemidaSDK.h"
#include "vanillalua/vanillalua.h"
#include <util/bitwise.h>
#include <util/math.h>
#include "../memory.h"
#include <xorstr.h>

#include <iostream>
#include <memory>
#include <vector>

namespace temple::functions
{
	namespace lua
	{
		temple_fn<void, std::uintptr_t, temple::lua::tvalue*> callTMres;

		temple_fn<void, std::uintptr_t, const temple::lua::tvalue*> callTM;

		temple_fn<temple::lua::tvalue*, std::uintptr_t, signed int> index2adr_f;
		temple_fn<temple::lua::tvalue*, std::uintptr_t, std::uintptr_t> luaH_set;
		temple_fn<temple::lua::tvalue*, std::uintptr_t, std::uintptr_t> luaH_setnum;
		temple_fn<std::uintptr_t, std::uintptr_t, std::uintptr_t> altfrealloc;
		temple_fn<std::uintptr_t, std::uintptr_t, int> newblock_func;
		temple_fn<std::uintptr_t, std::uintptr_t, int> altnewblock_func;
		temple_fn<void, std::uintptr_t, int> free_func;

		temple_fn<void, std::uintptr_t, std::uintptr_t> luaC_barrierf;
		temple_fn<void, std::uintptr_t, std::uintptr_t> luaC_barriert;
		temple_fn<int, std::uintptr_t, std::uintptr_t> f_call{ true };
		temple_fn<void, std::uintptr_t, const char*> fn_luaG_runerrorL;
		temple_fn<void, std::uintptr_t, char> fn_luaC_step;
		temple_fn<void, std::uintptr_t, const temple::lua::tvalue*, const temple::lua::tvalue*> luaG_indexerror;
		temple_fn <std::uintptr_t, std::uintptr_t, const temple::lua::tvalue*> luaT_objtypename;

		temple_fn<int, std::uintptr_t> require{ true };
		temple_fn<int, std::uintptr_t, std::uintptr_t> sandboxthread_f;
		temple_fn<int, std::uintptr_t> spawn{ true };
		temple_fn<int, std::uintptr_t, const char*> load{ true };
		temple_fn<int, std::uintptr_t, void(__cdecl*)(int, int)> luaD_rawrunprotected;
		temple_fn<void, std::uintptr_t, int> luaD_seterrorobj;
		temple_fn<void, std::uintptr_t, const char*> auxgetinfo;

		void luaG_runerrorL(std::uintptr_t thread, const char* msg)
		{
			//std::cout << "runerrorL lua error: " << msg << "\n";
			return fn_luaG_runerrorL(thread, msg);
		}

		void luaC_step(temple::lua::state thread, char n)
		{
			fn_luaC_step(*thread, n);
		}
		double xor_double(double n)
		{
			double result = 0;
			std::uint64_t xor_constant = *reinterpret_cast<std::uint64_t*>(memdata::vars::lua::xor_constant),
				number = 0;

			memcpy(&number, &n, sizeof std::uint64_t);

			number ^= xor_constant;

			memcpy(&result, &number, sizeof result);

			return result;
		}

		bool luaV_equalvector(float* v1, float* v2)
		{
			return v1[0] == v2[0] && v1[1] == v2[1] && v1[2] == v2[2];
		 }

		int luaO_rawequalObj(const temple::lua::tvalue* t1, const temple::lua::tvalue* t2)
		{
			if (t1->tt != t2->tt)
				return 0;

			if (t1->tt == memdata::lua::types::lua_tnil)
			{
				return 1;
			}
			else if (t1->tt == memdata::lua::types::lua_tboolean)
			{
				return t1->v.b == t2->v.b;
			}
			else if (t1->tt == memdata::lua::types::lua_tnumber)
			{
				return xor_double(t1->v.n) == xor_double(t2->v.n);
			}
			else if (t1->tt == memdata::lua::types::lua_tlightuserdata)
			{
				return t1->v.p == t2->v.p;
			}
			else if (t1->tt == memdata::lua::types::lua_tvector)
			{
				return luaV_equalvector(reinterpret_cast<float*>(const_cast<temple::lua::tvalue*>(t1)),
					reinterpret_cast<float*>(const_cast<temple::lua::tvalue*>(t2)));
			}
			else
			{
				return t1->v.gcobj == t2->v.gcobj;
			}
		}

		int luaO_rawequalKey(const temple::lua::tkey* t1, const temple::lua::tvalue* t2)
		{
			if (t1->tt != t2->tt)
				return 0;

			if (t1->tt == memdata::lua::types::lua_tnil)
			{
				return 1;
			}
			else if (t1->tt == memdata::lua::types::lua_tboolean)
			{
				return t1->value.b == t2->v.b;
			}
			else if (t1->tt == memdata::lua::types::lua_tnumber)
			{
				return xor_double(t1->value.n) == xor_double(t2->v.n);
			}
			else if (t1->tt == memdata::lua::types::lua_tlightuserdata)
			{
				return t1->value.p == t2->v.p;
			}
			else if (t1->tt == memdata::lua::types::lua_tvector)
			{
				return luaV_equalvector(reinterpret_cast<float*>(const_cast<temple::lua::tkey*>(t1)),
					reinterpret_cast<float*>(const_cast<temple::lua::tvalue*>(t2)));
			}
			else
			{
				return t1->value.gcobj == t2->v.gcobj;
			}
		}

		const temple::lua::tvalue* luaH_getstr(std::uintptr_t t, std::uintptr_t key)
		{
			auto n = table::hashstr(t, key);

			do
			{
				if ((*reinterpret_cast<std::uint8_t*>(n + memdata::offsets::lua::node::i_key) & 0xF) == memdata::lua::types::lua_tstring &&
					*reinterpret_cast<std::uintptr_t*>(n + memdata::offsets::lua::node::value) == key)
					return  reinterpret_cast<temple::lua::tvalue*>(n);

				auto key_size = (*reinterpret_cast<std::uint32_t*>(n + memdata::offsets::lua::node::i_key)) >> 4;

				if (!key_size)
				{
					return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
				}

				n += 32 * key_size;
			} while (n);

			return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
		}

		void luaC_wakethread(temple::lua::state thread)
		{
			auto prealloc_mark = *reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::prealloc_mark);

			if ((prealloc_mark & 2) != 0)
			{
				auto l_g = thread.l_g;

				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::prealloc_mark) = prealloc_mark & 0xFD;

				auto gcstate = *reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::gcstate);

				if (gcstate == 1 || gcstate == 2 || gcstate == 3)
				{
					*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::gclist) =
						*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::grayagain);

					*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::grayagain) = thread;

					*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::gcobj::marked) &= 0xFBu;
				}
			}
		}

		const temple::lua::tvalue* luaH_getnum(std::uintptr_t t, int key)
		{
			if (key - 1 < *reinterpret_cast<std::size_t*>(t + memdata::offsets::lua::table::sizearray))
				return reinterpret_cast<temple::lua::tvalue*>(memory::ptr_obf::deobf(t + memdata::offsets::lua::table::t_array,
					memdata::offsets::lua::table::t_array_obf)) + (key - 1);

			if (table::get_node(t) != memdata::vars::lua::dummynode)
			{
				std::uintptr_t n = table::hashnum(t, static_cast<double>(key));

				do
				{
					if ((*reinterpret_cast<std::uint8_t*>(n + memdata::offsets::lua::node::i_key) & 0xF) == memdata::lua::types::lua_tnumber &&
						xor_double(*reinterpret_cast<double*>(n + memdata::offsets::lua::node::value)) == static_cast<double>(key))
						return  reinterpret_cast<temple::lua::tvalue*>(n);

					auto key_size = ((*reinterpret_cast<std::uint32_t*>(n + memdata::offsets::lua::node::i_key)) >> 4);

					if (!key_size)
					{
						return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
					}

					n += 32 * key_size;
				} while (n);
			}
			return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
		}

		std::uintptr_t mainposition(std::uintptr_t t, const temple::lua::tvalue* key)
		{
			auto type = key->tt;

			if (type == memdata::lua::types::lua_tboolean)
			{
				return table::hashpow2(t, key->v.b);
			}
			else if (type == memdata::lua::types::lua_tvector)
			{
				return table::hashvec(t, key->v.v);
			}
			else if (type == memdata::lua::types::lua_tnumber)
			{
				return table::hashnum(t, xor_double(key->v.n));
			}
			else if (type == memdata::lua::types::lua_tstring)
			{
				return table::hashstr(t, key->v.gcobj);
			}
			else if (type == memdata::lua::types::lua_tlightuserdata)
			{
				return table::hashpointer(t, key->v.p);
			}
			else
			{
				return table::hashpointer(t, reinterpret_cast<const void*>(key->v.gcobj));
			}
		}

		const temple::lua::tvalue* luaH_get(std::uintptr_t t, const temple::lua::tvalue* key)
		{
			if (key->tt == memdata::lua::types::lua_tnil)
			{
				return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
			}
			else if (key->tt == memdata::lua::types::lua_tstring)
			{
				return luaH_getstr(t, key->v.gcobj);
			}
			else if (key->tt == memdata::lua::types::lua_tnumber)
			{
				double n = xor_double(key->v.n);

				if (static_cast<int>(n) == n)
					return luaH_getnum(t, static_cast<int>(n));
			}

			std::uintptr_t node = mainposition(t, key);

			do
			{
				//std::cout << "key loop of black " << std::dec << reinterpret_cast<temple::lua::tvalue*>(node + memdata::offsets::lua::node::value)->tt << "\n";
				if (luaO_rawequalObj(reinterpret_cast<temple::lua::tvalue*>(node + memdata::offsets::lua::node::value), key))
				{
					return  reinterpret_cast<temple::lua::tvalue*>(node);
				}
				else
				{
					auto key_size = ((*reinterpret_cast<std::uint32_t*>(node + memdata::offsets::lua::node::i_key)) >> 4);

					if (!key_size)
					{
						//std::cout << "key ended\n";
						return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
					}

					node += 32 * key_size;
				}
			} while (node);

			//std::cout << "ended\n";
			return reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
		}

		/*temple::lua::tvalue* newkey(temple::lua::state thread, std::uintptr_t t, const temple::lua::tvalue* key)
		{
			auto mp = mainposition(t, key);

			if (*reinterpret_cast<std::uintptr_t*>(mp + memdata::offsets::lua::node::gval) || mp == memdata::vars::lua::dummynode)
			{
				std::uintptr_t lastfree = table::lastfree(t);

				if(lastfree <= 0)
			}
		}

		temple::lua::tvalue* luaH_set(temple::lua::state thread, std::uintptr_t t, const temple::lua::tvalue* key)
		{
			const temple::lua::tvalue* p = luaH_get(t, key);

			*reinterpret_cast<std::uint8_t*>(t + memdata::offsets::lua::table::flags) = 0;

			if (p != reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject))
				return const_cast<temple::lua::tvalue*>(p);
			else
			{
				if (key->tt == memdata::lua::types::lua_tnil)
					luaG_runerrorL(*thread, "table index is nil");
				else if (key->tt == memdata::lua::types::lua_tnumber && key->v.n != key->v.n)
					luaG_runerrorL(*thread, "table index is NaN");

				return newkey(thread, t, key);
			}
		}*/

		const temple::lua::tvalue* luaT_gettm(std::uintptr_t t, std::uint32_t tm_event, std::uintptr_t ename)
		{
			auto tm = luaH_getstr(t, ename);

			if (tm->tt)
				return tm;

			*reinterpret_cast<std::uint8_t*>(t + memdata::offsets::lua::table::flags) |= 1 << tm_event;
			return 0;
		}

		void luau_poscall(temple::lua::state thread, temple::lua::tvalue* first)
		{
			/*
			auto ci = state::get_ci(*thread);
			auto cip = ci - 32;

			// copy return values into parent stack (but only up to nresults!), fill the rest with nil
			// TODO: it might be worthwhile to handle the case when nresults==b explicitly?
			auto res = reinterpret_cast<temple::lua::tvalue*>(callinfo::get_func(ci));
			auto vali = first;
			auto valend = thread->top;

			int i;
			for (i = ci->nresults; i != 0 && vali < valend; i--)
				setobj(res++, vali++);
			while (i-- > 0)
				setnilvalue(res++);

			// pop the stack frame
			L->ci = cip;
			L->base = cip->base;
			L->top = (ci->nresults == LUA_MULTRET) ? res : cip->top;*/
		}

		const temple::lua::tvalue* luaT_gettmbyobj(temple::lua::state thread, const temple::lua::tvalue* obj, std::uint32_t tms_event)
		{

			std::uintptr_t metatable = 0;

			if (obj->tt == memdata::lua::types::lua_tuserdata)
			{
				auto userdata = obj->v.gcobj;
				metatable = memory::ptr_obf::deobf(
					userdata + memdata::offsets::lua::userdata::metatable, memdata::offsets::lua::userdata::metatable_obf);
			}
			else if (obj->tt == memdata::lua::types::lua_ttable)
			{
				auto table = obj->v.gcobj;
				metatable = memory::ptr_obf::deobf(
					table + memdata::offsets::lua::table::metatable, memdata::offsets::lua::table::metatable_obf);
			}
			else
			{
				//std::cout << obj->tt << "\n";
				metatable = *reinterpret_cast<std::uintptr_t*>(thread.l_g + 4 * obj->tt + memdata::offsets::lua::state::l_g::metatable);
			}
			std::uintptr_t tmname = thread.l_g + 4 * tms_event + memdata::offsets::lua::state::l_g::tmname;

			tmname = memory::ptr_obf::deobf(tmname, memdata::offsets::lua::state::l_g::tmname_obf);

			return metatable ? luaH_getstr(metatable, tmname) : 
				reinterpret_cast<temple::lua::tvalue*>(memdata::vars::lua::nilobject);
		}

		void setobj(temple::lua::tvalue* obj1, const temple::lua::tvalue* obj2)
		{
			obj1->v = obj2->v;
			obj1->extra = obj2->extra;
			obj1->tt = obj2->tt;
		}
		
		std::uintptr_t luaM_realloc_(temple::lua::state thread, std::size_t size, std::uint8_t category)
		{
			auto l_g = thread.l_g;

			std::uintptr_t block = 0;

			if (size - 1 >= 0x200 || *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size) < 0)
			{
				auto result = altfrealloc(thread, l_g + memdata::offsets::lua::state::l_g::altrealloc_link,
					size + memdata::offsets::lua::gcobj::entry::size, size, 1);

				block = result + memdata::offsets::lua::gcobj::entry::size;

				*reinterpret_cast<std::uintptr_t*>(result + 28) -= *reinterpret_cast<std::uintptr_t*>(result + 20);
				++* reinterpret_cast<std::uintptr_t*>(result + 32);
			}
			else
			{
				block = altnewblock_func(*thread, *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size));
			}

			if (!block && size)
				luaG_runerrorL(*thread, xorstr("error during allocation"));

			*reinterpret_cast<std::size_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) += size;
			*reinterpret_cast<std::size_t*>(l_g + 4 * category + memdata::offsets::lua::state::l_g::alloc_category) += size;

			return block;
		}

		std::uintptr_t luaM_newarray(temple::lua::state thread, std::size_t size, std::uint8_t category)
		{
			auto l_g = thread.l_g;

			std::uintptr_t block = 0;


			if (size - 1 >= 0x200 || *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size) < 0)
			{
				auto frealloc = *reinterpret_cast<int(__cdecl**)(std::uintptr_t, std::uintptr_t, std::uintptr_t, std::size_t)>
					(l_g + memdata::offsets::lua::state::l_g::frealloc);

				block = frealloc(*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::ud), 0, 0, size);
			}
			else
				block = newblock_func(*thread, *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size));

			if (!block && size)
			{
				luaG_runerrorL(*thread, xorstr("error during allocation"));
			}

			*reinterpret_cast<std::size_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) += size;
			*reinterpret_cast<std::size_t*>(l_g + 4 * category + memdata::offsets::lua::state::l_g::alloc_category) += size;

			return block;
		}

		std::uintptr_t luaM_reallocvector(temple::lua::state thread, std::uintptr_t arr, std::size_t sizearray, std::size_t narray, std::uint8_t category)
		{
			std::uintptr_t block = 0;

			char narr;
			char sarr;

			if (narray - 1 >= 0x200)
			{
				narr = -1;
			}
			else
			{
				narr = *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + narray);
			}

			if (sizearray - 1 >= 0x200)
			{
				sarr = -1;
			}
			else
			{
				sarr = *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + sizearray);
			}
			
			if (narr >= 0)
			{
				newblock_func(*thread, narr);
			}
			else
			{
				auto frealloc = *reinterpret_cast<int(__cdecl**)(std::uintptr_t, std::uintptr_t, std::uintptr_t, std::uintptr_t, std::size_t)>
					(thread.l_g + memdata::offsets::lua::state::l_g::frealloc);

				if (sarr < 0)
				{
					block = frealloc(*thread, *reinterpret_cast<std::uintptr_t*>(thread.l_g + memdata::offsets::lua::state::l_g::ud), arr, sizearray, narray);

					if (!block && narray)
						luaG_runerrorL(*thread, xorstr("error during allocation"));

					*reinterpret_cast<std::size_t*>(thread.l_g + memdata::offsets::lua::state::l_g::totalbytes) += narray - sizearray;
					*reinterpret_cast<std::size_t*>(thread.l_g + 4 * category + memdata::offsets::lua::state::l_g::alloc_category) += narray - sizearray;

					return block;
				}

				block = frealloc(*thread, *reinterpret_cast<std::uintptr_t*>(thread.l_g + memdata::offsets::lua::state::l_g::ud), 0, 0, narray);
			}

			if (!block && narray)
				luaG_runerrorL(*thread, xorstr("error during allocation"));

			std::size_t sz = narray;

			if (sizearray < narray)
				sz = sizearray;

			memcpy(reinterpret_cast<void*>(block), reinterpret_cast<void*>(arr), sz);

			if (narr < 0)
			{
				auto freealloc = *reinterpret_cast<void(__cdecl**)(std::uintptr_t, std::uintptr_t, std::uintptr_t, std::uintptr_t, std::size_t)>
					(thread.l_g + memdata::offsets::lua::state::l_g::frealloc);

				freealloc(*thread, *reinterpret_cast<std::uintptr_t*>(thread.l_g + memdata::offsets::lua::state::l_g::ud), arr, sizearray, 0);
			}
			else
			{
				free_func(*thread, narr, arr);
			}

			*reinterpret_cast<std::size_t*>(thread.l_g + memdata::offsets::lua::state::l_g::totalbytes) += narray - sizearray;
			*reinterpret_cast<std::size_t*>(thread.l_g + 4 * category + memdata::offsets::lua::state::l_g::alloc_category) += narray - sizearray;

			return block;
		}

		std::uintptr_t luaM_free(temple::lua::state thread, std::uintptr_t result, std::size_t size, std::uint8_t category)
		{
			auto l_g = thread.l_g;

			std::uintptr_t block = 0;

			if (size - 1 >= 0x200 || *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size) < 0)
			{
				auto frealloc = *reinterpret_cast<void(__cdecl**)(std::uintptr_t, std::uintptr_t, std::uintptr_t, std::uintptr_t, std::size_t)>
					(l_g + memdata::offsets::lua::state::l_g::frealloc);

				frealloc(*thread, *reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::ud), result, size, 0);
			}
			else
				free_func(*thread, *reinterpret_cast<char*>(memdata::vars::lua::realloc_table + size), result);

			*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) -= size;
			*reinterpret_cast<std::uintptr_t*>(l_g + 4 * category + memdata::offsets::lua::state::l_g::alloc_category) -= size;

			return category;
		}

		void luaC_barrier(temple::lua::state thread, std::uintptr_t p, std::uintptr_t v)
		{
			if (gcobj::valis_white(v) && gcobj::is_black(p))
				luaC_barrierf(*thread, p, v);
		}

		void luaC_linkupval(temple::lua::state thread, std::uintptr_t uv)
		{
			auto l_g = thread.l_g;

			//* reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::gcobj::next) =
			//*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::rootgc);

			//*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::rootgc) = uv;

			if (gcobj::is_gray(uv))
			{
				auto gcstate = *reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::gcstate);

				if (gcstate == 1 || gcstate == 2)
				{
					//gcobj::gray2black(uv);
					luaC_barrier(thread, uv, *reinterpret_cast<std::uintptr_t*>(uv + memdata::offsets::lua::upvalue::v));
				}
				else
				{
					gcobj::make_white(l_g, uv);
				}
			}
		}

		void luaF_close(temple::lua::state thread, temple::lua::tvalue* level)
		{

			auto openupval = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::openupval);

			for (; openupval; openupval = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::openupval))
			{
				if (*reinterpret_cast<temple::lua::tvalue**>(openupval + memdata::offsets::lua::upvalue::v) < level)
					break;
				
				upvalue::unlink_upvalue(openupval);

				**reinterpret_cast<std::uintptr_t**>(openupval + 20) = *reinterpret_cast<std::uintptr_t*>(openupval + 16);
				std::uintptr_t old = *reinterpret_cast<std::uintptr_t*>(openupval + 16);
				if (old)
					*reinterpret_cast<std::uintptr_t*>(old + 20) = *reinterpret_cast<std::uintptr_t*>(openupval + 20);

				if (gcobj::is_dead(thread, openupval))
				{
					*reinterpret_cast<std::uintptr_t*>(openupval + memdata::offsets::lua::upvalue::v) = openupval + memdata::offsets::lua::upvalue::prev;
				}
				else
				{
					setobj(reinterpret_cast<temple::lua::tvalue*>(openupval + memdata::offsets::lua::upvalue::prev),
						*reinterpret_cast<temple::lua::tvalue**>(openupval + memdata::offsets::lua::upvalue::v));

					*reinterpret_cast<std::uintptr_t*>(openupval + memdata::offsets::lua::upvalue::v) =
						openupval + memdata::offsets::lua::upvalue::prev;

					luaC_linkupval(thread, openupval);
				}
				/*
				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::openupval) =
					*reinterpret_cast<std::uintptr_t*>(openupval + memdata::offsets::lua::gcobj::next);

				if (gcobj::is_dead(thread, openupval))
				{
					if (*reinterpret_cast<std::uintptr_t*>(openupval + memdata::offsets::lua::upvalue::v) !=
						openupval + memdata::offsets::lua::upvalue::u_value)
					{
						upvalue::unlink_upvalue(openupval);
					}
					luaM_free(thread, openupval, 32, *reinterpret_cast<std::uint8_t*>(openupval + memdata::offsets::lua::gcobj::category));
				}
				else
				{
					upvalue::unlink_upvalue(openupval);
					setobj(reinterpret_cast<temple::lua::tvalue*>(openupval + memdata::offsets::lua::upvalue::u_value),
						*reinterpret_cast<temple::lua::tvalue**>(openupval + memdata::offsets::lua::upvalue::v));

					*reinterpret_cast<std::uintptr_t*>(openupval + memdata::offsets::lua::upvalue::v) =
						openupval + memdata::offsets::lua::upvalue::u_value;

					luaC_linkupval(thread, openupval);
				}
				*/
			}
		}

		void luaD_reallocCI(temple::lua::state thread, int newsize)
		{
			auto oldci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);

			luaM_reallocvector(*thread, oldci, 24 * *reinterpret_cast<std::size_t*>(thread + memdata::offsets::lua::state::size_ci),
				24 * newsize, *reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::category));

			*reinterpret_cast<std::size_t*>(thread + memdata::offsets::lua::state::size_ci) = newsize;

			auto ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);
			*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci) = (ci - oldci) +
				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);

			*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::end_ci) =
				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci) +
				*reinterpret_cast<std::size_t*>(thread + memdata::offsets::lua::state::size_ci) - 1;
		}

		int luaD_pcall(temple::lua::state thread, std::uintptr_t func, std::uintptr_t u, unsigned int old_top, std::ptrdiff_t ef)
		{
			luaC_wakethread(thread);

			int status;
			unsigned short oldnCcalls = *reinterpret_cast<std::uint16_t*>(thread + memdata::offsets::lua::state::nccalls);
			std::ptrdiff_t old_ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci) -
				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);

			status = luaD_rawrunprotected(*thread, reinterpret_cast<void(__cdecl*)(int, int)>(func), u);

			if (status)
			{

				if (ef)
				{
					if(luaD_rawrunprotected(*thread, reinterpret_cast<void(__cdecl*)(int, int)>
						(memdata::functions::lua::callerrfunc.ptr), reinterpret_cast<std::uintptr_t>(state::get_stack(thread)) + ef))
					status = 5;
				}
				auto pcall_fn = *reinterpret_cast<void(__cdecl**)(std::uintptr_t)>
					(thread.l_g + memdata::offsets::lua::state::l_g::pcall_fn);

				if (pcall_fn)
				{
					pcall_fn(*thread);

					if (*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::status) == 6)
						return 0;
				}
				auto oldtop = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::stack) + old_top;

				luaF_close(thread, reinterpret_cast<temple::lua::tvalue*>(oldtop));
				luaD_seterrorobj(thread, status, reinterpret_cast<temple::lua::tvalue*>(oldtop));

				*reinterpret_cast<std::uint16_t*>(thread + memdata::offsets::lua::state::nccalls) = oldnCcalls;
				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci) =
					*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci) + old_ci;

				auto ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);

				*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base) =
					*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::base);

				if (*reinterpret_cast<std::size_t*>(thread + memdata::offsets::lua::state::size_ci) > 20000)
				{
					if (((*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci) -
						*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci)) / 24) + 1 > 20000)
					{
						luaD_reallocCI(*thread, 20000);
					}
				}
			}

			return status;
		}

		int pcall(temple::lua::state thread, int nargs, int nresults, int errfunc)
		{
			auto stack = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::stack);

			std::uintptr_t func = 0;

			if (errfunc)
			{
				auto obj = index2adr(thread, errfunc);

				func = reinterpret_cast<std::uintptr_t>(obj) - stack;
			}

			std::uintptr_t c[2];

			c[1] = nresults;
			c[0] = reinterpret_cast<std::uintptr_t>(thread.top - (nargs + 1));

			auto prealloc_mark = *reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::prealloc_mark) & 1;
			*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::prealloc_mark) |= 1;

			auto status = functions::lua::luaD_pcall(*thread, memdata::functions::lua::f_call.ptr, 
				reinterpret_cast<std::uintptr_t>(c), c[0] - stack, func);

			if (nresults == -1)
			{
				auto ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);
				auto top = reinterpret_cast<std::uintptr_t>(thread.top);
				if (top >= *reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top))
					*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top) = top;
			}

			if (!prealloc_mark)
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::prealloc_mark) &= 0xFEu;

			return status;
		}

		int yield(temple::lua::state thread, int nresults)
		{
			if (*reinterpret_cast<std::uint16_t*>(thread + memdata::offsets::lua::state::nccalls) >
				*reinterpret_cast<std::uint16_t*>(thread + memdata::offsets::lua::state::baseccalls))
			{
				luaG_runerrorL(*thread, xorstr("attempt to yield across metamethod/C-call boundary"));
			}

			thread.base = thread.top - nresults;
			*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::status) = 1;

			return -1;
		}

		void luaS_resize(temple::lua::state thread, std::size_t size)
		{
			auto l_g = thread.l_g;

			if (*reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::gcstate) == 4)
				return;

			if (size > 0x3FFFFFFF)
				return;

			auto newhash = luaM_newarray(thread, 4 * size, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::state::category));

			memset(reinterpret_cast<void*>(newhash), 0, 4 * ((4 * size) >> 2));

			for (std::size_t i = 0; i < *reinterpret_cast<std::uint32_t*>(
				l_g + memdata::offsets::lua::state::l_g::strt_size); i++)
			{
				auto hash = *reinterpret_cast<std::uintptr_t*>(*reinterpret_cast<std::uintptr_t*>(
					l_g + memdata::offsets::lua::state::l_g::strt_hash) + 4 * i);

				while (hash)
				{
					auto next = *reinterpret_cast<std::uintptr_t*>(hash + memdata::offsets::lua::gcobj::next);

					auto h = string::get_hash(hash);
					auto h1 = gcobj::l_mod(h, size);
					*reinterpret_cast<std::uintptr_t*>(hash + memdata::offsets::lua::gcobj::next) = *reinterpret_cast<std::uintptr_t*>(newhash + 4 * h1);
					*reinterpret_cast<std::uintptr_t*>(newhash + 4 * h1) = hash;

					hash = next;
				}
			}

			luaM_free(thread, *reinterpret_cast<std::uintptr_t*>(
				l_g + memdata::offsets::lua::state::l_g::strt_hash), 4 * size, *reinterpret_cast<std::uint8_t*>(
					thread + memdata::offsets::lua::state::category));

			*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::strt_size) = size;
			*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::strt_hash) = newhash;
		}

		std::uintptr_t newlstr(temple::lua::state thread, const char* str, std::size_t len)
		{
			std::uintptr_t l_g = thread.l_g;

			std::uint32_t h = len;

			std::size_t len_counter = len;
			char* nc_str = const_cast<char*>(str);

			std::uint32_t hsh_1 = 0;
			std::uint32_t hsh_2 = 0;

			if (len >= 32)
			{
				std::size_t count = (len - 20) / 12;

				do
				{
					std::uint32_t magic = *reinterpret_cast<std::uint32_t*>(nc_str + 8) + h;
					hsh_1 = (magic ^ (hsh_1 + *reinterpret_cast<std::uint32_t*>(nc_str))) -
						temple::util::bitwise::rotr32(magic, 14);
					hsh_2 = (hsh_1 ^ (hsh_2 + *reinterpret_cast<std::uint32_t*>(nc_str + 4))) -
						temple::util::bitwise::rotr32(hsh_1, 11);

					h = (hsh_2 ^ magic) - temple::util::bitwise::rotl32(hsh_2, 7);

					nc_str += 12;
					len_counter -= 12;

					--count;

				} while (count);
			}
				
			for (; len_counter; --len_counter)
			{
				h ^= (h << 5) + (h >> 2) + *reinterpret_cast<std::uint8_t*>(nc_str + len_counter - 1);
			}

			std::uint32_t o = *reinterpret_cast<std::uintptr_t*>(*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::strt_hash) + 4 *
				gcobj::l_mod(h, *reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::strt_size)));

			for (; o; o = *reinterpret_cast<std::uintptr_t*>(o + memdata::offsets::lua::gcobj::next))
			{
				if (memory::ptr_obf::deobf(o + memdata::offsets::lua::string::tsv_len,
					memdata::offsets::lua::string::tsv_len_obf) == len &&
					memcmp(str, reinterpret_cast<const char*>(o + memdata::offsets::lua::string::raw), len) == 0)
				{
					if (gcobj::is_dead(thread, o))
						gcobj::changewhite(o);
					return o;
				}
			}

			if (len > 0x40000000)
				lua::luaG_runerrorL(*thread, xorstr("memory allocation error: block too big"));

			std::uintptr_t ts = lua::luaM_realloc_(thread, len + 21, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::state::category));

			memory::ptr_obf::obf(ts + memdata::offsets::lua::string::tsv_len, len,
				memdata::offsets::lua::string::tsv_len_obf);

			memory::ptr_obf::obf(ts + memdata::offsets::lua::string::hash, h,
				memdata::offsets::lua::string::hash_obf);

			*reinterpret_cast<std::uint8_t*>(ts + memdata::offsets::lua::gcobj::marked) =
				*reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::currentwhite) & gcobj::white_bits;

			*reinterpret_cast<std::uint8_t*>(ts + memdata::offsets::lua::gcobj::tt) = memdata::lua::types::lua_tstring;

			*reinterpret_cast<std::uint8_t*>(ts + memdata::offsets::lua::gcobj::category) =
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::category);

			memcpy(reinterpret_cast<void*>(ts + memdata::offsets::lua::string::raw), str, len);

			*reinterpret_cast<std::uint8_t*>(ts + len + memdata::offsets::lua::string::raw) = 0;

			auto newstring_fn = *reinterpret_cast<int(__cdecl**)(std::uintptr_t, std::size_t)>
				(l_g + memdata::offsets::lua::state::l_g::newstring_fn);

			int magic_id = -1;

			if (newstring_fn)
				magic_id = newstring_fn(ts + memdata::offsets::lua::string::raw, len);

			*reinterpret_cast<std::int16_t*>(ts + memdata::offsets::lua::string::str_id) = magic_id;

			std::uintptr_t strt_entry = *reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::strt_hash) + 4 *
				(h & (*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::strt_size) - 1));

			*reinterpret_cast<std::uintptr_t*>(ts + memdata::offsets::lua::gcobj::next) = *reinterpret_cast<std::uint32_t*>(strt_entry);

			*reinterpret_cast<std::uintptr_t*>(strt_entry) = ts;

			*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::nuse) += 1;

			auto strt_size = *reinterpret_cast<std::size_t*>(l_g + memdata::offsets::lua::state::l_g::strt_size);
			auto strt_nuse = *reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::nuse);

			if (strt_nuse > strt_size && strt_size <= 0x3FFFFFFF)
				luaS_resize(*thread, strt_size * 2);

			return ts;
		}

		void luaC_barrierback(temple::lua::state thread, std::uintptr_t t)
		{
			auto l_g = thread.l_g;

			gcobj::black2gray(t);

			*reinterpret_cast<std::uintptr_t*>(t + memdata::offsets::lua::table::gclist) =
				*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::grayagain);

			*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::grayagain) = t;
		}

		int luaD_call(temple::lua::state thread, temple::lua::tvalue* fn, int nresults)
		{
			std::uintptr_t buff[2];

			buff[0] = reinterpret_cast<std::uintptr_t>(fn);
			buff[1] = nresults;

			return f_call(*thread, reinterpret_cast<std::uintptr_t>(&buff));
		}

		int lua_call(temple::lua::state thread, int nargs, int nresults)
		{
			auto func = functions::lua::index2adr(thread, -(nargs + 1));
			auto result = luaD_call(thread, func, nresults);

			if (nresults == -1)
			{
				auto ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);
				auto top = reinterpret_cast<std::uintptr_t>(thread.top);
				if (top >= *reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top))
					*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top) = top;
			}

			return result;
		}

		int arrayindex(temple::lua::tvalue* key) {
			if (key->tt == memdata::lua::types::lua_tnumber) {
				auto n = xor_double(key->v.n);
				auto k = static_cast<int>(n);
				if (static_cast<double>(k) == n)
					return k;
			}
			return -1;  /* `key' did not match some condition */
		}

		int findindex(temple::lua::state thread, std::uintptr_t t, temple::lua::tvalue* key)
		{
			int i;
			if (key->tt == memdata::lua::types::lua_tnil)
				return -1; /* first iteration */
			i = key->tt == memdata::lua::types::lua_tnumber ? arrayindex(key) : -1;
			if (0 < i && i <= table::get_sizearray(t)) /* is `key' inside array part? */
				return i - 1;               /* yes; that's the index (corrected to C) */
			else
			{
				std::uintptr_t n = mainposition(t, key);
				for (;;)
				{ /* check whether `key' is somewhere in the chain */
					/* key may be dead already, but it is ok to use it in `next' */
					if (luaO_rawequalKey(&reinterpret_cast<temple::lua::luanode*>(n)->key, key) || ((node::gkey(n) & 0xF) == 12 && gcobj::is_collectable(key) && node::key2tval(n)->v.gcobj == key->v.gcobj))
					{
						i = static_cast<int>((n - table::get_node(t, 0)) / 32); /* key index in hash table */
						/* hash elements are numbered after array ones */
						return i + table::get_sizearray(t);
					}
					if (node::gnext(n) == 0)
						break;
					n += node::gnext(n);
				}
				luaG_runerrorL(thread, xorstr("invalid key to 'next'")); /* key not found */
			}
		}

		int luaH_next(temple::lua::state thread, std::uintptr_t t, temple::lua::tvalue* key)
		{
			int i = findindex(thread, t, key); /* find original element */
			for (i++; i < table::get_sizearray(t); i++)
			{ /* try first array part */
				if (!(table::get_array(t, i)->tt == memdata::lua::types::lua_tnil))
				{ /* a non-nil value? */
					key->v.n = xor_double(static_cast<double>(i + 1));
					key->tt = memdata::lua::types::lua_tnumber;
					setobj(key + 1, table::get_array(t, i));
					return 1;
				}
			}
			for (i -= table::get_sizearray(t); i < table::size_node(t); i++)
			{ /* then hash part */
				if (!(node::gval(table::get_node(t, i))->tt == memdata::lua::types::lua_tnil))
				{ /* a non-nil value? */
					//setobj(key, node::key2tval(table::get_node(t, i)));
					auto node = reinterpret_cast<temple::lua::luanode*>(table::get_node(t, i));

					key->v = node->key.value;
					key->extra = node->key.extra;
					key->tt = node->key.tt;

					setobj(key + 1, node::gval(table::get_node(t, i)));
					return 1;
				}
			}
			return 0; /* no more elements */
		}

		int next(temple::lua::state thread, int idx) {
			int more;
			auto t = index2adr(thread, idx);
			more = luaH_next(thread, t->v.gcobj, thread.top - 1);
			if (more) {
				thread.top += 1;
			}
			else  /* no more elements */
				thread.top -= 1;  /* remove key */
	
			return more;
		}

		void luaV_settable(temple::lua::state thread, const temple::lua::tvalue* t, temple::lua::tvalue* key, temple::lua::tvalue* val)
		{
			temple::lua::tvalue temp;

			for (int loop = 0; loop < 100; loop++)
			{
				const temple::lua::tvalue* tm;
				if (t->tt == memdata::lua::types::lua_ttable)
				{
					std::uintptr_t h = t->v.gcobj;

					auto old_index = luaH_set(*thread, h, key);

					*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::lcni) =
						(reinterpret_cast<std::uintptr_t>(old_index) - memory::ptr_obf::deobf(h + memdata::offsets::lua::table::node,
							memdata::offsets::lua::table::node_obf)) >> 5;

					auto mt = table::get_metatable(h);

					if (old_index->tt != memdata::lua::types::lua_tnil || mt == 0 ||
						(table::get_flags(mt) & (1 << memdata::offsets::lua::state::l_g::newindex_tm)) != 0 || 
						(tm = luaT_gettm(mt, memdata::offsets::lua::state::l_g::newindex_tm,
							memory::ptr_obf::deobf(thread.l_g + memdata::offsets::lua::state::l_g::tmname_newindex,
								memdata::offsets::lua::state::l_g::tmname_obf))) == 0)
					{
						setobj(old_index, val);

						if (gcobj::is_collectable(val) && gcobj::is_black(h) && gcobj::is_white(val->v.gcobj))
						{
							luaC_barriert(*thread, h, val->v.gcobj);
						}
						return;
					}
				}
					tm = luaT_gettmbyobj(thread, t, memdata::offsets::lua::state::l_g::newindex_tm);

					if (tm->tt == memdata::lua::types::lua_tnil)
						luaG_indexerror(thread, t, key);

					if (tm->tt == memdata::lua::types::lua_tfunction)
					{
						/*auto& top = thread.top;
						setobj(top, tm);
						setobj(top + 1, t);
						setobj(top + 2, key);
						setobj(top + 3, val);

						if (state::get_stack_last(thread) - thread.top <= 4 * sizeof(temple::lua::tvalue))
							luaD_growstack(thread, 4);

						auto fn = top;

						top = fn + 4;

						return luaD_call(thread, fn, 0);*/

						callTM(*thread, tm, t, key, val);

						return;
					}
				setobj(&temp, tm);

				t = &temp;
			}

			luaG_runerrorL(*thread, xorstr("loop in settable"));
		}

		void luaV_gettable(temple::lua::state thread, const temple::lua::tvalue* t, const temple::lua::tvalue* key, temple::lua::tvalue* val)
		{
			for (int loop = 0; loop < 100; loop++)
			{
				const temple::lua::tvalue* tm;

				if (t->tt == memdata::lua::types::lua_ttable)
				{
					std::uintptr_t h = t->v.gcobj;

					auto old_index = luaH_get(h, key);

					*reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::lcni) =
						(reinterpret_cast<std::uintptr_t>(old_index) - memory::ptr_obf::deobf(h + memdata::offsets::lua::table::node,
							memdata::offsets::lua::table::node_obf)) >> 5;

					auto mt = table::get_metatable(h);

					if (old_index->tt != memdata::lua::types::lua_tnil 
						|| mt == 0 
						|| (table::get_flags(mt) & (1 << memdata::offsets::lua::state::l_g::index_tm)) != 0
						|| (tm = luaT_gettm(mt, memdata::offsets::lua::state::l_g::index_tm, memory::ptr_obf::deobf(thread.l_g +
							memdata::offsets::lua::state::l_g::tmname_index, memdata::offsets::lua::state::l_g::tmname_obf))) == 0)
					{
						setobj(val, old_index);
						return;
					}

				}
			
					tm = luaT_gettmbyobj(thread, t, memdata::offsets::lua::state::l_g::index_tm);

					if (tm->tt == memdata::lua::types::lua_tnil)
						luaG_indexerror(thread, t, key);

					if (tm->tt == memdata::lua::types::lua_tfunction)
					{
						/*auto result = (val - state::get_stack(thread)) / sizeof(temple::lua::tvalue);

						auto& top = thread.top;
						setobj(top, tm);
						setobj(top + 1, t);
						setobj(top + 2, key);

						if (state::get_stack_last(thread) - thread.top <= 3 * sizeof(temple::lua::tvalue))
							luaD_growstack(thread, 3);

						auto fn = top;

						top = fn + 3;

						luaD_call(thread, fn, 1);

						top -= 1;

						auto& res = state::get_stack(thread);

						setobj(res + result, top);*/

						callTMres(*thread, val, tm, t, key);

						return;
					}
				
				t = tm;
			}

			luaG_runerrorL(*thread, xorstr("loop in gettable"));
		}

		void setfield(temple::lua::state thread, int idx, const char* field)
		{
			auto obj = index2adr(thread, idx);

			temple::lua::tvalue key; 

			key.v.gcobj = newlstr(*thread, field, std::strlen(field));
			key.tt = memdata::lua::types::lua_tstring;

			auto top = thread.top - 1;

			functions::lua::luaV_settable(*thread, obj, &key, top);

			thread.top -= 1;
		}

		void getfield(temple::lua::state thread, int idx, const char* field)
		{
			auto obj = index2adr(thread, idx);

			temple::lua::tvalue key;

			key.v.gcobj = newlstr(*thread, field, std::strlen(field));
			key.tt = memdata::lua::types::lua_tstring;

			auto top = thread.top;

			functions::lua::luaV_gettable(*thread, obj, &key, top);

			thread.top += 1;
		}

		void rawgetfield(temple::lua::state thread, int idx, const char* field)
		{
			pushstring(thread, field);
			rawget(thread, (idx < 0 && idx > -10000) ? idx - 1 : idx);
		}

		void rawset(temple::lua::state thread, int idx)
		{
			auto t = index2adr(thread, idx);

			if (table::isreadonly(t->v.gcobj))
				luaG_runerrorL(thread, xorstr("Attempt to modify a readonly table"));

			setobj(luaH_set(thread, t->v.gcobj, thread.top - 2), thread.top - 1);
			//luaC_barriert(thread, t->v.gcobj, thread.top - 1);

			thread.top -= 2;
		}

		void rawseti(temple::lua::state thread, int idx, int n)
		{
			auto obj = index2adr(thread, idx);

			auto table = obj->v.gcobj;
			auto top = thread.top - 1;

			temple::lua::tvalue number;

			number.tt = memdata::lua::types::lua_tnumber;
			number.v.n = xor_double(n);

			setobj(luaH_setnum(*thread, table, &number), top);

			obj = index2adr(thread, -1);
			
			if (obj->tt >= memdata::lua::types::lua_tstring)
			{
				if ((functions::lua::gcobj::get_marked(table) & 4) != 0 &&
					(functions::lua::gcobj::get_marked(obj->v.gcobj) & 3) != 0)
				{
					luaC_barriert(*thread, table, obj->v.gcobj);
				}
			}

			thread.top -= 1;
		}

		int rawgeti(temple::lua::state thread, int idx, int n)
		{
			auto t = index2adr(thread, idx);
			setobj(thread.top, luaH_getnum(t->v.gcobj, n));
			
			thread.top += 1;

			return (thread.top - 1)->tt;
		}

		void gettable(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(thread, idx);

			setobj(thread.top - 1, luaH_get(obj->v.gcobj, thread.top - 1));

			luaV_gettable(thread, obj, thread.top - 1, thread.top - 1);
		}

		void rawget(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(thread, idx);

			setobj(thread.top - 1, luaH_get(obj->v.gcobj, thread.top - 1));
		}

		void setmetatable(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(thread, idx);

			std::uintptr_t mt = 0;
			auto top = thread.top - 1;

			if (top->tt)
				mt = top->v.gcobj;

			if (obj->tt == memdata::lua::types::lua_ttable)
			{
				auto table = obj->v.gcobj;

				if (*reinterpret_cast<std::uint8_t*>(table + memdata::offsets::lua::table::readonly))
					luaG_runerrorL(*thread, xorstr("Attempt to modify a readonly table"));

				memory::ptr_obf::obf(table + memdata::offsets::lua::table::metatable, mt,
					memdata::offsets::lua::table::metatable_obf);


				if (mt && gcobj::is_white(mt) && gcobj::is_black(table))
					luaC_barriert(*thread, table, mt);
			}
			else if (obj->tt == memdata::lua::types::lua_tuserdata)
			{
				auto userdata = obj->v.gcobj;

				memory::ptr_obf::obf(userdata + memdata::offsets::lua::userdata::metatable, mt,
					memdata::offsets::lua::userdata::metatable_obf);

				if (mt && gcobj::is_white(mt) && gcobj::is_black(userdata))
					luaC_barrierf(*thread, userdata, mt);
			}
			else
			{
				auto l_g = thread.l_g;
				*reinterpret_cast<std::uintptr_t*>(l_g + 4 * obj->tt + memdata::offsets::lua::state::l_g::metatable) = mt;
			}

			thread->top -= 1;
		}

		int getmetatable(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(thread, idx);

			std::uintptr_t metatable = 0;

			if (obj->tt == memdata::lua::types::lua_tuserdata)
			{
				auto userdata = obj->v.gcobj;
				metatable = memory::ptr_obf::deobf(
					userdata + memdata::offsets::lua::userdata::metatable, memdata::offsets::lua::userdata::metatable_obf);
			}
			else if (obj->tt == memdata::lua::types::lua_ttable)
			{
				auto table = obj->v.gcobj;
				metatable = memory::ptr_obf::deobf(
					table + memdata::offsets::lua::table::metatable, memdata::offsets::lua::table::metatable_obf);
			}
			else
			{
				auto l_g = thread.l_g;
				metatable = *reinterpret_cast<std::uintptr_t*>(l_g + 4 * obj->tt + memdata::offsets::lua::state::l_g::metatable);
			}

			if (metatable)
			{
				auto& top = thread->top;

				top->v.gcobj = metatable;
				top->tt = memdata::lua::types::lua_ttable;
				top++;

				return 1;
			}
			else
			{
				return 0;
			}
		}

		void luaC_link(temple::lua::state thread, std::uintptr_t ud, std::uint8_t tt)
		{
			auto l_g = thread.l_g;

			//*reinterpret_cast<std::uintptr_t*>(ud + memdata::offsets::lua::gcobj::next) =
			//	*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::rootgc);

			//*reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::rootgc) = ud;

			*reinterpret_cast<std::uint8_t*>(ud + memdata::offsets::lua::gcobj::marked) =
				(*reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::currentwhite)) & 3;

			*reinterpret_cast<std::uint8_t*>(ud + memdata::offsets::lua::gcobj::tt) = tt;

			*reinterpret_cast<std::uint8_t*>(ud + memdata::offsets::lua::gcobj::category) = 
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::category);
		}

		std::uintptr_t newlclosure(temple::lua::state thread, int nups, std::uintptr_t env, std::uintptr_t proto)
		{
			auto closure = lua::luaM_realloc_(*thread, 16 * (nups) + 24, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::state::category));

			luaC_link(thread, closure, memdata::lua::types::lua_tfunction);

			closure::set_env(closure, env);
			closure::set_isc(closure, false);
			closure::set_nups(closure, nups);
			closure::set_maxstacksize(closure, proto::get_maxstacksize(proto));
			closure::set_source(closure, proto);

			while (nups)
			{
				nups--;
				//auto& upvalue = *reinterpret_cast<temple::lua::tvalue**>(closure + memdata::offsets::lua::closure::l_ups + nups * 16);
				//upvalue->tt = memdata::lua::types::lua_tnil;
				*reinterpret_cast<int*>(closure + memdata::offsets::lua::closure::l_ups + nups * 16 + 12) = memdata::lua::types::lua_tnil;
			}

			return closure;
		}

		std::uintptr_t newthread(temple::lua::state thread)
		{
			auto l_g = thread.l_g;

			if (*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) >=
				*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::gcthreshold))
				luaC_step(thread, 1);

			luaC_wakethread(thread);

			std::uintptr_t new_thread = lua::luaM_realloc_(thread, 76, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::state::category));

			luaC_link(thread, new_thread, memdata::lua::types::lua_tthread);

			state::set_l_g(new_thread, l_g);

			/* Pre-init state */

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::stack)		= 0;
			state::set_stacksize(new_thread, 0);
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::openupval)	= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::size_ci)		= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::nccalls)		= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::status)		= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::ci)			= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::base_ci)		= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::namecall)		= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::lcni)			= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::extraspace)	= 0;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::unk_0)		= 0;
			*reinterpret_cast<std::uint8_t*>(new_thread + memdata::offsets::lua::state::category) =		
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::category);

			std::uintptr_t base_ci = lua::luaM_newarray(thread, 192, *reinterpret_cast<std::uint8_t*>( //BASIC_CI_SIZE * sizeof CallInfo
				new_thread + memdata::offsets::lua::gcobj::category)); 

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::base_ci) = base_ci;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::ci) = base_ci;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::end_ci) = base_ci + 168; //(BASIC_CI_SIZE - 1) * sizeof CallInfo
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::size_ci) = 8;
			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::stack) = 
				lua::luaM_newarray(thread, 720, *reinterpret_cast<std::uint8_t*>( //(BASIC_STACK_SIZE + EXTRA_STACK) * sizeof TValue
					new_thread + memdata::offsets::lua::gcobj::category)); 

			state::set_stacksize(new_thread, 45); //BASIC_STACK_SIZE + EXTRA_STACK;

			for (std::size_t i = 0; i < 720; i += 16)
				(reinterpret_cast<temple::lua::tvalue*>(*reinterpret_cast<std::uintptr_t*>(
					new_thread + memdata::offsets::lua::state::stack) + i))->tt = memdata::lua::types::lua_tnil;

			std::uintptr_t stack = *reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::stack);

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::top) = 
				*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::stack);

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::stack_last) = 
				stack + 16 * (state::get_stacksize(new_thread) - 5);

			std::uintptr_t ci = *reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::ci);
			*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::func) = stack;
			(*reinterpret_cast<temple::lua::tvalue**>(new_thread + memdata::offsets::lua::state::top))->tt =
				memdata::lua::types::lua_tnil;

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::top) += 16;
			*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::base) =
				*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::top);

			*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::base) = 
				*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::base);

			*reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top) = 
				*reinterpret_cast<std::uintptr_t*>(new_thread + memdata::offsets::lua::state::top) + 320; // LUA_MINSTACK * sizeof TValue

			*reinterpret_cast<std::uint64_t*>(new_thread + memdata::offsets::lua::state::g_env) =
				*reinterpret_cast<std::uint64_t*>(thread + memdata::offsets::lua::state::g_env);

			*reinterpret_cast<std::uint64_t*>(new_thread + memdata::offsets::lua::state::g_env + sizeof(std::uint64_t)) =
				*reinterpret_cast<std::uint64_t*>(thread + memdata::offsets::lua::state::g_env + sizeof(std::uint64_t));

			*reinterpret_cast<std::uint8_t*>(new_thread + memdata::offsets::lua::state::unk_1) =
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::unk_1);


			auto& top = thread->top;

			top->v.gcobj = new_thread;
			top->tt = memdata::lua::types::lua_tthread;
			top++;

			auto newthread_fn = *reinterpret_cast<void(__cdecl**)(std::uintptr_t, std::uintptr_t)>
				(thread.l_g + memdata::offsets::lua::state::l_g::newthread_fn);
			
			if (newthread_fn)
				newthread_fn(*thread, new_thread);

			return new_thread;
		}

		void pushcclosure(temple::lua::state thread, std::uintptr_t fn, int nups, const char* name, std::uintptr_t continuation)
		{
			auto l_g = thread.l_g;

			if (*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) >=
				*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::gcthreshold))
				functions::lua::luaC_step(thread, 1);

			luaC_wakethread(thread);

			auto ci = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);
			std::uintptr_t env = 0;

			if (ci == *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci))
				env = *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::g_env);
			else
				env = functions::lua::closure::get_env(functions::lua::state::get_closure(thread));

			auto closure = lua::luaM_realloc_(*thread, 16 * nups + 32, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::gcobj::category));

			luaC_link(thread, closure, memdata::lua::types::lua_tfunction);

			functions::lua::closure::set_env(closure, env);
			functions::lua::closure::set_nups(closure, nups);
			functions::lua::closure::set_isc(closure, true);
			functions::lua::closure::set_maxstacksize(closure, 20);

			functions::lua::closure::set_source(closure, fn);
			functions::lua::closure::set_continuation(closure, continuation);
			functions::lua::closure::set_name(*thread, closure, name);

			auto& top = thread->top;

			top = top - nups;

			while (nups--)
			{
				*reinterpret_cast<temple::lua::tvalue*>(closure + memdata::offsets::lua::closure::c_ups + (nups * 16)) =
					*(top + nups);
			}

			top->v.gcobj = closure;
			top->tt = memdata::lua::types::lua_tfunction;
			top++;
		}

		signed int upvalueindex(signed int idx)
		{
			return -10002 - idx;
		}

		void sandboxthread(temple::lua::state thread, std::uint16_t identity)
		{
			std::uintptr_t yo[2] = { 0, 0 };
			std::uint32_t ident[4] = { 0, 0, 0, 0 };

			ident[0] = identity;

			functions::lua::sandboxthread_f(*thread, reinterpret_cast<std::uintptr_t>(ident), reinterpret_cast<std::uintptr_t>(yo));
			
			pushvalue(thread, -10002);
			memory::temple_environments.emplace(index2adr(thread, -1)->v.gcobj);

			pop(thread, 1);
		}

		temple::lua::tvalue* index2adr(temple::lua::state thread, int idx)
		{
			if (idx > 0)
			{
				auto base = thread->base;
				return base + idx - 1;
			}
			else if (idx < 0 && idx > -10000)
			{
				auto top = thread->top;

				return top + idx;
			}
			else
			{
				return index2adr_f(*thread, idx);
			}
		}

		std::uintptr_t luaH_new(temple::lua::state thread, int narray, int nhash)
		{
			std::uintptr_t h = luaM_realloc_(thread, 32, *reinterpret_cast<std::uint8_t*>(
				thread + memdata::offsets::lua::state::category));

			luaC_link(thread, h, memdata::lua::types::lua_ttable);

			memory::ptr_obf::obf(h + memdata::offsets::lua::table::metatable,
				0, memdata::offsets::lua::table::metatable_obf);

			table::set_flags(h, -1);

			memory::ptr_obf::obf(h + memdata::offsets::lua::table::t_array,
				0, memdata::offsets::lua::table::t_array_obf);

			*reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::sizearray) = 0;
			*reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::lastfree) = 0;
			*reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::lsizenode) = 0xFF00;
			*reinterpret_cast<std::uint8_t*>(h + memdata::offsets::lua::table::readonly) = 0;

			memory::ptr_obf::obf(h + memdata::offsets::lua::table::node,
				memdata::vars::lua::dummynode, memdata::offsets::lua::table::node_obf);

			if (narray > 0x4000000)
				luaG_runerrorL(*thread, xorstr("table overflow"));

			if (narray > 0xFFFFFFF)
				luaG_runerrorL(*thread, xorstr("memory allocation error: block too big"));


			std::uintptr_t arr_vector = luaM_reallocvector(*thread, memory::ptr_obf::deobf(h + memdata::offsets::lua::table::t_array,
				memdata::offsets::lua::table::t_array_obf), 16 * *reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::sizearray),
				16 * narray, static_cast<std::uint32_t>(*reinterpret_cast<std::uint8_t*>(h + memdata::offsets::lua::gcobj::category)));

			memory::ptr_obf::obf(h + memdata::offsets::lua::table::t_array,
				arr_vector, memdata::offsets::lua::table::t_array_obf);

			for (int i = *reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::sizearray); i < narray; i++)
			{
				reinterpret_cast<temple::lua::tvalue*>(arr_vector + i * 16)->tt = 0;
			}

			*reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::sizearray) = narray;

			

			int lsize = 0;

			if (!nhash)
			{
				memory::ptr_obf::obf(h + memdata::offsets::lua::table::node,
					memdata::vars::lua::dummynode, memdata::offsets::lua::table::node_obf);
			}
			else
			{
				lsize = temple::util::math::ceillog2(nhash);
				if (lsize > 26)
					luaG_runerrorL(*thread, xorstr("table overflow"));

				nhash = 1 << lsize;

				if (nhash > 0x7FFFFFF)
					luaG_runerrorL(*thread, xorstr("memory allocation error: block too big"));

				std::uintptr_t node_vector = luaM_newarray(thread, 32 * nhash,
					*reinterpret_cast<std::uint8_t*>(h + memdata::offsets::lua::gcobj::category));

				memory::ptr_obf::obf(h + memdata::offsets::lua::table::node,
					node_vector, memdata::offsets::lua::table::node_obf);

				for (int i = 0; i < nhash; i++)
				{
					*reinterpret_cast<std::uintptr_t*>(node_vector + i * 32 + memdata::offsets::lua::node::gval) = 0;
					*reinterpret_cast<std::uintptr_t*>(node_vector + i * 32 + memdata::offsets::lua::node::i_key) = 0;
				}
			}

			*reinterpret_cast<std::uint8_t*>(h + memdata::offsets::lua::table::lsizenode) = lsize;
			*reinterpret_cast<std::uint8_t*>(h + memdata::offsets::lua::table::nodesize) = (1 << lsize) - 1;
			*reinterpret_cast<std::uint32_t*>(h + memdata::offsets::lua::table::lastfree) = nhash;

			return h;
		}

		void createtable(temple::lua::state thread, int narray, int nrec)
		{
			auto l_g = thread.l_g;

			if (*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::totalbytes) >=
				*reinterpret_cast<std::uint32_t*>(l_g + memdata::offsets::lua::state::l_g::gcthreshold))
				luaC_step(thread, 1);

			luaC_wakethread(thread);

			std::uintptr_t h = luaH_new(thread, narray, nrec);

			auto top = thread.top;
			top->v.gcobj = h;
			top->tt = memdata::lua::types::lua_ttable;
			thread.top += 1;
		}

		std::uintptr_t getstack(temple::lua::state thread, int level)
		{
			auto stack_size = (state::get_ci(thread) - state::get_base_ci(thread));

			if (level >= (stack_size / 24))
			{
				return 0;
			}
			else
			{
				return (stack_size - 24 * level) / 24;
			}
		}

		void getinfo(temple::lua::state thread, const char* what, temple::lua::lua_Debug* ar, int level)
		{
			std::uintptr_t fn = 0;
			std::uintptr_t stack_level = 0;

			if (level < 0)
			{
				auto func = thread.top + level;
				fn = func->v.gcobj;
			}
			else if (level < (state::get_ci(thread) - state::get_base_ci(thread)) / 24)
			{
				stack_level = state::get_ci(thread) - level * 24;

				fn = (*reinterpret_cast<temple::lua::tvalue**>(stack_level +
					memdata::offsets::lua::state::callinfo::func))->v.gcobj;
			}

			if (fn)
			{
				auxgetinfo(0, what, ar, fn, stack_level);

				if (std::strchr(what, 'f'))
				{
					auto& top = thread.top;

					top->v.gcobj = fn;
					top->tt = memdata::lua::types::lua_tfunction;
					top += 1;
				}
			}
		}

		std::string getobjecttype(temple::lua::state thread, int idx)
		{
			auto obj = functions::lua::index2adr(thread, idx);

			return std::string(reinterpret_cast<const char*>(
				luaT_objtypename(*thread, obj) + memdata::offsets::lua::string::raw));
		}

		const char* getnamecallmethod(temple::lua::state thread)
		{
			auto namecall = *reinterpret_cast<std::uintptr_t*>(*thread + memdata::offsets::lua::state::namecall);

			if (namecall)
				return reinterpret_cast<const char*>(namecall + memdata::offsets::lua::string::raw);
			else
				return nullptr;
		}

		void setreadonly(temple::lua::state thread, int idx, bool val)
		{
			auto tbl = index2adr(*thread, idx)->v.gcobj;

			*reinterpret_cast<std::uint8_t*>(tbl + memdata::offsets::lua::table::readonly) = val;
		}

		std::uintptr_t touserdata(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(thread, idx);

			auto type = obj->tt;

			if (type == memdata::lua::types::lua_tlightuserdata)
			{
				return obj->v.gcobj;
			}
			else if (type == memdata::lua::types::lua_tuserdata)
			{
				return obj->v.gcobj + memdata::offsets::lua::userdata::ud;
			}

			return 0;
		}

		double tonumber(temple::lua::state thread, int idx)
		{
			auto tbl = index2adr(*thread, idx);

			return xor_double(tbl->v.n);
		}

		std::int32_t tointeger32(temple::lua::state thread, int idx)
		{
			auto tbl = index2adr(*thread, idx);

			return static_cast<std::int32_t>(static_cast<std::uint64_t>(xor_double(tbl->v.n)) & 0xFFFFFFFF);
		}

		std::uint32_t touinteger32(temple::lua::state thread, int idx)
		{
			auto tbl = index2adr(*thread, idx);

			return static_cast<std::uint32_t>(static_cast<std::uint64_t>(xor_double(tbl->v.n)) & 0xFFFFFFFF);
		}

		int toboolean(temple::lua::state thread, int idx)
		{
			auto tbl = index2adr(*thread, idx);

			return tbl->v.b;
		}

		void insert(temple::lua::state thread, int idx)
		{
			auto top = thread->top;
			auto obj = index2adr(thread, idx);

			for (auto i = top; i > obj; i--)
				setobj(i, i - 1);

			setobj(obj, top);
		}

		std::uintptr_t newuserdata(temple::lua::state thread, std::size_t size, char tag)
		{

			if (*reinterpret_cast<std::uint32_t*>(thread.l_g + memdata::offsets::lua::state::l_g::totalbytes) >=
				*reinterpret_cast<std::uint32_t*>(thread.l_g + memdata::offsets::lua::state::l_g::gcthreshold))
				functions::lua::luaC_step(thread, 1);

			luaC_wakethread(thread);
			
			if (size > 0x7FFFFFE7)
				luaG_runerrorL(*thread, xorstr("memory allocation error: block too big"));

			auto ud = luaM_realloc_(thread, size + memdata::offsets::lua::userdata::ud,
				*reinterpret_cast<std::uint8_t*>(thread + memdata::offsets::lua::state::category));

			luaC_link(thread, ud, memdata::lua::types::lua_tuserdata);

			*reinterpret_cast<std::uint32_t*>(ud + memdata::offsets::lua::userdata::size) = size;
			*reinterpret_cast<std::uint8_t*>(ud + memdata::offsets::lua::userdata::tag) = tag;
			
			memory::ptr_obf::obf(ud + memdata::offsets::lua::userdata::metatable, 0, 
				memdata::offsets::lua::userdata::metatable_obf);

			auto& top = thread.top;

			top->v.gcobj = ud;
			top->tt = memdata::lua::types::lua_tuserdata;

			top++;

			return ud + memdata::offsets::lua::userdata::ud;
		}

		void pushlightuserdata(temple::lua::state thread, void* ud)
		{
			auto& top = thread->top;

			top->v.p = ud;
			top->tt = memdata::lua::types::lua_tlightuserdata;
			top++;
		}

		void pushlclosure(temple::lua::state thread, int nups, std::uintptr_t env, std::uintptr_t proto)
		{
			auto& top = thread->top;

			top->v.gcobj = newlclosure(*thread, nups, env, proto);
			top->tt = memdata::lua::types::lua_tfunction;
			top++;
		}

		void pushtable(temple::lua::state thread, std::uintptr_t table)
		{
			auto& top = thread->top;

			top->v.gcobj = table;
			top->tt = memdata::lua::types::lua_ttable;
			top++;
		}

		void pushvalue(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(*thread, idx);
			auto& top = thread->top;

			setobj(top, obj);

			top++;
		}

		void pushinteger(temple::lua::state thread, std::int32_t n)
		{
			auto& top = thread->top;

			top->v.n = xor_double(static_cast<double>(n & 0xFFFFFFFF));
			top->tt = memdata::lua::types::lua_tnumber;
			top++;
		}

		void pushuinteger(temple::lua::state thread, std::uint32_t n)
		{
			auto& top = thread->top;

			top->v.n = xor_double(static_cast<double>(n));
			top->tt = memdata::lua::types::lua_tnumber;
			top++;
		}

		void pushnumber(temple::lua::state thread, double n)
		{
			auto& top = thread->top;

			top->v.n = xor_double(n);
			top->tt = memdata::lua::types::lua_tnumber;
			top++;
		}

		void pushstring(temple::lua::state thread, const char* str)
		{
			auto& top = thread->top;

			if (!str)
				return pushnil(thread);

			top->v.gcobj = newlstr(*thread, str, std::strlen(str));
			top->tt = memdata::lua::types::lua_tstring;
			top++;
		}

		void pushlstring(temple::lua::state thread, const char* str, std::size_t len)
		{
			auto& top = thread->top;

			if (!str)
				return pushnil(thread);

			top->v.gcobj = newlstr(*thread, str, len);
			top->tt = memdata::lua::types::lua_tstring;
			top++;
		}

		void pushboolean(temple::lua::state thread, int b)
		{
			auto& top = thread->top;

			top->v.b = b & 1;
			top->tt = memdata::lua::types::lua_tboolean;
			top++;
		}

		void pushnil(temple::lua::state thread)
		{
			auto& top = thread->top;

			top->tt = memdata::lua::types::lua_tnil;
			top++;
		}

		const char* tostring(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(*thread, idx);

			return reinterpret_cast<const char*>(obj->v.gcobj + memdata::offsets::lua::string::raw);
		}

		int type(temple::lua::state thread, int idx)
		{
			auto obj = index2adr(*thread, idx);

			return obj->tt;
		}

		void setnilvalue(temple::lua::tvalue* obj)
		{
			obj->tt = 0;
		}

		int gettop(temple::lua::state thread)
		{
			return thread->top - thread->base;
		}

		void settop(temple::lua::state thread, int idx)
		{
			auto& top = thread->top;

			if (idx >= 0)
			{
				auto base = thread->base;

				while (top < base + idx)
					setnilvalue(top++);

				top = base + idx;
			}
			else
				top += idx + 1;

		}

		void pop(temple::lua::state thread, int idx)
		{
			settop(*thread, -(idx) - 1);
		}

		namespace state
		{
			temple::lua::tvalue*& get_top(temple::lua::state thread)
			{
				return *reinterpret_cast<temple::lua::tvalue**>(thread + memdata::offsets::lua::state::top);
			}

			temple::lua::tvalue*& get_base(temple::lua::state thread)
			{
				return *reinterpret_cast<temple::lua::tvalue**>(thread + memdata::offsets::lua::state::base);
			}

			temple::lua::extraspace* get_extraspace(temple::lua::state thread)
			{
				return *reinterpret_cast<temple::lua::extraspace**>(thread + memdata::offsets::lua::state::extraspace);
			}

			std::uintptr_t get_closure(temple::lua::state thread)
			{
				return **reinterpret_cast<std::uintptr_t**>((*reinterpret_cast<std::uintptr_t*>(
					thread + memdata::offsets::lua::state::ci)) + memdata::offsets::lua::state::callinfo::func);
			}

			void set_identity(temple::lua::state thread, std::uint32_t identity)
			{
				std::uintptr_t tls = 0;

				__asm
				{
					mov eax, fs:2Ch
					mov tls, eax
				}

				*reinterpret_cast<std::uint32_t*>(*reinterpret_cast<std::uintptr_t*>(tls) + 48) = identity;

				thread->extraspace->identity = identity;
			}

			std::uint32_t get_identity(temple::lua::state thread)
			{
				return thread->extraspace->identity;
			}
			
			std::uintptr_t get_l_g(temple::lua::state thread)
			{
				return memory::ptr_obf::deobf(thread + memdata::offsets::lua::state::l_g::l_g,
					memdata::offsets::lua::state::l_g::l_g_obf);
			}

			void set_l_g(temple::lua::state thread, std::uintptr_t l_g)
			{
				memory::ptr_obf::obf(thread + memdata::offsets::lua::state::l_g::l_g, l_g,
					memdata::offsets::lua::state::l_g::l_g_obf);
			}

			void set_stacksize(temple::lua::state thread, std::size_t sz)
			{
				memory::ptr_obf::obf(thread + memdata::offsets::lua::state::stacksize, sz,
					memdata::offsets::lua::state::stacksize_obf);
			}

			std::uintptr_t get_stacksize(temple::lua::state thread)
			{
				return memory::ptr_obf::deobf(thread + memdata::offsets::lua::state::stacksize,
					memdata::offsets::lua::state::stacksize_obf);
			}

			temple::lua::tvalue* get_stack_last(temple::lua::state thread)
			{
				return *reinterpret_cast<temple::lua::tvalue**>(thread + memdata::offsets::lua::state::stack_last);
			}

			temple::lua::tvalue*& get_stack(temple::lua::state thread)
			{
				return *reinterpret_cast<temple::lua::tvalue**>(thread + memdata::offsets::lua::state::stack);
			}

			void set_stack(temple::lua::state thread, temple::lua::tvalue* stack)
			{
				*reinterpret_cast<temple::lua::tvalue**>(thread + memdata::offsets::lua::state::stack) = stack;
			}

			std::uintptr_t get_genv(temple::lua::state thread)
			{
				return *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::g_env);
			}

			bool is_temple_thread(temple::lua::state thread) //make better method later lol
			{
				auto rbx_identity = state::get_identity(thread);
				auto temple_identity = temple::config::lua::identity;
				if (rbx_identity == temple_identity)
					return true;
				else
					return false;
			}

			std::uintptr_t get_ci(temple::lua::state thread)
			{
				return *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::ci);
			}
			std::uintptr_t get_base_ci(temple::lua::state thread)
			{
				return *reinterpret_cast<std::uintptr_t*>(thread + memdata::offsets::lua::state::base_ci);
			}
		}

		namespace callinfo
		{
			std::uintptr_t get_func(std::uintptr_t ci)
			{
				return *reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::func);
			}
			std::uintptr_t get_base(std::uintptr_t ci)
			{
				return *reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::base);
			}
			std::uintptr_t get_top(std::uintptr_t ci)
			{
				return *reinterpret_cast<std::uintptr_t*>(ci + memdata::offsets::lua::state::callinfo::top);
			}
		}
		
		namespace l_g
		{
			std::uintptr_t get_rootgc(std::uintptr_t l_g)
			{
				//return *reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::rootgc);

				return *reinterpret_cast<std::uintptr_t*>(l_g + memdata::offsets::lua::state::l_g::altrealloc_link);
			}

			std::uintptr_t luaC_white(std::uintptr_t l_g)
			{
				return *reinterpret_cast<std::uint8_t*>(l_g + memdata::offsets::lua::state::l_g::currentwhite) & gcobj::white_bits;
			}
		}

		namespace upvalue
		{
			void unlink_upvalue(std::uintptr_t uv)
			{
				auto next = *reinterpret_cast<std::uintptr_t*>(uv + memdata::offsets::lua::upvalue::next);
				auto prev = *reinterpret_cast<std::uintptr_t*>(uv + memdata::offsets::lua::upvalue::prev);

				*reinterpret_cast<std::uintptr_t*>(next + memdata::offsets::lua::upvalue::prev) = prev;
				*reinterpret_cast<std::uintptr_t*>(prev + memdata::offsets::lua::upvalue::next) = next;
			}
		}

		namespace gcobj
		{
			std::uint8_t white_bits = 3;
			std::uint8_t black_bits = 2;
			std::uint8_t mask_marks = 0xF8;

			std::uintptr_t get_next(std::uintptr_t gcobj)
			{
				//return *reinterpret_cast<std::uintptr_t*>(gcobj + memdata::offsets::lua::gcobj::next);


				return *reinterpret_cast<std::uintptr_t*>((gcobj - memdata::offsets::lua::gcobj::entry::size)
					+ memdata::offsets::lua::gcobj::entry::next_gcobj);
			}

			std::uint8_t get_marked(std::uintptr_t gcobj)
			{
				return *reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked);
			}

			void set_marked(std::uintptr_t gcobj, std::uint8_t marked)
			{
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) = marked;
			}

			std::uint8_t get_tt(std::uintptr_t gcobj)
			{
				return *reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::tt);
			}
			
			bool is_white(std::uintptr_t gcobj)
			{
				return get_marked(gcobj) & white_bits;
			}

			bool is_collectable(std::uintptr_t gcobj)
			{
				return get_tt(gcobj) >= memdata::lua::types::lua_tstring;
			}

			bool is_collectable(temple::lua::tvalue* obj)
			{
				return obj->tt >= memdata::lua::types::lua_tstring;
			}

			bool valis_white(std::uintptr_t gcobj)
			{
				return is_collectable(gcobj) && is_white(gcobj);
			}
			
			bool is_black(std::uintptr_t gcobj)
			{
				return get_marked(gcobj) & 4;
			}

			std::uint8_t other_white(temple::lua::state thread)
			{
				return *reinterpret_cast<std::uint8_t*>(thread.l_g + memdata::offsets::lua::state::l_g::currentwhite) ^ white_bits;
			}

			bool is_dead(temple::lua::state thread, std::uintptr_t gcobj)
			{
				return *reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) & other_white(thread) & white_bits;
			}

			void unhook_gc(temple::lua::state thread, std::uintptr_t gcobj)
			{
				*reinterpret_cast<std::uintptr_t*>(thread.l_g + memdata::offsets::lua::state::l_g::rootgc) = 
					*reinterpret_cast<std::uintptr_t*>(gcobj + memdata::offsets::lua::gcobj::next);
			}

			void mark_white(temple::lua::state thread, std::uintptr_t gcobj)
			{
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) =
					(*reinterpret_cast<std::uint8_t*>(thread.l_g + memdata::offsets::lua::state::l_g::currentwhite)) & 3;
			}

			void changewhite(std::uintptr_t gcobj)
			{
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) ^= 3;
			}

			void reset_bits(std::uintptr_t gcobj, std::uint8_t m)
			{
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) &= ~m;
			}

			void black2gray(std::uintptr_t gcobj)
			{
				reset_bits(gcobj, 1 << black_bits);
			}

			std::uint32_t l_mod(std::uint32_t s, std::size_t size)
			{
				return s & (size - 1);
			}

			bool is_gray(std::uintptr_t gcobj)
			{
				return !is_black(gcobj) && !is_white(gcobj);
			}

			void gray2black(std::uintptr_t gcobj)
			{
				auto marked = *reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked);
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) = marked | (1 << black_bits);
			}

			void make_white(std::uintptr_t l_g, std::uintptr_t gcobj)
			{
				auto marked = *reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked);
				*reinterpret_cast<std::uint8_t*>(gcobj + memdata::offsets::lua::gcobj::marked) =
					(marked & mask_marks) | l_g::luaC_white(l_g);
			}
		}

		namespace userdata
		{
			/*std::uint8_t get_size(std::uintptr_t ud)
			{
				return *reinterpret_cast<std::uint8_t*>(ud + memdata::offsets::lua::userdata::size);
			}*/

			std::uintptr_t get_metatable(std::uintptr_t ud)
			{
				return memory::ptr_obf::deobf(ud + memdata::offsets::lua::userdata::metatable,
					memdata::offsets::lua::userdata::metatable_obf);
			}
		}

		namespace table
		{
			bool isreadonly(std::uintptr_t t)
			{
				return *reinterpret_cast<bool*>(t + memdata::offsets::lua::table::readonly);
			}
			std::size_t size_node(std::uint32_t t)
			{
				return 1 << (*reinterpret_cast<std::uint8_t*>(t + memdata::offsets::lua::table::lsizenode));
			}

			std::uint32_t nodesize(std::uint32_t t)
			{
				return *reinterpret_cast<std::uint32_t*>(t + memdata::offsets::lua::table::nodesize);
			}

			std::uintptr_t hashpow2(std::uintptr_t t, std::uint32_t n)
			{
				std::uintptr_t node = memory::ptr_obf::deobf(t + memdata::offsets::lua::table::node,
					memdata::offsets::lua::table::node_obf);

				return node + 32 * gcobj::l_mod(n, size_node(t));
			}

			std::uintptr_t hashmod(std::uintptr_t t, std::uint32_t n)
			{
				std::uintptr_t node = memory::ptr_obf::deobf(t + memdata::offsets::lua::table::node,
					memdata::offsets::lua::table::node_obf);

				return node + 32 * ((n) % ((size_node(t) - 1) | 1));
			}

			std::uintptr_t hashpointer(std::uintptr_t t, const void* p)
			{
				unsigned int h = unsigned(uintptr_t(p));

				// MurmurHash3 32-bit finalizer
				h ^= h >> 16;
				h *= 0x85ebca6bu;
				h ^= h >> 13;
				h *= 0xc2b2ae35u;
				h ^= h >> 16;

				return hashpow2(t, h);
			}

			std::uintptr_t get_node(std::uintptr_t t)
			{
				return memory::ptr_obf::deobf(t + memdata::offsets::lua::table::node,
					memdata::offsets::lua::table::node_obf);
			}

			std::uintptr_t get_node(std::uintptr_t t, int index)
			{
				std::uintptr_t node = get_node(t);

				return node + 32 * index;
			}

			std::uintptr_t hashstr(std::uintptr_t t, std::uintptr_t str)
			{
				std::uintptr_t str_hash = memory::ptr_obf::deobf(str + memdata::offsets::lua::string::hash,
					memdata::offsets::lua::string::hash_obf);

				std::uintptr_t node = get_node(t);

				return node + 32 * gcobj::l_mod(str_hash, size_node(t));
			}
			std::uintptr_t hashnum(std::uintptr_t t, double n)
			{
				unsigned int i[2];
				memcpy(i, &n, sizeof(i));

				// mask out sign bit to make sure -0 and 0 hash to the same value
				uint32_t h1 = i[0];
				uint32_t h2 = i[1] & 0x7fffffff;

				// finalizer from MurmurHash64B
				const uint32_t m = 0x5bd1e995;

				h1 ^= h2 >> 18;
				h1 *= m;
				h2 ^= h1 >> 22;
				h2 *= m;
				h1 ^= h2 >> 17;
				h1 *= m;
				h2 ^= h1 >> 19;
				h2 *= m;

				// ... truncated to 32-bit output (normally hash is equal to (uint64_t(h1) << 32) | h2, but we only really need the lower 32-bit half)
				return hashpow2(t, h2);

			}

			std::uintptr_t hashvec(std::uintptr_t t, const float* v)
			{
				unsigned int i[3];
				memcpy(i, v, sizeof(i));

				// convert -0 to 0 to make sure they hash to the same value
				i[0] = (i[0] == 0x8000000) ? 0 : i[0];
				i[1] = (i[1] == 0x8000000) ? 0 : i[1];
				i[2] = (i[2] == 0x8000000) ? 0 : i[2];

				// scramble bits to make sure that integer coordinates have entropy in lower bits
				i[0] ^= i[0] >> 17;
				i[1] ^= i[1] >> 17;
				i[2] ^= i[2] >> 17;

				// Optimized Spatial Hashing for Collision Detection of Deformable Objects
				unsigned int h = (i[0] * 73856093) ^ (i[1] * 19349663) ^ (i[2] * 83492791);

#if LUA_VECTOR_SIZE == 4
				i[3] = (i[3] == 0x8000000) ? 0 : i[3];
				i[3] ^= i[3] >> 17;
				h ^= i[3] * 39916801;
#endif

				return hashpow2(t, h);
			}

			std::uintptr_t get_metatable(std::uintptr_t t)
			{
				return memory::ptr_obf::deobf(t + memdata::offsets::lua::table::metatable, memdata::offsets::lua::table::metatable_obf);
			}

			std::uint8_t get_flags(std::uintptr_t t)
			{
				return *reinterpret_cast<std::uint8_t*>(t + memdata::offsets::lua::table::flags);
			}

			void set_flags(std::uintptr_t t, std::uint8_t flags)
			{
				*reinterpret_cast<std::uint8_t*>(t + memdata::offsets::lua::table::flags) = flags;
			}

			std::uint32_t get_sizearray(std::uintptr_t t)
			{
				return *reinterpret_cast<std::uint32_t*>(t + memdata::offsets::lua::table::sizearray);
			}

			std::uintptr_t get_array(std::uintptr_t t)
			{
				return memory::ptr_obf::deobf(t + memdata::offsets::lua::table::t_array,
					memdata::offsets::lua::table::t_array_obf);
			}
			
			temple::lua::tvalue* get_array(std::uintptr_t t, int index)
			{
				return reinterpret_cast<temple::lua::tvalue*>(get_array(t) + 16 * index);
			}
		}

		namespace node
		{
			std::uint8_t ttype(std::uint8_t key)
			{
				return (key & 0xF);
			}
			std::uint8_t gkey(std::uintptr_t n)
			{
				return *reinterpret_cast<std::uint8_t*>(n + memdata::offsets::lua::node::i_key);
			}
			temple::lua::tvalue* gval(std::uintptr_t n)
			{
				return reinterpret_cast<temple::lua::tvalue*>(n);
			}

			temple::lua::tvalue* key2tval(std::uintptr_t n)
			{
				return reinterpret_cast<temple::lua::tvalue*>(n + memdata::offsets::lua::node::value);
			}

			std::uintptr_t gnext(std::uintptr_t n)
			{
				auto key_size = ((*reinterpret_cast<std::uint32_t*>(n + memdata::offsets::lua::node::i_key)) >> 4);

				if (!key_size)
				{
					return 0;
				}

				return 32 * key_size;
			}
		}

		namespace string
		{
			const char* get_buffer(std::uintptr_t tsv)
			{
				return reinterpret_cast<const char*>(tsv + memdata::offsets::lua::string::raw);
			}
			std::size_t get_length(std::uintptr_t tsv)
			{
				return memory::ptr_obf::deobf(tsv + memdata::offsets::lua::string::tsv_len,
					memdata::offsets::lua::string::tsv_len_obf);
			}

			std::size_t get_hash(std::uintptr_t tsv)
			{
				return memory::ptr_obf::deobf(tsv + memdata::offsets::lua::string::hash,
					memdata::offsets::lua::string::hash_obf);;
			}
		}

		namespace closure
		{
			bool islclosure(std::uintptr_t closure)
			{
				return !*reinterpret_cast<bool*>(closure + memdata::offsets::lua::closure::isc);
			}

			bool iscclosure(std::uintptr_t closure)
			{
				return *reinterpret_cast<bool*>(closure + memdata::offsets::lua::closure::isc);
			}

			void set_isc(std::uintptr_t closure, bool b)
			{
				*reinterpret_cast<bool*>(closure + memdata::offsets::lua::closure::isc) = b;
			}

			bool get_isc(std::uintptr_t closure)
			{
				return *reinterpret_cast<bool*>(closure + memdata::offsets::lua::closure::isc);
			}

			void set_maxstacksize(std::uintptr_t closure, std::uint8_t maxstacksize)
			{
				*reinterpret_cast<std::uint8_t*>(closure + memdata::offsets::lua::closure::maxstacksize) = maxstacksize;
			}

			std::uint8_t get_maxstacksize(std::uintptr_t closure)
			{
				return *reinterpret_cast<std::uint8_t*>(closure + memdata::offsets::lua::closure::maxstacksize);
			}

			void setcclosure(std::uintptr_t closure, bool b)
			{
				*reinterpret_cast<bool*>(closure + memdata::offsets::lua::closure::isc) = b;
			}

			std::uintptr_t get_env(std::uintptr_t closure)
			{
				return *reinterpret_cast<std::uintptr_t*>(closure + memdata::offsets::lua::closure::env);
			}

			void set_env(std::uintptr_t closure, std::uintptr_t env)
			{
				*reinterpret_cast<std::uintptr_t*>(closure + memdata::offsets::lua::closure::env) = env;
			}

			std::uintptr_t get_source(std::uintptr_t closure)
			{
				return memory::ptr_obf::deobf(closure + memdata::offsets::lua::closure::source,
					memdata::offsets::lua::closure::source_obf);
			}
			void set_source(std::uintptr_t closure, std::uintptr_t source)
			{
				memory::ptr_obf::obf(closure + memdata::offsets::lua::closure::source,
					source, memdata::offsets::lua::closure::source_obf);
			}

			std::uintptr_t get_proto(std::uintptr_t closure) // Does the same thing as get_source
			{                                                // but different naming for better code readability
				return get_source(closure);
			}

			const char* get_name(std::uintptr_t closure)
			{
				if (iscclosure(closure))
				{
					return reinterpret_cast<const char*>(memory::ptr_obf::deobf(
						closure + memdata::offsets::lua::closure::name, memdata::offsets::lua::closure::name_obf));
				}
				else
				{
					auto proto = get_proto(closure);

					auto ts = memory::ptr_obf::deobf(
						proto + memdata::offsets::lua::proto::name, memdata::offsets::lua::proto::name_obf);

					if (!ts)
						return nullptr;
					else
						return reinterpret_cast<const char*>(ts + memdata::offsets::lua::string::raw);
				}
			}

			void set_name(std::uintptr_t thread, std::uintptr_t closure, const char* name)
			{
				if (iscclosure(closure))
				{
					memory::ptr_obf::obf(closure + memdata::offsets::lua::closure::name,
						reinterpret_cast<std::uintptr_t>(name), memdata::offsets::lua::closure::name_obf);
				}
				else
				{
					auto proto = get_proto(closure);

					memory::ptr_obf::obf(proto + memdata::offsets::lua::proto::name,
						newlstr(thread, name, std::strlen(name)), memdata::offsets::lua::proto::name_obf);
				}
			}

			void set_continuation(std::uintptr_t closure, std::uintptr_t continuation)
			{
				memory::ptr_obf::obf(closure + memdata::offsets::lua::closure::continuation,
					continuation, memdata::offsets::lua::closure::continuation_obf);
			}

			std::uintptr_t get_continuation(std::uintptr_t closure)
			{
				return memory::ptr_obf::deobf(closure + memdata::offsets::lua::closure::continuation,
					memdata::offsets::lua::closure::continuation_obf);
			}

			std::uint8_t get_nups(std::uintptr_t closure)
			{
				return *reinterpret_cast<std::uint8_t*>(closure + memdata::offsets::lua::closure::nups);
			}

			void set_nups(std::uintptr_t closure, std::uint8_t nups)
			{
				*reinterpret_cast<std::uint8_t*>(closure + memdata::offsets::lua::closure::nups) = nups;
			}

			int get_ups(std::uintptr_t closure, int n, temple::lua::tvalue& obj)
			{
				int nups = get_nups(closure);

				if (nups)
				{

					if (n > nups || n < 1)
					{
						obj.tt = memdata::lua::types::lua_tnone;

						return 0;
					}

					if (iscclosure(closure))
					{
						temple::lua::tvalue* upvalues_list = reinterpret_cast<temple::lua::tvalue*>(closure + memdata::offsets::lua::closure::c_ups);

						obj = upvalues_list[n - 1];

						return 1;
					}
					else
					{

						auto upvalues_list = reinterpret_cast<temple::lua::tvalue*>(closure + memdata::offsets::lua::closure::l_ups);
						auto upval = upvalues_list[n - 1];

						if (upval.tt == memdata::lua::types::lua_tupval)
						{
							auto upvalue = reinterpret_cast<temple::lua::upval*>(upval.v.gcobj);
							auto val = upvalue->v;
							if (val)
								obj = *val;
						}
						else
							obj = upval;

						return 1;

					}
				}

				obj.tt = memdata::lua::types::lua_tnone;

				return 0;
			}

			int set_ups(std::uintptr_t closure, int n, temple::lua::tvalue val)
			{
				int nups = get_nups(closure);

				if (nups)
				{
					if (n > nups || n < 1)
						return 0;

					if (iscclosure(closure))
					{
						temple::lua::tvalue* upvalues_list = reinterpret_cast<temple::lua::tvalue*>(closure + memdata::offsets::lua::closure::c_ups);

						upvalues_list[n - 1] = val;
					}
					else
					{
						temple::lua::upval** upvalues_list = reinterpret_cast<temple::lua::upval**>(closure + memdata::offsets::lua::closure::l_ups);
						auto& upval = upvalues_list[(n - 1) * 4]->v;

						upval->v = val.v;
						upval->tt = val.tt;
					}
					return 1;
				}
				else
					return 0;
			}
			
			bool is_temple_closure(temple::lua::state thread, std::uintptr_t closure)
			{
				if (customfn::newcclosurehandler::is_temple_newcclosure(thread, closure))
				{
					return true;
				}
				else if (checks::callcheck::is_temple_cclosure(closure))
				{
					return true;
				}
				else if (functions::lua::closure::islclosure(closure) && memory::temple_environments.find(closure::get_env(closure)) != memory::temple_environments.end())
				{
					//std::cout << "penis balls\n";
					return true;
				}
				else
				{
					return false;
				}
			}
		}

		namespace proto
		{
			int get_sizek(std::uintptr_t proto)
			{
				return *reinterpret_cast<int*>(proto + memdata::offsets::lua::proto::sizek);
			}

			int get_sizep(std::uintptr_t proto)
			{
				return *reinterpret_cast<int*>(proto + memdata::offsets::lua::proto::sizep);
			}

			std::uint8_t get_maxstacksize(std::uintptr_t proto)
			{
				return *reinterpret_cast<std::uint8_t*>(proto + memdata::offsets::lua::proto::maxstacksize);
			}

			std::uint8_t get_nups(std::uintptr_t proto)
			{
				return *reinterpret_cast<std::uint8_t*>(proto + memdata::offsets::lua::proto::nups);
			}

			std::size_t get_sizecode(std::uintptr_t proto)
			{
				return *reinterpret_cast<std::size_t*>(proto + memdata::offsets::lua::proto::sizecode);
			}

			void set_code(std::uintptr_t proto, std::size_t n, std::uint32_t i)
			{
				auto code = memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::k,
					memdata::offsets::lua::proto::proto_obf);
				auto* code_list = reinterpret_cast<std::uint32_t*>(code);

				*(code_list + n) = i;
			}

			int get_k(std::uintptr_t proto, int n, temple::lua::tvalue& result)
			{
				int sizek = get_sizek(proto);

				if (sizek)
				{

					if (n > sizek || n < 1)
					{
						result.tt = memdata::lua::types::lua_tnone;

						return 0;
					}

					std::uintptr_t k_list = memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::k,
						memdata::offsets::lua::proto::proto_obf);
					temple::lua::tvalue* constant_list = reinterpret_cast<temple::lua::tvalue*>(k_list);

					result = constant_list[n - 1];

					return 1;
				}
				else
				{
					result.tt = memdata::lua::types::lua_tnone;

					return 0;
				}
			}

			int set_k(std::uintptr_t proto, int n, temple::lua::tvalue obj)
			{
				int sizek = get_sizek(proto);

				if (sizek)
				{
					if (n > sizek || n < 1)
						return 0;

					std::uintptr_t k_list = memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::k,
						memdata::offsets::lua::proto::proto_obf);
					temple::lua::tvalue* constant_list = reinterpret_cast<temple::lua::tvalue*>(k_list);

					auto& constant = constant_list[n - 1];

					constant.tt = obj.tt;
					constant.v = obj.v;

					return 1;
				}
				else
					return 0;
			}

			int get_p(std::uintptr_t thread, std::uintptr_t proto, int n, temple::lua::tvalue& result)
			{
				int sizep = get_sizep(proto);

				if (sizep)
				{
					if (n > sizep || n < 1)
						return 0;

					std::uintptr_t p_list = memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::p,
						memdata::offsets::lua::proto::proto_obf);
					std::uintptr_t* proto_list = reinterpret_cast<std::uintptr_t*>(p_list);

					auto proto = proto_list[n - 1];

					result.tt = memdata::lua::types::lua_tfunction;
					result.v.gcobj = newlclosure(thread, get_nups(proto), index2adr(thread, -10002)->v.gcobj, proto);

					return 1;
				}
				else
					return 0;
			}

			int set_p(std::uintptr_t proto, int n, std::uintptr_t newproto)
			{
				int sizep = get_sizep(proto);

				if (sizep)
				{
					if (n > sizep || n < 1)
						return 0;

					std::uintptr_t p_list = memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::p,
						memdata::offsets::lua::proto::proto_obf);
					std::uintptr_t* proto_list = reinterpret_cast<std::uintptr_t*>(p_list);

					auto& proto = proto_list[n - 1];

					proto = newproto;

					return 1;
				}
				else
					return 0;
			}

			const char* get_source(std::uintptr_t proto)
			{
				return reinterpret_cast<const char*>(memory::ptr_obf::deobf(proto + memdata::offsets::lua::proto::source,
					memdata::offsets::lua::proto::source_obf) + memdata::offsets::lua::string::raw);
			}
		}
	}

	namespace reflection
	{
		temple_fn<void, std::uintptr_t, const std::shared_ptr<std::uintptr_t>&> instancebridge_push{ true };
	}

	namespace fflag
	{
		temple_fn<int, const std::string&, const std::string&, int, char> set_fflag_f;

		int set_fflag(const std::string& flag, const std::string& value)
		{
			return functions::fflag::set_fflag_f(flag, value, 63, 0);
		}
	}

	namespace instance
	{
		temple_fn<unsigned int> get_property_descriptor;

		temple_fn<void, std::uintptr_t, std::uintptr_t> get_instance_property{ true };
		temple_fn<void, std::uintptr_t, std::uintptr_t> set_instance_property;
		temple_fn<void, std::uintptr_t, float, std::uintptr_t> fireclickdetector;
		temple_fn<void, std::uintptr_t> fireproximityprompt;
		temple_fn<void> getnetworkowner;

		std::string get_name(rbx::instance* instance)
		{
			return instance->name;
		}

		std::string get_classname(rbx::instance* instance)
		{
			auto fn = *reinterpret_cast<std::string*(__thiscall**)(std::uintptr_t)>(*reinterpret_cast<std::uintptr_t*>(instance) + 20);

			return *fn(reinterpret_cast<std::uintptr_t>(instance));
		}

		rbx::instance* find_first_child_by_class(rbx::instance* instance, const std::string& classname)
		{
			const auto children = instance->children.get();
			for (const auto& child : *children)
				if (get_classname(child.get()) == classname)
					return child.get();

			return nullptr;
		}

		rbx::instance* find_first_child(rbx::instance* instance, const std::string& name)
		{
			const auto children = instance->children.get();
			for (const auto& child : *children)
				if (child->name == name)
					return child.get();

			return nullptr;
		}
	}

	namespace game
	{
		bool is_loaded(std::uintptr_t game)
		{
			return *reinterpret_cast<bool*>(game + memdata::offsets::game::is_loaded);
		}
	}

	void ready_functions()
	{
		VM_START;

		lua::sandboxthread_f.load_data(memdata::functions::lua::sandboxthread);
		lua::index2adr_f.load_data(memdata::functions::lua::index2adr);
		lua::spawn.load_data(memdata::functions::lua::spawn);
		lua::load.load_data(memdata::functions::lua::load);

		lua::luaC_barrierf.load_data(memdata::functions::lua::luaC_barrierf);
		lua::luaC_barriert.load_data(memdata::functions::lua::luaC_barriert);
		lua::fn_luaG_runerrorL.load_data(memdata::functions::lua::luaG_runerrorL);
		lua::luaH_setnum.load_data(memdata::functions::lua::luaH_setnum);
		lua::f_call.load_data(memdata::functions::lua::f_call);
		lua::luaH_set.load_data(memdata::functions::lua::luaH_set);
		lua::altfrealloc.load_data(memdata::functions::lua::altfrealloc);
		lua::newblock_func.load_data(memdata::functions::lua::newblock_func);
		lua::altnewblock_func.load_data(memdata::functions::lua::altnewblock_func);
		lua::free_func.load_data(memdata::functions::lua::free_func);
		lua::luaG_indexerror.load_data(memdata::functions::lua::luaG_indexerror);
		lua::luaT_objtypename.load_data(memdata::functions::lua::luaT_objtypename);
		lua::luaD_rawrunprotected.load_data(memdata::functions::lua::luaD_rawrunprotected);
		lua::luaD_seterrorobj.load_data(memdata::functions::lua::luaD_seterrorobj);
		lua::callTM.load_data(memdata::functions::lua::callTM);
		lua::callTMres.load_data(memdata::functions::lua::callTMres);
		lua::auxgetinfo.load_data(memdata::functions::lua::auxgetinfo);

		lua::fn_luaC_step.load_data(memdata::functions::lua::luaC_step);

		reflection::instancebridge_push.load_data(memdata::functions::reflection::instancebridge_push);

		fflag::set_fflag_f.load_data(memdata::functions::fflag::set_fflag);

		instance::getnetworkowner.load_data(memdata::functions::instance::getnetworkowner);

		instance::fireclickdetector.load_data(memdata::functions::instance::fireclickdetector);
		instance::fireproximityprompt.load_data(memdata::functions::instance::fireproximityprompt);
		instance::get_property_descriptor.load_data(memdata::functions::instance::get_property_descriptor);

		instance::get_instance_property.load_data(memdata::functions::instance::get_instance_property);
		instance::set_instance_property.load_data(memdata::functions::instance::set_instance_property);

		VM_END;;
	}
}