#pragma once
#include "functionhandler/functionhandler.h"
#include <rbx/structs/instance.h>
#include <rbx/structs/lua.h>

#include <cstdint>
#include <cstddef>
#include <string>
#include <memory>

namespace temple::functions
{
	namespace lua
	{
		extern temple_fn<void, std::uintptr_t, const temple::lua::tvalue*, const temple::lua::tvalue*> luaG_indexerror;
		extern temple_fn <std::uintptr_t, std::uintptr_t, const temple::lua::tvalue*> luaT_objtypename;
		extern temple_fn<int, std::uintptr_t, const char*> load;
		extern temple_fn<int, std::uintptr_t> spawn;
		extern temple_fn<int, std::uintptr_t> require;

		void luaG_runerrorL(std::uintptr_t thread, const char* msg);
		void luaC_wakethread(temple::lua::state thread);

		void getinfo(temple::lua::state, const char*, temple::lua::lua_Debug*, int);
		std::uintptr_t newthread(temple::lua::state);
		std::uintptr_t newlstr(temple::lua::state, const char*, std::size_t);

		void createtable(temple::lua::state, int, int);

		void sandboxthread(temple::lua::state, std::uint16_t);

		void rawgetfield(temple::lua::state, int, const char*);

		void getfield(temple::lua::state, int, const char*);
		void setfield(temple::lua::state, int, const char*);

		void gettable(temple::lua::state, int);

		void rawseti(temple::lua::state, int, int);
		int rawgeti(temple::lua::state, int, int);
		void rawset(temple::lua::state, int);
		void rawget(temple::lua::state, int);

		void setmetatable(temple::lua::state, int);
		int getmetatable(temple::lua::state, int);

		int pcall(temple::lua::state, int, int, int);
		int lua_call(temple::lua::state, int, int);

		int yield(temple::lua::state, int);

		std::uintptr_t newlclosure(temple::lua::state, int, std::uintptr_t, std::uintptr_t);

		const char* getnamecallmethod(temple::lua::state);
		void sandboxthread(temple::lua::state, std::uint16_t);

		std::uintptr_t getstack(temple::lua::state, int);
		temple::lua::tvalue* index2adr(temple::lua::state, int);
		void setreadonly(temple::lua::state, int, bool);

		std::string getobjecttype(temple::lua::state, int);

		std::uintptr_t touserdata(temple::lua::state, int);
		double tonumber(temple::lua::state, int);
		std::int32_t tointeger32(temple::lua::state, int);
		std::uint32_t touinteger32(temple::lua::state, int);
		int toboolean(temple::lua::state, int);

		int next(temple::lua::state, int);
		void insert(temple::lua::state, int);

		std::uintptr_t newuserdata(temple::lua::state, std::size_t, char);
		void pushlightuserdata(temple::lua::state, void*);
		void pushlclosure(temple::lua::state, int, std::uintptr_t, std::uintptr_t);
		void pushcclosure(temple::lua::state thread, std::uintptr_t fn, int nups, const char* name, std::uintptr_t continuation);
		signed int upvalueindex(signed int);
		void pushtable(temple::lua::state, std::uintptr_t);
		void pushvalue(temple::lua::state, int);
		void pushboolean(temple::lua::state, int);
		void pushlstring(temple::lua::state, const char*, std::size_t);
		void pushstring(temple::lua::state, const char*);
		void pushinteger(temple::lua::state, std::int32_t);
		void pushuinteger(temple::lua::state thread, std::uint32_t n);
		void pushnumber(temple::lua::state, double);
		void pushnil(temple::lua::state);

		const char* tostring(temple::lua::state, int);
		int type(temple::lua::state, int);

		int  gettop(temple::lua::state);
		void settop(temple::lua::state, int);
		void pop   (temple::lua::state, int);

		namespace state
		{
			temple::lua::tvalue*& get_top(temple::lua::state);
			temple::lua::tvalue*& get_base(temple::lua::state);

			temple::lua::extraspace* get_extraspace(temple::lua::state);
			std::uintptr_t get_closure(temple::lua::state);

			void set_identity(temple::lua::state, std::uint32_t);
			std::uint32_t get_identity(temple::lua::state);
			std::uintptr_t get_l_g(temple::lua::state);
			void set_l_g(temple::lua::state, std::uintptr_t);

			std::uintptr_t get_stacksize(temple::lua::state);
			void set_stacksize(temple::lua::state, std::size_t);

			temple::lua::tvalue* get_stack_last(temple::lua::state);
			temple::lua::tvalue*& get_stack(temple::lua::state);
			bool is_temple_thread(temple::lua::state);
			std::uintptr_t get_ci(temple::lua::state);
			std::uintptr_t get_base_ci(temple::lua::state);
		}

		namespace callinfo
		{
			std::uintptr_t get_func(std::uintptr_t);
			std::uintptr_t get_base(std::uintptr_t);
			std::uintptr_t get_top(std::uintptr_t);
		}

		namespace userdata
		{
			std::uint8_t get_size(std::uintptr_t);

			std::uintptr_t get_metatable(std::uintptr_t);
		}

		namespace table
		{
			bool isreadonly(std::uintptr_t);
			std::uintptr_t get_node(std::uintptr_t);
			std::uintptr_t get_node(std::uintptr_t, int);
			std::size_t size_node(std::uint32_t);
			std::uint32_t nodesize(std::uint32_t);
			std::uintptr_t hashpow2(std::uintptr_t, std::uint32_t);
			std::uintptr_t hashpointer(std::uintptr_t, const void*);
			std::uintptr_t hashvec(std::uintptr_t t, const float*);
			std::uintptr_t hashmod(std::uintptr_t, std::uint32_t);

			std::uintptr_t hashstr(std::uintptr_t, std::uintptr_t);

			std::uintptr_t hashnum(std::uintptr_t, double);

			std::uintptr_t get_metatable(std::uintptr_t);

			std::uint8_t get_flags(std::uintptr_t);

			void set_flags(std::uintptr_t, std::uint8_t);

			std::uint32_t get_sizearray(std::uintptr_t);
			std::uintptr_t get_array(std::uintptr_t);
			temple::lua::tvalue* get_array(std::uintptr_t, int);
		}

		namespace node
		{
			//std::uint8_t ttype(std::uint8_t);
			std::uint8_t gkey(std::uintptr_t);
			temple::lua::tvalue* gval(std::uintptr_t);
			temple::lua::tvalue* key2tval(std::uintptr_t);
			std::uintptr_t gnext(std::uintptr_t);
		}

		namespace string
		{
			const char* get_buffer(std::uintptr_t);
			std::size_t get_length(std::uintptr_t);
			std::size_t get_hash(std::uintptr_t);
		}

		namespace l_g
		{
			std::uintptr_t get_rootgc(std::uintptr_t);
		}

		namespace upvalue
		{
			void unlink_upvalue(std::uintptr_t);
		}

		namespace gcobj
		{
			extern std::uint8_t white_bits;
			extern std::uint8_t black_bits;
			extern std::uint8_t mask_marks;

			std::uintptr_t get_next(std::uintptr_t);
			std::uint8_t get_marked(std::uintptr_t);
			std::uint8_t get_tt(std::uintptr_t);

			bool is_white(std::uintptr_t);
			bool is_collectable(std::uintptr_t);
			bool is_collectable(temple::lua::tvalue*);
			bool valis_white(std::uintptr_t);
			bool is_black(std::uintptr_t);
			void black2gray(std::uintptr_t);

			std::uint8_t other_white(temple::lua::state);
			bool is_dead(temple::lua::state, std::uintptr_t);
			void changewhite(std::uintptr_t);
			std::uint32_t l_mod(std::uint32_t, std::size_t);

			bool is_gray(std::uintptr_t);
			void gray2black(std::uintptr_t);
			void make_white(std::uintptr_t, std::uintptr_t);

			void unhook_gc(temple::lua::state, std::uintptr_t);
			void mark_white(temple::lua::state, std::uintptr_t);
		}

		namespace closure
		{
			bool islclosure(std::uintptr_t);
			bool iscclosure(std::uintptr_t);
			void setcclosure(std::uintptr_t, bool);

			std::uintptr_t get_env(std::uintptr_t);
			void set_env(std::uintptr_t, std::uintptr_t);

			std::uintptr_t get_source(std::uintptr_t);
			void set_source(std::uintptr_t, std::uintptr_t);
			std::uintptr_t get_proto(std::uintptr_t);

			void set_continuation(std::uintptr_t, std::uintptr_t);
			std::uintptr_t get_continuation(std::uintptr_t);

			std::uint8_t get_nups(std::uintptr_t);
			void set_nups(std::uintptr_t, std::uint8_t);

			void set_isc(std::uintptr_t, bool);
			void set_maxstacksize(std::uintptr_t, std::uint8_t);
			std::uint8_t get_maxstacksize(std::uintptr_t);

			const char* get_name(std::uintptr_t);
			void set_name(std::uintptr_t, std::uintptr_t, const char*);

			int get_ups(std::uintptr_t, int, temple::lua::tvalue&);
			int set_ups(std::uintptr_t, int, temple::lua::tvalue);

			bool is_temple_closure(temple::lua::state thread, std::uintptr_t);
		}

		namespace proto
		{
			int get_sizek(std::uintptr_t);
			int get_sizep(std::uintptr_t);
			std::uint8_t get_nups(std::uintptr_t);
			std::uint8_t get_maxstacksize(std::uintptr_t);

			int get_k(std::uintptr_t, int, temple::lua::tvalue&);
			int set_k(std::uintptr_t, int, temple::lua::tvalue);

			int get_p(std::uintptr_t, std::uintptr_t, int, temple::lua::tvalue&);
			int set_p(std::uintptr_t, int, std::uintptr_t);

			const char* get_source(std::uintptr_t);

			std::size_t get_sizecode(std::uintptr_t);
			void set_code(std::uintptr_t, std::size_t, std::uint32_t);
		}
	}

	namespace reflection
	{
		extern temple_fn<void, std::uintptr_t, const std::shared_ptr<std::uintptr_t>&> instancebridge_push;
	}

	namespace fflag
	{
		int set_fflag(const std::string&, const std::string&);
	}

	namespace instance
	{
		extern temple_fn<unsigned int> get_property_descriptor;

		extern temple_fn<void, std::uintptr_t, std::uintptr_t> get_instance_property;
		extern temple_fn<void, std::uintptr_t, std::uintptr_t> set_instance_property;
		extern temple_fn<void, std::uintptr_t, float, std::uintptr_t> fireclickdetector;
		extern temple_fn<void, std::uintptr_t> fireproximityprompt;
		extern temple_fn<void> getnetworkowner;

		std::string get_classname(rbx::instance*);
		std::string get_name(rbx::instance*);
		rbx::instance* find_first_child_by_class(rbx::instance*, const std::string&);
		rbx::instance* find_first_child(rbx::instance*, const std::string&);
	}

	namespace game
	{
		bool is_loaded(std::uintptr_t);
	}

	void ready_functions();
}

