#pragma once

#include <rbx/structs/instance.h>
#include <rbx/structs/lua.h>
#include "../memory.h"

#include <cstdint>
#include <cstddef>

namespace temple::memdata
{
	using memory::ptr_obf::obfuscation_type;

	extern std::uintptr_t base;

	namespace classes
	{
		extern std::uintptr_t game_pointer;
		extern std::uintptr_t gamebasicsettings_pointer;

		extern std::uintptr_t gamebasicsettings;
		extern std::uintptr_t game;
		extern rbx::instance* datamodel;
		extern rbx::instance* script_context;

		namespace lua
		{
			extern temple::lua::state rstate;
			extern temple::lua::state state;
		}
	}

	namespace functions
	{
		enum class calling_convention
		{
			cc_unk,
			cc_cdecl,
			cc_stdcall,
			cc_fastcall,
			cc_thiscall,
			cc_rbxcall
		};

		struct func_data
		{
			std::uintptr_t ptr = 0;
			calling_convention cc = calling_convention::cc_unk;

			func_data() {}
			func_data(std::uintptr_t p, calling_convention c) : ptr(p), cc(c) {}
		};

		namespace lua
		{
			extern func_data load;
			extern func_data spawn;
			extern func_data index2adr;
			extern func_data luaE_newthread;
			extern func_data luaC_step;
			extern func_data sandboxthread;
			extern func_data f_call;
			extern func_data callerrfunc;
			extern func_data luaG_runerrorL;
			extern func_data luaC_barrierf;
			extern func_data luaC_barriert;
			extern func_data luaH_set;
			extern func_data luaG_indexerror;
			extern func_data luaT_objtypename;
			extern func_data luaD_rawrunprotected;
			extern func_data luaH_setnum;
			extern func_data auxgetinfo;
			extern func_data altfrealloc;
			extern func_data altnewblock_func;
			extern func_data newblock_func;
			extern func_data luaD_seterrorobj;
			extern func_data free_func;
			extern func_data callTMres;
			extern func_data callTM;
			extern func_data index;
			extern func_data namecall;
			extern func_data require;
		}

		namespace reflection
		{
			extern func_data instancebridge_push;
		}

		namespace fflag
		{
			extern func_data set_fflag;
		}

		namespace instance
		{
			extern func_data getnetworkowner;

			extern func_data fireclickdetector;
			extern func_data fireproximityprompt;

			extern func_data get_property_descriptor;

			extern func_data get_instance_property;
			extern func_data set_instance_property;
		}
	}

	namespace lua::types
	{
		extern std::int32_t lua_tnone,
			lua_tnil,
			lua_tboolean,
			lua_tlightuserdata,
			lua_tvector,
			lua_tnumber,
			lua_tstring,
			lua_tfunction,
			lua_tthread,
			lua_ttable,
			lua_tuserdata,
			lua_tupval;
	}

	namespace vars
	{
		namespace lua
		{
			extern std::uintptr_t xor_constant;
			extern std::uintptr_t hash_type;
			extern std::uintptr_t nilobject;
			extern std::uintptr_t dummynode;
			extern std::uintptr_t realloc_table;
			extern std::uintptr_t connection_tag;
			extern std::uintptr_t reallocation_alt;
		}

		namespace instance
		{
			extern std::uintptr_t property_table;
		}

		namespace rbx
		{
			extern std::uintptr_t task_scheduler;
		}
	}

	namespace offsets
	{
		namespace lua
		{
			namespace state
			{
				extern std::ptrdiff_t state;

				extern obfuscation_type state_obf;

				extern std::ptrdiff_t status;
				extern std::ptrdiff_t prealloc_mark;
				extern std::ptrdiff_t nccalls;
				extern std::ptrdiff_t baseccalls;

				extern std::ptrdiff_t top;
				extern std::ptrdiff_t base;
				extern std::ptrdiff_t g_env;

				extern std::ptrdiff_t namecall;
				extern std::ptrdiff_t stack;
				extern std::ptrdiff_t stacksize;
				extern std::ptrdiff_t stack_last;

				extern obfuscation_type stacksize_obf;

				extern std::ptrdiff_t extraspace;

				extern std::ptrdiff_t ci;
				extern std::ptrdiff_t base_ci;
				extern std::ptrdiff_t end_ci;
				extern std::ptrdiff_t size_ci;

				extern std::ptrdiff_t category;

				extern std::ptrdiff_t openupval;
				extern std::ptrdiff_t lcni;

				extern std::ptrdiff_t gclist;

				extern std::ptrdiff_t unk_0;
				extern std::ptrdiff_t unk_1;

				namespace callinfo
				{
					extern std::ptrdiff_t base;
					extern std::ptrdiff_t top;
					extern std::ptrdiff_t func;
				}

				namespace l_g
				{
					extern std::ptrdiff_t l_g;
					extern obfuscation_type l_g_obf;

					extern std::ptrdiff_t strt_hash;
					extern std::ptrdiff_t nuse;
					extern std::ptrdiff_t strt_size;

					extern std::ptrdiff_t grayagain;
					extern std::ptrdiff_t currentwhite;
					extern std::ptrdiff_t totalbytes;
					extern std::ptrdiff_t gcthreshold;
					extern std::ptrdiff_t gcstate;

					extern std::ptrdiff_t newthread_fn;
					extern std::ptrdiff_t newstring_fn;
					extern std::ptrdiff_t pcall_fn;

					extern std::ptrdiff_t rootgc;

					extern std::ptrdiff_t frealloc;
					extern std::ptrdiff_t ud;

					extern std::ptrdiff_t alloc_category;
					extern std::ptrdiff_t altrealloc_link;

					extern std::ptrdiff_t metatable;

					extern std::ptrdiff_t tmname;
					extern obfuscation_type tmname_obf;

					extern std::ptrdiff_t tmname_index;
					extern std::ptrdiff_t tmname_newindex;

					extern std::ptrdiff_t index_tm;
					extern std::ptrdiff_t newindex_tm;
				}
			}

			namespace upvalue
			{
				extern std::ptrdiff_t v;
				extern std::ptrdiff_t u_value;
				extern std::ptrdiff_t next;
				extern std::ptrdiff_t prev;
				extern std::ptrdiff_t n_next;
				extern std::ptrdiff_t n_prev;
			}

			namespace gcobj
			{
				extern std::ptrdiff_t next;
				extern std::ptrdiff_t marked;
				extern std::ptrdiff_t tt;
				extern std::ptrdiff_t category;

				namespace entry
				{
					extern std::ptrdiff_t size;
					extern std::ptrdiff_t next_gcobj;
				}
			}

			namespace closure
			{
				extern std::ptrdiff_t isc;
				extern std::ptrdiff_t nups;
				extern std::ptrdiff_t maxstacksize;
				extern std::ptrdiff_t env;
				extern std::ptrdiff_t source;

				extern std::ptrdiff_t l_ups;
				extern std::ptrdiff_t c_ups;

				extern std::ptrdiff_t name;
				extern std::ptrdiff_t continuation;

				extern obfuscation_type name_obf;
				extern obfuscation_type continuation_obf;
				extern obfuscation_type source_obf;
			}

			namespace proto
			{
				extern std::ptrdiff_t sizep;
				extern std::ptrdiff_t sizek;
				extern std::ptrdiff_t nups;
				extern std::ptrdiff_t maxstacksize;
				extern std::ptrdiff_t sizecode;

				extern std::ptrdiff_t code;
				extern std::ptrdiff_t k;
				extern std::ptrdiff_t p;

				extern std::ptrdiff_t source;
				extern obfuscation_type source_obf;

				extern std::ptrdiff_t name;
				extern obfuscation_type name_obf;

				extern obfuscation_type proto_obf;
			}

			namespace table
			{
				extern std::ptrdiff_t readonly;
				extern std::ptrdiff_t flags;
				extern std::ptrdiff_t lsizenode;
				extern std::ptrdiff_t lastfree;
				extern std::ptrdiff_t nodesize;
				extern std::ptrdiff_t sizearray;
				extern std::ptrdiff_t gclist;

				extern std::ptrdiff_t node;
				extern obfuscation_type node_obf;

				extern std::ptrdiff_t t_array;
				extern obfuscation_type t_array_obf;

				extern std::ptrdiff_t metatable;
				extern obfuscation_type metatable_obf;
			}

			namespace node
			{
				extern std::ptrdiff_t i_key;
				extern std::ptrdiff_t value;
				extern std::ptrdiff_t gval;
			}

			namespace userdata
			{
				extern std::ptrdiff_t size;
				extern std::ptrdiff_t tag;
				extern std::ptrdiff_t ud;

				extern std::ptrdiff_t metatable;
				extern obfuscation_type metatable_obf;
			}

			namespace string
			{
				extern std::ptrdiff_t raw;

				extern std::ptrdiff_t str_id;

				extern std::ptrdiff_t hash;
				extern obfuscation_type hash_obf;

				extern std::ptrdiff_t tsv_len;
				extern obfuscation_type tsv_len_obf;
			}
		}

		namespace weakthreadref
		{
			extern std::ptrdiff_t thread_node;
			extern std::ptrdiff_t node_first;
			extern std::ptrdiff_t node_next;
			extern std::ptrdiff_t state_off1;
			extern std::ptrdiff_t state_off2;
		}

		namespace instance
		{
			extern std::ptrdiff_t start_of_children;
			extern std::ptrdiff_t name;

			namespace property_descriptor
			{
				extern std::ptrdiff_t scriptable;
			}

			namespace event_instance
			{
				extern std::ptrdiff_t source;
				extern std::ptrdiff_t first_node;
			}

			namespace signal_connection
			{
				extern std::ptrdiff_t next;
			}

			namespace localscript
			{
				extern std::ptrdiff_t protected_string;
			}

			namespace modulescript
			{
				extern std::ptrdiff_t protected_string;
			}
		}

		namespace task_scheduler
		{
			extern std::ptrdiff_t frame_delay;
		}

		namespace game
		{
			extern std::ptrdiff_t is_loaded;
			extern std::ptrdiff_t non_networked;
		}
	}

	namespace memloc
	{
		extern std::uintptr_t ret_location;
		extern std::uintptr_t createevent_trigger;
		extern std::uintptr_t trustcheck_return;
	}

	void base_addresses();
}