#include "memory.h"

#include <themida/ThemidaSDK.h>
#include "functions/functions.h"
#include <temple/config/config.h>
#include "memdata/memdata.h"
#include <xorstr.h>

#include <unordered_map>
#include <iostream>
#include <thread>
#include <chrono>

namespace temple::memory
{
	namespace ptr_obf
	{

		std::uintptr_t xor_deobfuscate(std::uintptr_t ptr)
		{
			return *reinterpret_cast<std::uintptr_t*>(ptr) ^ ptr;
		}

		std::uintptr_t add_deobfuscate(std::uintptr_t ptr)
		{
			return *reinterpret_cast<std::uintptr_t*>(ptr) - ptr;
		}

		std::uintptr_t sub_deobfuscate(std::uintptr_t ptr)
		{
			return *reinterpret_cast<std::uintptr_t*>(ptr) + ptr;
		}

		std::uintptr_t inverse_sub_deobfuscate(std::uintptr_t ptr)
		{
			return ptr - *reinterpret_cast<std::uintptr_t*>(ptr);
		}

		std::uintptr_t deobf(std::uintptr_t ptr, obfuscation_type type)
		{
			switch (type)
			{
			case obfuscation_type::exor:
			{
				return xor_deobfuscate(ptr);
			}
			case obfuscation_type::add:
			{
				return add_deobfuscate(ptr);
			}
			case obfuscation_type::sub:
			{
				return sub_deobfuscate(ptr);
			}
			case obfuscation_type::inverse_sub:
			{
				return inverse_sub_deobfuscate(ptr);
			}
			}

			return 0;
		}

		void xor_obfuscate(std::uintptr_t ptr, std::uintptr_t value)
		{
			*reinterpret_cast<std::uintptr_t*>(ptr) = ptr ^ value;
		}

		void add_obfuscate(std::uintptr_t ptr, std::uintptr_t value)
		{
			*reinterpret_cast<std::uintptr_t*>(ptr) = value + ptr;
		}

		void sub_obfuscate(std::uintptr_t ptr, std::uintptr_t value)
		{
			*reinterpret_cast<std::uintptr_t*>(ptr) = value - ptr;
		}

		void inverse_sub_obfuscate(std::uintptr_t ptr, std::uintptr_t value)
		{
			*reinterpret_cast<std::uintptr_t*>(ptr) = ptr - value;
		}

		void obf(std::uintptr_t ptr, std::uintptr_t value, obfuscation_type type)
		{
			switch (type)
			{
			case obfuscation_type::exor:
			{
				xor_obfuscate(ptr, value);
				break;
			}
			case obfuscation_type::add:
			{
				add_obfuscate(ptr, value);
				break;
			}
			case obfuscation_type::sub:
			{
				sub_obfuscate(ptr, value);
				break;
			}
			case obfuscation_type::inverse_sub:
			{
				inverse_sub_obfuscate(ptr, value);
				break;
			}
			}
		}
	}

	std::unordered_set<std::uintptr_t> temple_environments;

	std::uintptr_t get_game()
	{
		return *reinterpret_cast<std::uintptr_t*>(memdata::classes::game_pointer);
	}

	std::uintptr_t get_gamebasicsettings()
	{
		return *reinterpret_cast<std::uintptr_t*>(memdata::classes::gamebasicsettings_pointer);
	}

	std::uintptr_t get_global_function(std::uintptr_t thread, const char* name)
	{
		functions::lua::getfield(thread, -10002, name);

		auto result = functions::lua::closure::get_source(functions::lua::index2adr(
			thread, -1)->v.gcobj);

		functions::lua::pop(thread, 1);

		return result;
	}

	void prepare_memory()
	{

		memdata::classes::game = *reinterpret_cast<std::uintptr_t*>(memdata::classes::game_pointer);
		memdata::classes::datamodel = reinterpret_cast<rbx::instance*>(memdata::classes::game + 0xC);

		/* Non-Networked DataModel Bypass */

		*reinterpret_cast<std::uint32_t*>(memdata::classes::game + memdata::offsets::game::non_networked) = 3;

		memdata::classes::script_context = functions::instance::find_first_child_by_class(memdata::classes::datamodel, xorstr("ScriptContext"));

		//std::cout << functions::instance::get_classname(memdata::classes::script_context) << "\n";

		auto thread = ptr_obf::deobf(reinterpret_cast<std::uintptr_t>(
			memdata::classes::script_context) + memdata::offsets::lua::state::state, memdata::offsets::lua::state::state_obf);

		memdata::classes::lua::rstate = thread;

		memdata::classes::lua::state = functions::lua::newthread(*memdata::classes::lua::rstate);
		//std::cout << "compadre esta mi muerto des los pantalones\n";
		functions::lua::sandboxthread(memdata::classes::lua::state, temple::config::lua::identity);

		functions::lua::pushvalue(thread, -10002);
		functions::lua::setfield(thread, -10002, xorstr("_G"));
		//std::cout << "caca\n";

		memdata::functions::lua::require.ptr = get_global_function(memdata::classes::lua::state, xorstr("require"));
		memdata::functions::lua::spawn.ptr = get_global_function(memdata::classes::lua::state, xorstr("spawn"));

		//std::cout << std::hex << "spawn should be " << memdata::functions::lua::spawn.ptr - memdata::base + 0x400000 << "\n";

		functions::lua::require.load_data(memdata::functions::lua::require);
		functions::lua::spawn.load_data(memdata::functions::lua::spawn);

	}

	void validate_memory_locations(std::vector<std::ptrdiff_t> numbers)
	{
		memdata::offsets::lua::state::state = numbers[0];
		memdata::functions::lua::load.ptr = numbers[1];
	}

	bool check_teleport()
	{
		if (memdata::classes::game != get_game() && functions::game::is_loaded(get_game()))
		{
			//std::cout << "teleported!\n";
			prepare_memory();

			return true;
		}
		else
		{
			return false;
		}
	}
}