#pragma once
#include <cstdint>
#include <memory>

namespace temple::lua
{
	struct lua_Debug
	{
		const char* name;           /* (n) */
		const char* what;           /* (s) `Lua', `C', `main', `tail' */
		const char* source;         /* (s) */
		int linedefined;            /* (s) */
		int currentline;            /* (l) */
		unsigned char nupvals;      /* (u) number of upvalues */
		unsigned char nparams;      /* (a) number of parameters */
		char isvararg;              /* (a) */
		char short_src[256]; /* (s) */
		void* userdata;             /* only valid in luau_callhook */
	};


	union value_t
	{
		std::uintptr_t gcobj;
		int b;
		double n;
		void* p;
		float v[2];
	};

	struct tvalue
	{
		value_t v;
		int extra;
		int tt;
	}; 

	typedef struct upval {
		std::uintptr_t gc;
		std::uint8_t tt;
		std::uint8_t marked;
		tvalue* v;
		std::uintptr_t unk;
		std::uintptr_t prev;
		std::uintptr_t next;
	} rUpVal;

	struct extraspace
	{
		std::uint32_t padding[6];
		std::uint32_t identity;
		std::uint32_t padding2[5];
		std::shared_ptr<std::uintptr_t> script;
	};

	typedef struct tkey
	{
		value_t value;
		int extra;
		unsigned tt : 4;
		int next : 28; /* for chaining */
	};

	typedef struct luanode
	{
		tvalue val;
		tkey key;
	} luanode;

	struct state
	{
	protected:
		std::uintptr_t ptr = 0;

		public:
		tvalue* &get_top();
		void set_top(tvalue*);
		tvalue* &get_base();
		void set_base(tvalue*);
		extraspace* get_extraspace();
		std::uintptr_t get_closure();
		std::uintptr_t get_l_g();

		__declspec(property(get = get_top, put = set_top)) tvalue*& top;
		__declspec(property(get = get_base, put = set_base)) tvalue*& base;
		__declspec(property(get = get_extraspace)) extraspace* extraspace;
		__declspec(property(get = get_closure)) std::uintptr_t closure;
		__declspec(property(get = get_l_g)) std::uintptr_t l_g;

		operator std::uintptr_t() const
		{
			return ptr;
		}

		std::uintptr_t operator *() const
		{
			return ptr;
		}

		state* operator ->()
		{
			return this;
		}

		std::uintptr_t operator +(std::ptrdiff_t off) const
		{
			return ptr + off;
		}

		std::uintptr_t operator -(std::ptrdiff_t off)
		{
			return ptr - off;
		}

		state(std::uintptr_t ptr) : ptr(ptr) {}
};
}