#include "bitwise.h"

#include <limits.h> 
#include <assert.h>
#include <memory>
#include <string>
#include <iostream>

namespace temple::util::bitwise
{
	std::uint32_t rotl32(std::uint32_t n, unsigned int c)
	{
		const unsigned int mask = (CHAR_BIT * sizeof(n) - 1);

		c &= mask;
		return static_cast<std::uint32_t>((n << c) | (n >> (((~c + 1)) & mask)));
	}

	std::uint32_t rotr32(std::uint32_t n, unsigned int c)
	{
		const unsigned int mask = (CHAR_BIT * sizeof(n) - 1);

		c &= mask;
		return static_cast<std::uint32_t>((n >> c) | (n << (((~c + 1)) & mask)));
	}

	std::uint64_t rotl64(std::uint64_t n, unsigned int c)
	{
		const unsigned int mask = (CHAR_BIT * sizeof(n) - 1);

		c &= mask;
		return static_cast<std::uint64_t>((n << c) | (n >> (((~c + 1)) & mask)));
	}

	std::uint64_t rotr64(std::uint64_t n, unsigned int c)
	{
		const unsigned int mask = (CHAR_BIT * sizeof(n) - 1);

		c &= mask;
		return static_cast<std::uint64_t>((n >> c) | (n << (((~c + 1)) & mask)));
	}

	std::uint64_t bswap64(std::uint64_t n)
	{
		std::uint8_t res[8];
		auto n_d = reinterpret_cast<std::uint8_t*>(&n);

		res[0] = n_d[7];
		res[1] = n_d[6];
		res[2] = n_d[5];
		res[3] = n_d[4];
		res[4] = n_d[3];
		res[5] = n_d[2];
		res[6] = n_d[1];
		res[7] = n_d[0];

		return *reinterpret_cast<std::uint64_t*>(res);
	}

	std::uint32_t bswap(std::uint32_t n)
	{
		std::uint8_t res[4];
		auto n_d = reinterpret_cast<std::uint8_t*>(&n);

		res[0] = n_d[3];
		res[1] = n_d[2];
		res[2] = n_d[1];
		res[3] = n_d[0];

		return *reinterpret_cast<std::uint32_t*>(res);
	}

	std::string to_hex(std::uint64_t n)
	{
		std::string hex_table = "0123456789ABCDEF";
		std::string result = "";

		while (n)
		{
			result = hex_table[n % 16] + result;
			n /= 16;
		}

		return result;
	}
}
