#include "console.h"
#include <xorstr.h>

#include <Windows.h>
#include <iostream>
#include <cstdint>

namespace temple::util::console
{
	void create()
	{
		FILE* stream;

		AllocConsole();
		freopen_s(&stream, xorstr("conout$"), "w", stdout);
		freopen_s(&stream, xorstr("conout$"), "w", stderr);
		freopen_s(&stream, xorstr("conin$"), "r", stdin);

		auto freeconsole = reinterpret_cast<std::byte*>(GetProcAddress(GetModuleHandleA("kernelbase"), "FreeConsole"));
		DWORD old_protect;
		auto addy = freeconsole + 5;

		VirtualProtect(addy, 2, PAGE_EXECUTE_READWRITE, &old_protect);
		*reinterpret_cast<std::uint16_t*>(addy) = 0xC35D;
		VirtualProtect(addy, 2, old_protect, &old_protect);
	}

	void destroy()
	{
		auto freeconsole = reinterpret_cast<std::byte*>(GetProcAddress(GetModuleHandleA("kernelbase"), "FreeConsole"));
		DWORD old_protect;
		auto addy = freeconsole + 5;

		VirtualProtect(addy, 2, PAGE_EXECUTE_READWRITE, &old_protect);
		*reinterpret_cast<std::uint16_t*>(addy) = 0xEC83;
		VirtualProtect(addy, 2, old_protect, &old_protect);

		fclose(stdout);
		fclose(stderr);
		fclose(stdin);

		FreeConsole();
	}

	int get_color_code(const std::string& color_code)
	{
		if (color_code == "@@BLACK@@")
		{
			return 0;
		}
		else if (color_code == "@@BLUE@@")
		{
			return 1;
		}
		else if (color_code == "@@GREEN@@")
		{
			return 2;
		}
		else if (color_code == "@@CYAN@@")
		{
			return 3;
		}
		else if (color_code == "@@RED@@")
		{
			return 4;
		}
		else if (color_code == "@@MAGENTA@@")
		{
			return 5;
		}
		else if (color_code == "@@BROWN@@")
		{
			return 6;
		}
		else if (color_code == "@@LIGHT_GRAY@@")
		{
			return 7;
		}
		else if (color_code == "@@DARK_GRAY@@")
		{
			return 8;
		}
		else if (color_code == "@@LIGHT_BLUE@@")
		{
			return 9;
		}
		else if (color_code == "@@LIGHT_GREEN@@")
		{
			return 10;
		}
		else if (color_code == "@@LIGHT_CYAN@@")
		{
			return 11;
		}
		else if (color_code == "@@LIGHT_RED@@")
		{
			return 12;
		}
		else if (color_code == "@@LIGHT_MAGENTA@@")
		{
			return 13;
		}
		else if (color_code == "@@YELLOW@@")
		{
			return 14;
		}
		else if (color_code == "@@WHITE@@")
		{
			return 15;
		}

		return -1;
	}

	void set_color(int attribute)
	{
		auto console = GetStdHandle(STD_OUTPUT_HANDLE);
		
		SetConsoleTextAttribute(console, attribute);
	}

	void print(const std::string& output, bool endline)
	{
		if (output.substr(0, 2) == "@@")
		{
			auto code = get_color_code(output);

			if (code != -1)
			{
				set_color(code);

				return;
			}
		}

		auto& out = std::cout; // to make sure i dont accidentally comment it

		out << output;
		
		if (endline)
		{
			out << "\n";
		}
	}



	void set_title(const std::string& title)	
	{
		SetConsoleTitleA(title.c_str());
	}
}