#pragma once
#include <xorstr.h>

#include <string>
#include <vector>

namespace temple::overlay::elements
{
	class base;
	class text;
	class quad;
	class line;
	class square;
	class circle;
	class triangle;

	enum class element_type
	{
		text,
		quad,
		line,
		square,
		circle,
		triangle,
		unknown
	};

	class vector2
	{
	public:
		float x = 0, y = 0;

		vector2(float x, float y) : x(x), y(y) {}
		vector2() {}
	};

	class color3
	{
	public:
		uint8_t a = 255, r = 0, g = 0, b = 0;

		uint32_t to_hex();

		color3(uint8_t r, uint8_t g, uint8_t b) : r(r), g(g), b(b) {}
		color3(uint8_t a, uint8_t r, uint8_t g, uint8_t b) : a(a), r(r), g(g), b(b) {}
		color3() {}
	};

	extern std::vector<base*> elements;

	class base
	{
	protected:
		element_type type = element_type::unknown;

	public:
		bool visible = true;
		color3 color;
		element_type get_type();
		void remove();
		void move(size_t);

		base() {};
	};

	class line : public base
	{
	public:
		uint32_t thickness = 1.0f;

		vector2 p1, p2;

		line() { this->type = element_type::line; }
	};

	class circle : public base
	{
	public:
		uint32_t thickness = 1.0f;
		uint32_t edges = 250;
		uint32_t radius = 0;
		bool filled = false;
		vector2 position;

		circle() { this->type = element_type::circle; }
	};

	class square : public base
	{
	public:
		uint32_t thickness = 1.0f;
		bool filled = false;
		vector2 position;
		vector2 size;

		square() { this->type = element_type::square; }
	};

	class triangle : public base
	{
	public:
		uint32_t thickness = 1.0f;
		bool filled = false;
		vector2 p1;
		vector2 p2;
		vector2 p3;

		triangle() { this->type = element_type::triangle; }
	};

	class quad : public base
	{
	public:
		uint32_t thickness = 1.0f;
		bool filled = false;
		vector2 p1;
		vector2 p2;
		vector2 p3;
		vector2 p4;

		quad() { this->type = element_type::quad; }
	};

	class text : public base
	{
	public:
		std::string content;
		vector2 position;
		bool center = false;
		bool outline = false;
		double size = 12;
		vector2 text_bounds;
		color3 outline_color;
		std::string font = xorstr("UI");


		text() { this->type = element_type::text; }
	};
}

