#include "auth.h"

#include <themida/ThemidaSDK.h>
#include "hwid_gen/hwid_gen.h"
#include <crypto/crypto.h>
#include <http/http.h>
#include <xorstr.h>

#include <Windows.h>
#include <optional>
#include <vector>
#include <iostream>
#include <chrono>

namespace temple::auth
{
    auth_session::auth_session() {
        VM_START;

        server = crypto::pk_encryption(crypto::from_hex(xorstr("dc1592bd900595a2003fbc9cedb1ab7007758fcc3c83ce66d5bd1b1b8cc6932c")));
        client = crypto::pk_encryption(crypto::from_hex(xorstr("31ab7cc6af1c1971a71c825d317919b8b911a8b4984af32565bbe5232f2cab25")),
            crypto::from_hex(xorstr("8598422a71ab725130898977468ea594798725845056cc3ecd2cf7b80062a6c7")));
        signed_client = crypto::signature(crypto::from_hex(xorstr("5d303060b096089d65bcbbf383b77f9c840009be72dabe4314c07abcd253e8af")),
            crypto::from_hex(xorstr("c2e227f320fae7f5d088ad10e78f5575387feb596e2312816942aa61e92035cc5d303060b096089d65bcbbf383b77f9c840009be72dabe4314c07abcd253e8af")));
        signed_server = crypto::signature(crypto::from_hex(xorstr("0fa5b06e3aebb49b6900626b3e29fe1fc71a027e111ed084317751aaa3bea241")));

        VM_END;;
    }


    constexpr auto token_size = 200;
    constexpr auto temple_reg_key = "Software\\Temple";
    constexpr auto token_value = "Tok";
    
    constexpr auto username_size = 64;
    constexpr auto username_value = "Username";

    std::optional<HKEY> registry_open_key(HKEY parent_key, char* key)
    {
        DWORD dw;
        HKEY key_handle;
        bool success = RegCreateKeyExA(parent_key, key, 0, nullptr, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, nullptr, &key_handle, &dw) == ERROR_SUCCESS;

        if (!success)
            return std::nullopt;

        return key_handle;
    }
    
    bool registry_write_username(const std::string& uname)
    {
        if (uname.length() > username_size)
            return false;

        auto open_key = registry_open_key(HKEY_CURRENT_USER, xorstr("Software\\Temple"));

        if (!open_key.has_value())
            return false;

        auto success = RegSetValueExA(*open_key, xorstr("Username"), 0, REG_BINARY, reinterpret_cast<const BYTE*>(uname.c_str()), uname.size()) == ERROR_SUCCESS;
        RegCloseKey(*open_key);

        return success;
    }

    std::optional<std::string> registry_get_username()
    {
        auto open_key = registry_open_key(HKEY_CURRENT_USER, xorstr("Software\\Temple"));

        if (!open_key.has_value())
            return std::nullopt;

        std::string uname_buffer(username_size, '\0');
        DWORD size = username_size;

        auto success = RegQueryValueExA(*open_key, xorstr("Username"), nullptr, nullptr, reinterpret_cast<BYTE*>(uname_buffer.data()), &size) == ERROR_SUCCESS;
        RegCloseKey(*open_key);

        if (!success)
            return std::nullopt;

        return uname_buffer.substr(0, size);
    }
    
    bool registry_write_token(const std::string& token)
    {
        if (token.length() != token_size)
            return false;

        auto open_key = registry_open_key(HKEY_CURRENT_USER, xorstr("Software\\Temple"));

        if (!open_key.has_value())
            return false;

        auto success = RegSetValueExA(*open_key, xorstr("Tok"), 0, REG_BINARY, reinterpret_cast<const BYTE*>(token.c_str()), token_size) == ERROR_SUCCESS;
        RegCloseKey(*open_key);

        return success;
    }

    std::optional<std::string> registry_get_token()
    {
        auto open_key = registry_open_key(HKEY_CURRENT_USER, xorstr("Software\\Temple"));

        if (!open_key.has_value())
            return std::nullopt;

        std::string token_buffer(token_size, '\0');
        DWORD size = token_size;

        auto success = RegQueryValueExA(*open_key, xorstr("Tok"), nullptr, nullptr, reinterpret_cast<BYTE*>(token_buffer.data()), &size) == ERROR_SUCCESS;
        RegCloseKey(*open_key);

        if (!success)
            return std::nullopt;

        return token_buffer;
    }

    std::vector<http::parameter> encrypt_parameters(std::vector<http::parameter> unsafe_parameters, const auth_session& session, bool dual_layered)
    {
        VM_START;

        auto time = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();
        auto time_60 = time - (time % 60);

        unsafe_parameters.push_back({ "t", std::to_string(time) });

        std::vector<http::parameter> encrypted_parameters;
        std::string signed_string;

        for (const auto& param : unsafe_parameters)
        {
            auto encrypted = param.value;

            if (dual_layered)
                encrypted = session.swapped_server.encrypt(encrypted);

            auto key = crypto::to_hex(crypto::sha256(param.key + std::to_string(time_60)));
            auto value = crypto::to_hex(session.server.encrypt(encrypted));
            encrypted_parameters.push_back({ key, value });

            signed_string += key + value;
        }
        if (dual_layered)
        {
            auto key = crypto::to_hex(crypto::sha256(xorstr("sid") + std::to_string(time_60)));
            auto value = crypto::to_hex(session.server.encrypt(session.session_id));
            encrypted_parameters.push_back({ key, value });

            signed_string += key + value;
        }
        encrypted_parameters.push_back(
            { crypto::to_hex(crypto::sha256(xorstr("signature") + std::to_string(time_60))), crypto::to_hex(session.signed_client.sign(signed_string)) }
        );

        VM_END;;

        return encrypted_parameters;
    }

    bool decrypt_response(std::string& result, auth_session& session, bool dual_layered)
    {
        VM_START;
        //std::cout << result << " nuggarman\n";
        if (result.length() < crypto_sign_BYTES)
            return false;

        result = crypto::from_hex(result);

        auto signature = result.substr(0, crypto_sign_BYTES);
        auto message = result.substr(crypto_sign_BYTES);

        if (!crypto::signature::verify(message, signature, session.signed_server))
            return false;

        auto opt_result = session.client.decrypt(message);

        if (!opt_result.has_value())
            return false;

        message = *opt_result;

        if (dual_layered)
        {
            auto opt_result = session.swapped_client.decrypt(message);

            if (!opt_result.has_value())
                return false;

            message = *opt_result;
        }

        result = message;

        auto sys_time = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();

        std::int64_t rsp_time = *reinterpret_cast<std::int64_t*>(message.data());

        if (!(sys_time < rsp_time + 60 && sys_time > rsp_time - 60))
            return false;

        result = result.substr(sizeof std::int64_t);

        VM_END;;

        return true;
    }
    bool key_exchange(auth_session& session)
    {
        VM_START;

        auto opt_result = http::post(xorstr("https://temple.cx/api/index.php"), encrypt_parameters(
            {
                {xorstr("protocol"), xorstr("exchkey")},
                {xorstr("pkey"), crypto::to_hex(session.swapped_client.get_public())}
            }, session));

        if (!opt_result.has_value()) {
            //std::cout << "Failed to connect\n";
            return false;
        }

        std::string result = *opt_result;

        if (!decrypt_response(result, session)) {
            //std::cout << "Failed to verify server response\n";
            return false;
        }


        session.session_id = result.substr(0, 32);
        session.swapped_server = crypto::pk_encryption(result.substr(32));

        VM_END;

        return true;
    }
    std::optional<std::vector<std::ptrdiff_t>> token_login(auth_session& session, const std::string& token)
    {
        VM_START;

        std::string tk_login_result;

        auto opt_result = http::post(xorstr(("https://temple.cx/api/index.php")), encrypt_parameters({
            {xorstr("protocol"), xorstr("tklog")},
            {xorstr("tk"), token},
            {xorstr("hwid"), hwid_gen::get_hwid()}
            }, session, true));

        if (!opt_result.has_value()) {
            //std::cout << "Token login failed\n";
            return std::nullopt;
        }

        tk_login_result = *opt_result;
        //std::cout << tk_login_result << " <- obama\n";
        if (!decrypt_response(tk_login_result, session, true)) {
            //std::cout << "Failed login!!1\n";
            return std::nullopt;
        }

        if (tk_login_result.substr(10, crypto_hash_sha512_BYTES * 2) != crypto::to_hex(crypto::sha512(tk_login_result.substr(0, 10) + token))) { // Final check;
            //std::cout << "Failed login!!1d\n";
            return std::nullopt;
        }

        std::vector<std::ptrdiff_t> lotto_numbers;

        auto auth_data = tk_login_result.substr(10 + crypto_hash_sha512_BYTES * 2);
        auto var = *reinterpret_cast<std::uint32_t*>(auth_data.data() + 0);
        auto var2 = *reinterpret_cast<std::uint32_t*>(auth_data.data() + 4);
        lotto_numbers.push_back(var);
        lotto_numbers.push_back(var2);

        VM_END;

        return lotto_numbers;
    }
    bool login(const std::string& username, const std::string& password)
    {
        VM_START;

        auth_session session; // Holds all the session crypto keys

        //std::cout << "Starting authentication..\n";

        if (!key_exchange(session))
            return false;

        auto token = http::post(xorstr("https://temple.cx/api/index.php"), encrypt_parameters({
            {xorstr("protocol"),xorstr("rfshtok")},
            {xorstr("unm"), username},
            {xorstr("pwd"), password},
            {xorstr("hwid"), hwid_gen::get_hwid()}
            }, session, true));

        if (!token.has_value()) {
            //std::cout << "Token failed\n";
            return false;
        }

        if (!decrypt_response(*token, session, true)) {
            //std::cout << "Failed to verify server response2\n";
            return false;
        }

        registry_write_token(*token);

        auto res = token_login(session, *token).has_value();
        
        if(res)
            registry_write_username(username);

        VM_END;

        return res;
    }

    std::optional<std::vector<std::ptrdiff_t>> is_logged_in()
    {
        VM_START;
        auto token = registry_get_token();

        if (!token.has_value())
            return std::nullopt;

        auth_session session;

        if (!key_exchange(session))
            return std::nullopt;

        auto res = token_login(session, *token);

        VM_END;

        return res;
    }
}

extern "C" __declspec(dllexport) bool __cdecl is_logged_in()
{
    return temple::auth::is_logged_in().has_value();
}
std::string to_string_limit(const char* cstring, std::size_t limit)
{
    std::size_t length;

    for (length = 0; length < limit + 1; length++)
    {
        if (cstring[length] == 0) break;
    }

    return std::string(cstring, length);
}
extern "C" __declspec(dllexport) bool __cdecl ui_login(const char* u, const char* p)
{
    return temple::auth::login(to_string_limit(u, 64), to_string_limit(p, 128));
}
