#include <auth/hwid_gen/hwid_gen.h>
#include <crypto/crypto.h>
#include <http/http.h>
#include <xorstr.h>

#include "store.h"
#include "../auth/auth.h"


namespace temple::store
{
    std::optional<std::string> get_script(const std::string& qualified_script_id) {
        auto token = temple::auth::registry_get_token();

        if (!token.has_value())
            return std::nullopt;

        temple::auth::auth_session session;

        if (!temple::auth::key_exchange(session))
            return std::nullopt;

        // Begin request to server

        auto opt_result = http::post(xorstr("https://temple.cx/store/rtscr.php"), temple::auth::encrypt_parameters({
            {xorstr("gid"), qualified_script_id},
            {xorstr("tk"), *token},
            {xorstr("hwid"), temple::auth::hwid_gen::get_hwid()}
            }, session, true));
        if (!opt_result.has_value()) {
            ////std::cout << "Token login failed\n";
            return std::nullopt;
        }

        std::string tk_login_result = *opt_result;
       // //std::cout << tk_login_result << " <- obama\n";
        if (!temple::auth::decrypt_response(tk_login_result, session, true)) {
          //  //std::cout << "Failed login!!1\n";
            return std::nullopt;
        }

        if (tk_login_result.substr(10, crypto_hash_sha512_BYTES * 2) != crypto::to_hex(crypto::sha512(tk_login_result.substr(0, 10) + *token))) { // Final check
          //  //std::cout << "Failed login!!1d\n";
            return std::nullopt;
        }

        return tk_login_result.substr(crypto_hash_sha512_BYTES * 2 + 10);
    }
}