#include "instructionconversion.h"

#include "luauops/luauops.h"

extern "C"
{
#include <lua/lopcodes.h>
}

#include <iostream>
#include <fstream>

//std::ofstream out("C:\\Users\\dicue\\Desktop\\Temple_Instr.txt");
//std::ofstream jout("C:\\Users\\dicue\\Desktop\\Temple_Jumps.txt");

#include <deque>
#include <utility>

namespace convert
{
	using temple::transpiler::luau::luau_instruction;
	using temple::transpiler::luau_opcodes;

	luau_instruction abc(luau_opcodes op, std::uint8_t a, std::uint8_t b, std::uint8_t c)
	{
		luau_instruction instr;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.a = a;
		instr.instruction.b = b;
		instr.instruction.c = c;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)a << " " << (int)b << " " << (int)c << "\n";

		return instr;
	}
	luau_instruction abx(luau_opcodes op, std::uint8_t a, std::uint16_t bx)
	{
		luau_instruction instr;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.a = a;
		instr.instruction.bx = bx;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)a << " " << (int)bx << "\n";

		return instr;
	}
	luau_instruction asbx(luau_opcodes op, std::uint8_t a, std::int16_t sbx)
	{
		luau_instruction instr;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.a = a;
		instr.instruction.sbx = sbx;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)a << " " << (int)sbx << "\n";

		return instr;
	}

	luau_instruction abcn(luau_opcodes op, std::uint8_t a, std::uint8_t b, std::uint8_t c, std::uint32_t N)
	{
		luau_instruction instr;
		instr.has_pseudo = true;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.a = a;
		instr.instruction.b = b;
		instr.instruction.c = c;

		instr.pseudo.abcd = N;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)a << " " << (int)b << " " << (int)c << " " << (int)N << "\n";

		return instr;
	}


	luau_instruction asbxk(luau_opcodes op, std::uint8_t a, std::int16_t sbx, std::uint32_t K)
	{
		luau_instruction instr = 0;
		instr.has_pseudo = true;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.a = a;
		instr.instruction.sbx = sbx;

		instr.pseudo.abcd = K;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)sbx << " " << (int)K << "\n";

		return instr;
	}
	luau_instruction sax(luau_opcodes op, int24_t sax)
	{
		luau_instruction instr;

		instr.instruction.i = static_cast<std::uint8_t>(op);
		instr.instruction.sax = sax;

		////std::cout <<  (int)static_cast<std::uint8_t>(op) << " " << (int)sax<< "\n";

		return instr;
	}
}

namespace temple::transpiler
{
	void instruction_convertor::push_instruction(luau::luau_instruction instruction)
	{
		instruction_list.push_back(instruction.instruction.iabc);

		if (instruction.has_pseudo)
			instruction_list.push_back(instruction.pseudo.abcd);
	}

	std::uint8_t instruction_convertor::constant_to_stack(std::uint32_t r, int idx = 1)
	{
		std::uint8_t newr = p->maxstacksize - idx;

		if (INDEXK(r) > 30000)
			push_instruction(convert::abcn(luau_opcodes::loadkx, newr, 0, 0, INDEXK(r)));
		else
			push_instruction(convert::abx(luau_opcodes::loadk, newr, INDEXK(r)));

		return newr;
	}

	void instruction_convertor::add_jump(int24_t sax)
	{
		push_instruction(convert::sax(luau_opcodes::jmp, sax));
	}

	void instruction_convertor::offset_jumps()
	{
		for (std::size_t i = 0; i < jump_list.size(); i++)
		{
			auto t_i = jump_list[i];

			const auto& it = jump_table.find(t_i);

			if (it == jump_table.end())
				continue;

			auto v_i = it->second;

			luau::instruction jmp(instruction_list[t_i]);

			////std::cout <<  "op: " << luaP_opnames[GET_OPCODE(p->code[v_i])] << "\n";
			////std::cout <<  "current line: " << p->lineinfo[v_i] << "\n";
			////std::cout <<  "current vanilla instr: " << v_i << "\n";
			////std::cout << "vanilla offset: " << jmp.sax << "\n";

			v_i += jmp.sax + 1;

			////std::cout <<  "jump vanilla instr: " << v_i << "\n";

			auto luaui = iequiv_list[v_i];

			////std::cout <<  "jump luau instr: " << luaui << "\n";
			////std::cout << "current luau instr: " << t_i << "\n";

			jmp.sax = luaui - (t_i + 1);

			////std::cout <<  "jump offset: " << jmp.sax << "\n";

			instruction_list[t_i] = jmp.iabc;
		}
	}

	void instruction_convertor::add_jump_entry(int n)
	{
		jump_list.push_back(instruction_list.size() - 1);
		jump_table[instruction_list.size() - 1] = n;
	}

	void output_instruction(Instruction instr)
	{
		////std::cout <<  luaP_opnames[GET_OPCODE(instr)] << " ";

		switch (getOpMode(GET_OPCODE(instr)))
		{
			case iABC:
			{
				////std::cout <<  "iABC " << GETARG_A(instr) << " " << GETARG_B(instr) << " " << GETARG_C(instr) << "\n";
				break;
			}
			case iABx:
			{
				////std::cout <<  "iABx " << GETARG_A(instr) << " " << GETARG_Bx(instr) << "\n";
				break;
			}
			case iAsBx:
			{
				////std::cout <<  "iAsBx " << GETARG_A(instr) << " " << GETARG_sBx(instr) << "\n";
				break;
			}
		}
	}

	std::vector<Instruction> instruction_convertor::convert_instructions()
	{
		std::deque<std::pair<std::uint8_t, Instruction>> self_map;
		std::size_t closure_close = p->nups ? 1 : 0;

		if (p->is_vararg)
		{
			push_instruction(convert::abc(luau_opcodes::clearstackv, p->numparams, 0, 0));
		}

		std::size_t last_instr = 0;

		for (int n = 0; n < p->sizecode; n++)
		{
			const auto i = p->code[n];

			if (iequiv_list.size() <= n)
				iequiv_list.push_back(instruction_list.size());

			////std::cout <<  "current instr v " << n << " luau " << instruction_list.size() << "\n";
			////std::cout <<  "line: " << p->lineinfo[n] << " "; output_instruction(i);

			switch (GET_OPCODE(i))
			{
				case OP_MOVE:
				{
					push_instruction(convert::abc(luau_opcodes::move, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_LOADK:
				{
				
					if(GETARG_Bx(i) > 30000)
						push_instruction(convert::abcn(luau_opcodes::loadkx, GETARG_A(i), 0, 0, GETARG_Bx(i)));
					else
						push_instruction(convert::abx(luau_opcodes::loadk, GETARG_A(i), GETARG_Bx(i)));

					break;
				}
				case OP_LOADNIL:
				{
					//for (std::uint8_t k = GETARG_B(i); k >= GETARG_A(i); k--)
					//	push_instruction(convert::abx(luau_opcodes::loadnil, k, 0));

					auto rb = GETARG_B(i);
					auto ra = GETARG_A(i);
					do {
						push_instruction(convert::abx(luau_opcodes::loadnil, rb--, 0));
					} while (rb >= ra);

					break;
				}
				case OP_LOADBOOL:
				{
					push_instruction(convert::abc(luau_opcodes::loadbool, GETARG_A(i), GETARG_B(i), GETARG_C(i)));

					break;
				}
				case OP_SETGLOBAL:
				{
					push_instruction(convert::abcn(luau_opcodes::setglobal, GETARG_A(i), 0, 0, GETARG_Bx(i)));

					break;
				}
				case OP_GETGLOBAL:
				{
					push_instruction(convert::abcn(luau_opcodes::getglobal, GETARG_A(i), 0, 0, GETARG_Bx(i)));

					break;
				}
				case OP_SETTABLE:
				{
					std::uint8_t reg = GETARG_C(i);

					if (ISK(GETARG_C(i)))
						reg = constant_to_stack(reg);

					if (ISK(GETARG_B(i)) && p->k[INDEXK(GETARG_B(i))].tt == LUA_TSTRING)
					{
						push_instruction(convert::abcn(luau_opcodes::settablek, reg, GETARG_A(i), 0, INDEXK(GETARG_B(i))));
					}
					else
					{
						std::uint8_t reg2 = GETARG_B(i);

						if (ISK(GETARG_B(i)))
							reg2 = constant_to_stack(reg2, 2);


						push_instruction(convert::abc(luau_opcodes::settabler, reg, GETARG_A(i), reg2));
					}

					break;
				}
				case OP_GETTABLE:
				{
					if (ISK(GETARG_C(i)) && p->k[INDEXK(GETARG_C(i))].tt == LUA_TSTRING)
					{
						push_instruction(convert::abcn(luau_opcodes::gettablek, GETARG_A(i), GETARG_B(i), 0, INDEXK(GETARG_C(i))));
					}
					else
					{
						std::uint8_t reg = GETARG_C(i);

						if (ISK(GETARG_C(i)))
							reg = constant_to_stack(reg);

						push_instruction(convert::abc(luau_opcodes::gettabler, GETARG_A(i), GETARG_B(i), reg));
					}

					break;
				}
				case OP_TAILCALL:
				case OP_CALL:
				 {

					if (!self_map.empty() && self_map.back().first == GETARG_A(i))
					{
						const auto c_i = self_map.back().second;

						push_instruction(convert::abcn(luau_opcodes::self, GETARG_A(c_i), GETARG_B(c_i), 0, INDEXK(GETARG_C(c_i))));

						self_map.pop_back();
					}

					push_instruction(convert::abc(luau_opcodes::call, GETARG_A(i), GETARG_B(i), GETARG_C(i)));


					break;
				}
				case OP_SELF: {

					if (ISK(GETARG_C(i)))
					{
						self_map.push_back(std::make_pair(GETARG_A(i), i));
					}
					else
					{
						push_instruction(convert::abc(luau_opcodes::move, GETARG_A(i) + 1, GETARG_B(i), 0));
						push_instruction(convert::abc(luau_opcodes::gettabler, GETARG_A(i), GETARG_B(i), GETARG_C(i)));
					}

					break;

				}
				case OP_CLOSURE: {

					push_instruction(convert::abx(luau_opcodes::closure, GETARG_A(i), GETARG_Bx(i)));

					
					closure_close++;

					if (p->sizep)
					{
						int nups = p->p[GETARG_Bx(i)]->nups;

						for (int idx = 1; idx <= nups; idx++)
						{
							auto c_i = p->code[n + idx];

							iequiv_list.push_back(instruction_list.size());

							switch (GET_OPCODE(c_i))
							{
							case OP_GETUPVAL:
							{
								push_instruction(convert::abc(luau_opcodes::capture, 2, GETARG_B(c_i), 0));

								break;
							}
							case OP_MOVE:
							{
								push_instruction(convert::abc(luau_opcodes::capture, 1, GETARG_B(c_i), 0));

								break;
							}
							}
						}
						n += nups;
					}
					break;
				}
				case OP_GETUPVAL:
				{
					push_instruction(convert::abc(luau_opcodes::getupval, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_SETUPVAL:
				{
					push_instruction(convert::abc(luau_opcodes::setupval, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_NEWTABLE:
				{
					push_instruction(convert::abcn(luau_opcodes::newtable, GETARG_A(i), GETARG_C(i) ? ceillog2(luaO_fb2int(GETARG_C(i))) : 0, 0, luaO_fb2int(GETARG_B(i))));

					break;
				}
				case OP_SETLIST:
				{
					std::uint32_t c = GETARG_C(i);
					if (!c)
						c = p->code[++n];

					push_instruction(convert::abcn(luau_opcodes::setlist, GETARG_A(i), GETARG_A(i) + 1, GETARG_B(i) ? GETARG_B(i) + 1 : 0, (c - 1) * LFIELDS_PER_FLUSH + 1));

					break;
				}
				case OP_NOT:
				{
					push_instruction(convert::abc(luau_opcodes::_not, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_LEN:
				{
					push_instruction(convert::abc(luau_opcodes::len, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_ADD:
				{
					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::addk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::addr, GETARG_A(i), val, GETARG_C(i)));

					break;
				}
				case OP_SUB:
				{
					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::subk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::subr, GETARG_A(i), val, GETARG_C(i)));

					break;
				}
				case OP_MOD:
				{
					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::modk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::modr, GETARG_A(i), val, GETARG_C(i)));

					break;
				}
				case OP_POW:
				{
					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::powk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::powr, GETARG_A(i), val, GETARG_C(i)));

					break;
				}
				case OP_MUL:
				{
					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::mulk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::mulr, GETARG_A(i), val, GETARG_C(i)));


					break;
				}
				case OP_DIV: {

					std::uint8_t val = GETARG_B(i);

					if (ISK(GETARG_B(i)))
						val = constant_to_stack(val);

					if (ISK(GETARG_C(i)))
						push_instruction(convert::abc(luau_opcodes::divk, GETARG_A(i), val, INDEXK(GETARG_C(i))));
					else
						push_instruction(convert::abc(luau_opcodes::divr, GETARG_A(i), val, GETARG_C(i)));

					break;
				}
				case OP_UNM:
				{
					push_instruction(convert::abc(luau_opcodes::unm, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_CONCAT:
				{
					push_instruction(convert::abc(luau_opcodes::concat, GETARG_A(i), GETARG_B(i), GETARG_C(i)));

					break;
				}
				case OP_JMP:
				{

					add_jump(GETARG_sBx(i));

					add_jump_entry(n);
					//iequiv_list.push_back(instruction_list.size() - 1);

					break;
				}
				case OP_EQ:
				{

					std::uint8_t b = GETARG_B(i);
					std::uint8_t c = GETARG_C(i);

					if (ISK(GETARG_B(i)))
						b = constant_to_stack(b);


					if (ISK(GETARG_C(i)))
						c = constant_to_stack(c, 2);

					//iequiv_list.back() = instruction_list.size();

					push_instruction(convert::asbxk(GETARG_A(i) ? luau_opcodes::neq : luau_opcodes::eq, b, 2, c));

					add_jump(1);

					break;
				}
				case OP_LE:
				{
					std::uint8_t b = GETARG_B(i);
					std::uint8_t c = GETARG_C(i);

					if (ISK(GETARG_B(i)))
						b = constant_to_stack(b);

					if (ISK(GETARG_C(i)))
						c = constant_to_stack(c, 2);

					//iequiv_list.back() = instruction_list.size();

					push_instruction(convert::asbxk(GETARG_A(i) ? luau_opcodes::bt : luau_opcodes::le, c, 2, b));

					add_jump(1);

					break;
				}
				case OP_LT:
				{
					std::uint8_t b = GETARG_B(i);
					std::uint8_t c = GETARG_C(i);

					if (ISK(GETARG_B(i)))
						b = constant_to_stack(b);

					if (ISK(GETARG_C(i)))
						c = constant_to_stack(c, 2);

					//iequiv_list.back() = instruction_list.size();

					push_instruction(convert::asbxk(GETARG_A(i) ? luau_opcodes::be : luau_opcodes::lt, c, 2, b));

					add_jump(1);

					break;
				}
				case OP_TEST:
				{
					push_instruction(convert::asbx(GETARG_C(i) ? luau_opcodes::test : luau_opcodes::ntest, GETARG_A(i), 1));

					add_jump(1);
					break;
				}
				case OP_TESTSET:
				{
					push_instruction(convert::asbx(GETARG_C(i) ? luau_opcodes::test : luau_opcodes::ntest, GETARG_B(i), 1));

					add_jump(2);

					push_instruction(convert::abc(luau_opcodes::move, GETARG_A(i), GETARG_B(i), 0));

					break;
				}
				case OP_FORPREP:
				{
					std::uint8_t a = GETARG_A(i);

					push_instruction(convert::abc(luau_opcodes::move, p->maxstacksize - 1, a, 0));
					push_instruction(convert::abc(luau_opcodes::move, a, a + 1, 0)); // limit
					push_instruction(convert::abc(luau_opcodes::move, a + 1, a + 2, 0)); // step
					push_instruction(convert::abc(luau_opcodes::move, a + 2, p->maxstacksize - 1, 0)); //index

					push_instruction(convert::asbx(luau_opcodes::forprep, GETARG_A(i), 1));

					add_jump(1);

					add_jump(GETARG_sBx(i));

					add_jump_entry(n);
					iequiv_list.back() = instruction_list.size() - 1;

					push_instruction(convert::abc(luau_opcodes::move, GETARG_A(i) + 3, GETARG_A(i) + 2, 0));

					iequiv_list.push_back(instruction_list.size() - 1);

					break;
				}
				case OP_FORLOOP:
				{
					push_instruction(convert::asbx(luau_opcodes::forloop, GETARG_A(i), 1));

					add_jump(1);

					add_jump(GETARG_sBx(i));

					add_jump_entry(n);

					break;
				}
				case OP_TFORLOOP:
				{
					push_instruction(convert::asbxk(luau_opcodes::tforloop, GETARG_A(i), 2, GETARG_C(i)));

					add_jump(1);

					break;
				}
				case OP_VARARG:
				{
					push_instruction(convert::abc(luau_opcodes::vararg, GETARG_A(i), GETARG_B(i), 0));

					break;
				}	
				case OP_CLOSE: {
					push_instruction(convert::abc(luau_opcodes::close, GETARG_A(i), 0, 0));
					break;
				}
				case OP_RETURN:
				{
					if (closure_close)
					{
						push_instruction(convert::abc(luau_opcodes::close, 0, 0, 0));
						closure_close--;
					}
					push_instruction(convert::abc(luau_opcodes::retn, GETARG_A(i), GETARG_B(i), GETARG_C(i)));

					break;
				}
			}

			if(is_debug)
			for (; last_instr < instruction_list.size(); last_instr++)
				line_info.push_back(p->lineinfo[n]);
		}

		offset_jumps();

		for (std::size_t i = 0; i < iequiv_list.size(); i++)
		{
			////std::cout << i << " " << iequiv_list[i] << "\n";
		}

		return instruction_list;
	}

	std::vector<int> instruction_convertor::get_line_info()
	{
		return line_info;
	}
}