#pragma once
#include <math.h>
#include <type_traits>
#include <cstring>

#undef min
#undef max

#pragma pack(push, 1)

template <size_t Bits, typename T>
inline constexpr T sign_extend(const T& v) noexcept {
    static_assert(std::is_integral<T>::value, "T is not integral");
    static_assert((sizeof(T) * 8u) >= Bits, "T is smaller than the specified width");
    if constexpr ((sizeof(T) * 8u) == Bits)
    {
        return v;
    }
    else
    {
        using S = struct { signed Val : Bits; };
        return reinterpret_cast<const S*>(&v)->Val;
    }
}

struct int24_t
{
    static int min()
    {
        return -(static_cast<int>(pow(2, 8.00 * sizeof(int24_t)) / 2));
    }
    static int max()
    {
        return (static_cast<int>(pow(2, 8.00 * sizeof(int24_t)) / 2) - 1);
    }

    std::uint8_t val[3];

    int24_t(std::uint32_t v)
    {
        operator=(v);
    }

    int24_t() = default;

    std::int32_t operator=(std::int32_t rhs)
    {
        std::memcpy(val, &rhs, 3);
        return rhs;
    }

    operator std::int32_t()
    {
        std::int32_t res;
        std::memcpy(&res, val, 3);
        return sign_extend<24>(res);
    }
};

#pragma pack(pop)