#include "serializer.h"
#include "compiler/compiler.h"
#include "instructionconversion/instructionconversion.h"
#include "luaucompiler/Luau/UCompiler.h"

extern "C"
{
#include <lua/lua.h>
#include <lua/lstate.h>
#include <lua/lauxlib.h>
}

#include <fstream>

#define TEMPLE_TRANSPILER false

namespace temple::transpiler
{
	enum class constant_type
	{
		nil,
		boolean,
		number,
		string,
	};

	void serializer::write_proto_root(vstream& vs, vstream& vp, Proto* p)
	{
		proto_map[p] = proto_size++;

		p->maxstacksize += 2;

		vp.write_byte(p->maxstacksize);
		vp.write_byte(p->numparams);
		vp.write_byte(p->nups);
		vp.write_byte(p->is_vararg);

		//WriteInstructionz
		
		instruction_convertor ic(p, is_debug);

		auto iv = ic.convert_instructions();

		vp.write_compressed_int(iv.size());

		for (std::size_t i = 0; i < iv.size(); i++)
			vp.write_int(iv[i]);

		vp.write_compressed_int(p->sizek);
			
		for (int i = 0; i < p->sizek; i++)
		{
			TValue* k = &p->k[i];
			switch (k->tt)
			{
				case LUA_TNIL:
				{
					vp.write_byte(static_cast<std::uint8_t>(constant_type::nil));
					break;
				}
				case LUA_TBOOLEAN:
				{
					vp.write_byte(static_cast<std::uint8_t>(constant_type::boolean));
					vp.write_byte(k->value.b);
					break;
				}
				case LUA_TNUMBER:
				{
					vp.write_byte(static_cast<std::uint8_t>(constant_type::number));
					vp.write_double(k->value.n);
					break;
				}
				case LUA_TSTRING:
				{
					TString* ts = &k->value.gc->ts;

					vs.write_compressed_int(ts->tsv.len);
					vs.write_string(getstr(ts), ts->tsv.len);

					vp.write_byte(static_cast<std::uint8_t>(constant_type::string));
					vp.write_compressed_int(++string_size);
					break;
				}
			}
		}

		vp.write_compressed_int(p->sizep);

		for (int i = 0; i < p->sizep; i++)
			vp.write_compressed_int(proto_map[p->p[i]]);

		if (p->name)
		{
			auto name_len = p->name->tsv.len;
			vs.write_compressed_int(name_len);
			vs.write_string(getstr(p->name), name_len);
			vp.write_compressed_int(++string_size);
		}
		else
		{
			vp.write_compressed_int(0);
		}

		vp.write_byte(is_debug ? 1 : 0);

		if (is_debug)
		{
			vp.write_byte(0);

			auto line_info = ic.get_line_info();

			std::size_t lastline = 0;

			for (std::size_t i = 0; i < line_info.size(); i++)
				vp.write_byte(0);

			for (std::size_t i = 0; i < line_info.size(); i++)
			{
				if (line_info[i] != lastline)
				{
					vp.write_int(line_info[i] - lastline);
				}
				else
				{
					vp.write_int(0);
				}
				lastline = line_info[i];
			}
		}

		vp.write_byte(0);
	}

	void serializer::write_proto_children(vstream& vs, vstream& vp, Proto* p)
	{
		for (int i = 0; i < p->sizep; i++)
			write_proto_children(vs, vp, p->p[i]);

		write_proto_root(vs, vp, p);
	}

	void serializer::write_proto(vstream& vs, Proto* p)
	{
		vstream stringlist;
		vstream protolist;

		write_proto_children(stringlist, protolist, p);

		vs.write_compressed_int(string_size);
		vs.add(stringlist);
		vs.write_compressed_int(proto_size);
		vs.add(protolist);
	}

	std::string serializer::serialize(const std::string& s)
	{
		/*
		auto L = lua_open();

		compiler c;
		auto p = c.compile_script(L, s);

		vstream vs;

		if (!p)
		{
			std::string result;
			auto res = std::string(lua_tostring(L, -1));

			res.insert(res.begin(), '\0');

			return res;// std::string(res.length() + 1, ("\0" + res).c_str());
		}

		vs.write_byte(1);

		write_proto(vs, p);

		vs.write_compressed_int(proto_map[p]);

		lua_close(L);

		std::string pb = vs.to_string();

		return pb;*/

		return Luau::compile(s);
	}
}