#include "vectorstream.h"

namespace temple::transpiler
{
	void vstream::write_byte(std::uint8_t b)
	{
		stream.push_back(b);
	}

	void vstream::write_bytes(std::uint8_t* b, const std::size_t s)
	{
		for (std::size_t i = 0; i < s; i++)
			vstream::write_byte(b[i]);
	}

	void vstream::write_int(std::uint32_t i)
	{
		write_bytes(reinterpret_cast<std::uint8_t*>(&i), sizeof std::uint32_t);
	}

	void vstream::write_double(double d)
	{
		std::uint32_t* i = reinterpret_cast<std::uint32_t*>(&d);

		write_int(i[0]);
		write_int(i[1]);
	}

	void vstream::write_compressed_int(std::uint32_t i)
	{
		do
		{
			std::uint8_t b = i & 0x7F;
			i >>= 7;

			if (i)
				b |= 0x80;

			write_byte(b);

		} while (i);
	}

	void vstream::write_string(const char* c, std::size_t l)
	{
		for (std::size_t i = 0; i < l; i++)
			write_byte(c[i]);
	}

	void vstream::add(vstream& vs)
	{
		stream.insert(stream.end(), vs.stream.begin(), vs.stream.end());
	}

	std::string vstream::to_string()
	{
		return std::string(stream.begin(), stream.end());
	}
}